;;; Code for action representation.

(in-package medic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Local variables
;;;

;; Only split operators, and then only for sequential plans.
;; It's incorrect to split fluents.


(defvar *action-representation* 'regular)
(defvar *action-representations* '(regular
				   simple-split
				   simple-split-factored
				   overloaded-split
				   overloaded-split-factored
				   bitwise))
(deftype action-representation ()
  `(member ,@*action-representations*))
;; The old "partially" vs. "fully" factoreed splitting correspond to q vs.
;; regular factoring.


(defun factored-splitting-p ()
  (member *action-representation*
	  '(simple-split-factored overloaded-split-factored)))
(defun overloaded-splitting-p ()
  (member *action-representation*
	  '(overloaded-split overloaded-split-factored)))

(defvar *split-table* nil
  "assoc list of (sprop . list-of-sprops)
To split, replace first sprop by conjunction of others.
To unsplit, first sprop is true iff all others are true.")
(defvar *split-untouched* nil
  "list of sprops representing unsplit operators used by overloaded splitting
An operator is unsplit if only 1 arg or time != :during.")

;; Used to be called split-clauses
;; This also sets *split-table*, *split-untouched*
(defun act-rep-clauses ()
  (setq *split-table* '())
  (setq *split-untouched* '())
  (cond ((eq *action-representation* 'regular)
	 (regular-act-rep-clauses))
	((member *action-representation* '(simple-split
					   simple-split-factored))
	 (simple-split-act-rep-clauses))
	((member *action-representation* '(overloaded-split
					   overloaded-split-factored))
	 (overloaded-split-act-rep-clauses))
	((eq *action-representation* 'bitwise)
	 (bitwise-act-rep-clauses))
	(t
	 (error "Bad value for *action-representation* ~s" *action-representation*))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Regular (unsplit) representation
;;;

(defun regular-act-rep-clauses ()
  (values '() '()))

;; (defun split-formula-none (formula)
;;   formula)
;; 
;; (defun split-cnf-clauses-none (clauses)
;;   clauses)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Splitting
;;;

;; Yuck!  This used to work only on sprops, but binding instantiation
;; occurs before conversion to sprops, and we want this to happen before
;; instantiation.

;; Returns the new split or nil.
;; Return in order, so factored splitting can easily extract first argument.
(defun split-action (action)
  (let ((result (split-action-from-args
		 (sprop-name action) (sprop-args action) (sprop-time action))))
    (and result
	 (cons action (mapcar #'(lambda (lst)
				  (apply #'make-sprop lst))
			      result)))))

;; Returns its result in list, not sprop, form.  Grrr!
(defun split-action-from-args (name non-time-args time-arg)
  (declare (list non-time-args))
  (assert (every #'symbol-or-number-p non-time-args))
  (case *action-representation*
    ((none bitwise)
     (error "Shouldn't be calling splitting"))
    ((simple-split simple-split-factored)
     (split-action-simple-from-args name non-time-args time-arg))
    ((overloaded-split overloaded-split-factored)
     (split-action-overloaded-from-args name non-time-args time-arg))
    (otherwise (error "What action representation? ~s" *action-representation*))))
;; Testing (be sure to set *action-representation* to test all cases):
;; (split-action-from-args 'PUTON '(?X ?Y ?Z) :during)
;; (split-action-from-args 'TABLE '(?W) :during)


(defun make-actionis (op)
  (case *action-representation*
    ((regular)
     (error "Are you sure you want actionis with regular action representation?"))
    ((bitwise)
     (error "Are you sure you want actionis with bitwise action representation?"))
    ((simple-split simple-split-factored)
     (assert (keyword-or-number-p (op-time op)))
     (if (null (op-formals op))
         ;; action has no arguments
         (make-sprop (op-name op) nil :during nil)
       ;; action has arguments
       (let ((disjuncts (mapcar #'(lambda (arg)
				    (simple-split-argnum-sprop op 1 arg :during))
				(first (op-obj-lst op)))))
	 (if (length-one disjuncts)
	     (car disjuncts)
	   (cons :or disjuncts)))))
    ((overloaded-split overloaded-split-factored)
     (if (null (op-formals op))
         ;; action has no arguments
         (make-sprop (op-name op) nil :during nil)
       ;; action has arguments
       (make-actionat-sprop (op-name op) :during)))
    (otherwise
     (error "What action representation? ~s" *action-representation*))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Simple splitting
;;;

(defun simple-split-argnum-name (name argnum)
  (intern (format nil "~S!ARG~D" name argnum)))

(defun simple-split-act-rep-clauses ()
  (let ((substitutions '()))
    ;; actions is a list of non-negated sprops
    (dolist (action *ground-actions*)
      (let ((sub (mapcar #'(lambda (l) (make-sprop-from-list l nil))
			 (split-action-simple-from-args
			  (sprop-name action) (sprop-args action)
			  (sprop-time action)))))
	(if sub
	    (push (cons action sub) substitutions)
	  (push action *split-untouched*))))
    (setq *split-table* substitutions)
    (values '()				; no new formulas
	    substitutions)))

;; Returns its result in list, not sprop, form.  Grrr!
(defun split-action-simple-from-args (name non-time-args time-arg)
  (let ((num-args (length non-time-args)))
    (if (or (< num-args 2) (not time-arg) (not (eq :during time-arg)))
	nil
      (let ((argnum 0))
	(mapcar #'(lambda (arg)
		    (incf argnum)
		    (list (simple-split-argnum-name name argnum) arg time-arg))
		non-time-args)))))
;; Testing
;; (split-action-simple-from-args 'PUTON '(?X ?Y ?Z) :during)
;; (split-action-simple-from-args '(TABLE ?W :during))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Overloaded splitting
;;;

(defun overloaded-argnum-name (argnum)
  (intern (format nil "ARG!~D" argnum)))

(defun overloaded-split-act-rep-clauses ()
  (let ((substitutions '()))
    ;; actions is a list of lists like (PUTON ?X ?Y ?Z)
    ;; split-action-simple assumes the last argument is the time
    (dolist (action *ground-actions*)
      (let ((sub (mapcar #'(lambda (l) (make-sprop-from-list l nil))
			 (split-action-overloaded-from-args
			  (sprop-name action) (sprop-args action)
			  (sprop-time action)))))
	(if sub
	    (push (cons action sub) substitutions)
	  (push action *split-untouched*))))
    (setq *split-table* substitutions)
    (values '()				; no new formulas
	    substitutions)))

;; Never split zero-ary operators;
;; split unary operators only if doing factoring;
;; always split binary and higher-arity operators.

;; Returns its result in list, not sprop, form.  Grrr!
(defun split-action-overloaded-from-args (name non-time-args time-arg)
  (let ((num-args (length non-time-args)))
    (if (or (not time-arg)		; subsumed by the below
	    (not (eq :during time-arg))
	    (zerop num-args)
	    (and (= 1 num-args) (not (factored-splitting-p))))
	nil
      (cons (list 'action name time-arg)
	    (let ((argnum 0))
	      (mapcar #'(lambda (arg)
			  (incf argnum)
			  (list (overloaded-argnum-name argnum) arg time-arg))
		      non-time-args))))))
;; (split-action-overloaded-from-args 'PUTON '(?X ?Y ?Z) :during)
;; (split-action-overloaded-from-args 'TABLE '(?W) :during)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Bitwise representation
;;;

;; returns a "truth table" of size 2^n of the form
;; ((bit2 :AFTER)) (bit1 :BEFORE))
;; ((bit2 :AFTER) (not (bit1 :BEFORE)))
;; ((not (bit2 :AFTER)) (bit1 :BEFORE))
;; ((not (bit2 :AFTER)) (bit1 :BEFORE)))

;; It would be just as easy to write this non-recursively.  For small tables,
;; it really doesn't matter.
(defun get-bit-table (n)
  (if (= n 1)
      (let ((bit1-during (make-sprop (intern "BIT1") '() :during nil)))
	(list (list bit1-during) (list (sprop-negate bit1-during))))
    (let ((next-res (get-bit-table (1- n)))
	  (bit-during (make-sprop (intern (format nil "BIT~a" n)) '() :during nil)))
      (append
       (mapcar #'(lambda (elem)
		   (cons bit-during elem))
	       next-res)
       (let ((not-bit-during (sprop-negate bit-during)))
	 (mapcar #'(lambda (elem)
		     (cons not-bit-during elem))
		 next-res))))))
;; (get-bit-table 3)


;; The resulting table covers all possible bit patterns.
(defun bitwise-act-rep-clauses ()
  ;; Is it sensible to call this for all values of *action-representation*?
  (let* ((actions *ground-actions*)
	 (num-actions (length actions)))
    (if (= num-actions 1)
	(values () ())
      (let* ((table-log-size (ceiling (log num-actions 2)))
	     (table-size (expt 2 table-log-size))
	     (doubled-up (- table-size num-actions))
	     ;; table contains lists of lists of propositions (not formulas)
	     (table (get-bit-table table-log-size))
	     (doubled-entries (loop for i from 1 to doubled-up
				  collect (prog1 (butlast (first table))
					    (assert (equal (butlast (first table))
							   (butlast (second table))))
					    (setq table (cddr table))))))
	(assert (= num-actions (+ (length doubled-entries) (length table))))
	(values
	 '()
	 (mapcar #'cons actions (append doubled-entries table)))))))
