;;; -*- Mode: LISP; Syntax: Common-lisp; Package: USER; Base: 10 -*-
;; Originally /projects/ai/ruler/clause.lisp.
;; Guts torn out by Michael Ernst, to eliminate some functionality and so this
;; doesn't define a read macro which changes the interpretation of symbols
;; prefixed by ?.

;; (in-package 'user)
(in-package medic)

;;; Wff -> clause form conversion
;;; Written by Mitch Sundt, Dec 1988
;;; Modified by Dan Weld, Oct 1989

;; " Copyright 1988, 1989 University of Washington
;; 
;;   All rights reserved. Use of this software is permitted for non-commercial
;;   research purposes, and it may be copied only for that use.  All copies must
;;   include this copyright message.  This software is made available AS IS, and
;;   the authors make no warranty about the software or its performance."

;;; Overall wff-to-clausal form conversion routine
;;; In general, it APPEARS that form-cnf generates a list of clauses.  Each
;;; clause is either a single literal, or a disjunction of several literals
;;; starting with OR.
;;; 
;;; Examples:
;;;  > (form-cnf 'forall (?x) (exists (?y ?z) (or (and ?x ?y) ?z))))
;;;  ((OR ?X6366 (SK-Z65 ?X6366)) (OR (SK-Y64 ?X6367) (SK-Z65 ?X6367)))
;;;
;;;  > (form-cnf '(forall (?x ?y) (or ?x (and (p ?x ?y) (p (f ?x) (f ?y))))))
;;;  ((OR (P ?X5760 ?Y5859) ?X5760) (OR (P (F ?X5762) (F ?Y5861)) ?X5762))
;;;
;;;  > (form-cnf '(forall (?x ?y) (and (p ?x ?y) ?x)))
;;;  ((P ?X4648 ?Y4749) ?X4650)
;;;
;;; NOTE: Requires that (and ...) take no more than two arguments.
;;;   extra arguments are ignored.  Nested ands are ok, of course.
;;;
;;; This is currently implemented as 3 recursive functions called by one
;;; overall driver (above).  Six component steps have been folded into
;;; the deepest function, 'skolemate', while the two outer functions
;;; remove all the :ands' and rename all the variables again.
(defun FORM-CNF ( object )
  (let ((conjuncts (form-clause-list (skolemate (binarize object)))))
    (cons :and conjuncts)))


;;; This function does all the steps from 1 through 7:
;;; I.e. it does not JUST skolemize - it partially converts to clause form
;;;	2. Reduce scopes of negations ...
;;;				Do this by tracking parity and only
;;;				inserting nots before literals.
;;;				Select appropriate connective based upon
;;;				parity of not.
;;;	3. and 4. Standardize variables and Skolemize ...
;;;				Just form skolem functions or univ. var terms
;;;				based on true quantifier form.  This is
;;;				not much more complex than doing them
;;;				separately ... trick is identifying universal
;;;				variables and maintaining list in proper
;;;				nesting order.
;;;	5. and 7. Convert to Prenex form and Drop Universal quantifiers ...
;;;				Just delete all quantifier clauses
;;;				as we standardize their variables.
;;;	6. Form Conjunctive Normal Form ...
;;;				This is easy ... do a bottom-up approach.
;;;				Assume Left and Right terms are in cnf
;;;				since they are returned from a deeper
;;;				recursive call.  Push the current connective
;;;				down in these terms until the full term
;;;				is in cnf.
;;; Data structures: Var-list is a list of pairs.  The car of each pair
;;; is a variable symbol as it appears in the original input wff
;;; If the cdr is an atom (i.e. the pair was (?x . ?x13)) then the
;;; var was universally quantified, otherwise it was existentially quantified.
;;; In this case (i.e. the cdr is a list), the car of the cdr is the name of
;;; skolem function and the rest are the arguments to the skolem function
;;; Thus if the pair is a list of two elements, the var will be replaced by a
;;; Skolem constant.
;;; Note that just because variables LOOK like symbols whose first character
;;; is a "?" doesn't mean that is what they are!  In fact they are pcvar
;;; structures defined in the ruler package in the file DNET.LISP
(defun SKOLEMATE ( wff &optional (not-parity nil) )
  (cond ((atom wff)
	 wff)
	((member (car wff) '(:and :or))   ; Process ands, ors
	 (push-disjunction
	  (true-connective (car wff) not-parity)
	  (skolemate (second wff) not-parity)
	  (skolemate (third wff) not-parity)))

	((eq (car wff) :not)                   ; Handle negations
	 (skolemate (second wff) (not not-parity)))

	(t                                        ; Predicates
	 (if not-parity
	     (list :not wff)
	     wff))))



;;;
;;; Pre: op	::= OR | AND
;;;	 left	::= literal | cnf-expr
;;;	 right	::= literal | cnf-expr
;;;
;;;		These values define what should be returned when expression
;;;		is NOT in cnf form.  We already know that left and right
;;;		are in cnf form, so we look at 'op to see if it is an :or.
;;;		If it is, and either 'left or 'right is a '(:and ...)
;;;		phrase, we pull the :and up, and push the :or down on
;;;		both the left and right children (until it is once again
;;;		in cnf form).
;;;
;;; Post: return term in cnf form
;;;
(defun PUSH-DISJUNCTION ( op left right )
  (cond ((and (eq op :or) (listp left) (eq (car left) :and))
         (list :and
               (push-disjunction :or (second left) right)
               (push-disjunction :or (third left) right)))
        ((and (eq op :or) (listp right) (eq (car right) :and))
         (list :and
               (push-disjunction :or (second right) left)
               (push-disjunction :or (third right) left)))
        (t
         (list op left right))))


;;; Pre: wff in cnf form, with arbitrary groupings of ands
;;; Post: list of clauses (:or' clauses or predicates) from the wff
;; (defun FORM-CLAUSE-LIST ( object )
;;   (cond ((and (listp object) (eq (car object) :and))
;;          (append (form-clause-list (second object))
;;                  (form-clause-list (third object))))
;; 	(t
;; 	 (list object))))

;; More efficient version, doesn't do lots of appends
(defun FORM-CLAUSE-LIST ( object )
  (let ((result '()))
    (labels ((fcl-helper ( obj )
	       (cond ((and (listp obj) (eq (car obj) :and))
		      (fcl-helper (second obj))
		      (fcl-helper (third obj)))
		     (t
		      (push (let ((disjuncts (form-disjunct-list obj)))
			      (if (null (cdr disjuncts))
				  (car disjuncts)
				(cons :or disjuncts)))
			    result)))))
      (fcl-helper object)
      result)))
;; (form-clause-list '(a b c))
;; (form-clause-list '(:or (a b) (c d)))
;; (form-clause-list '(:and (a b) (c d)))
;; (form-clause-list '(:and (a b) (:or (:and (c d) (e f)) (g h))))
;; (form-clause-list '(:and (a b) (:and (:or (c d) (e f)) (g h))))
;; (form-clause-list '(:and (a b) (:and (:and (c d) (e f)) (g h))))

(defun FORM-disjunct-LIST ( object )
  (let ((result '()))
    (labels ((fcl-helper ( obj )
	       (cond ((and (listp obj) (eq (car obj) :or))
		      (fcl-helper (second obj))
		      (fcl-helper (third obj)))
		     (t
		      (push obj result)))))
      (fcl-helper object)
      result)))

;;; Returns the connective this should be in view of parity of external nots
;;;	... may be used as a predicate returning true iff this is a connective
(defun TRUE-CONNECTIVE ( connective not-parity )
  (if not-parity
      (second (member connective '(:and :or :and)))
      (first  (member connective '(:and :or)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Need binarize if converting to CNF using clause.lisp.
;;; If using normal.lisp, there is no need to binarize.

; binarize and/or clauses NOTE:  THIS FUNCTION DIES IF ARGS IS NULL
;; Given a function and a list of arguments, apply the function to the
;; arguments two at a time.
(defun binarize-application (func args)
  (assert (not (null args)))
  (reduce #'(lambda (a1 a2) (list func a1 a2)) args))

;; ;; This is equivalent to a call to the built-in Lisp function reduce.
;; (defun binarize (func args)
;;   (cond
;;    ((null args)
;;     (error "binarize called with null second argument"))
;;    ((null (cdr args))			; (= 1 (length args))
;;     (car args))
;;    ((null (cdr (cdr args)))		; (= 2 (length args))
;;     (cons func args))
;;    (t (list func (car args) (binarize func (cdr args))))))

; recursively binarize an entire formula
;; Given a formula, convert AND and OR with arity not 2 into binary
;; applications.  Converts inside NOT forms, but leaves other lists alone.
(defun binarize (form)
  (cond ((atom form) form)
	((equal '(:and) form) t)
	((equal '(:or) form) nil)
	((eq :not (car form))
	 (list :not (binarize (cadr form))))
	((eq :and (car form))
	 (assert (not (null (cdr form))))
	 (let ((args (remove t (mapcar #'binarize (cdr form)))))
	   (cond ((null args)
		  t)
		 ((member nil args)
		  nil)
		 (t
		  (binarize-application (car form) args)))))
	((eq :or (car form))
	 (assert (not (null (cdr form))))
	 (let ((args (remove nil (mapcar #'binarize (cdr form)))))
	   (cond ((null args)
		  nil)
		 ((member t args)
		  t)
		 (t
		  (binarize-application (car form) args)))))
	(t form)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (form-cnf '(:and (:or (a b) (c d)) (:or (e f) (g h))))
;; (form-cnf '(:or (:and (a b) (c d)) (:and (e f) (g h))))
;; (form-cnf '(:or (:and (a b) (c d) (e f) (g h))))
;; (form-cnf '(:and (:and (a b) (c d)) (:and (e f) (g h))))
