;; convert.lisp -- main function is init-action-and-goal-clauses
(in-package medic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User parameters
;;;

;; can either be 'ordinary or 'extra-vars
(defvar *partial-action-method* 'ordinary)

;; can be 'exists->or or 'exists-skolemize
;; determines which way exists in preconditions get translated
(defvar *exists-strategy* 'exists-skolemize)
;;(defvar *exists-strategy* 'exists->or)


;; Used by alpha-rename to keep track of what quantified variables
;; have been used so far...
(defvar *used-variables* nil)

(defvar *types-strategy* 'eliminate
  "Whether or not we are using the types optimization.
Can be NO-TYPES, TIME-ONLY, or ELIMINATE.")
(defvar *types-strategies* '(no-types time-only eliminate))
(deftype types-strategy ()
  `(member ,@*types-strategies*))

(defvar *infer-types* t
  "non-nil if additional types should be added to operator preconds.
This should always be safe to do, regardless of action representation, splitting, etc.")

(defvar *no-pair-consistency-when-factoring* nil)
(defvar *no-pair-consistency-ever* nil)

;; Not sure whether this makes any difference when doing type elimination.
(defvar *use-unioned-obj-list-naively* nil)

(defvar *use-unioned-obj-list-naively-2* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Local variables
;;;

;; Yuck, there are way too many variables here.  Perhaps, for those used
;; just once, eliminate and move the code that computes them to the client.

(defvar *no-op-action* '(n-o-o-p))
(defvar *no-op-operator*)
(defvar *no-op-sprop*)

;;*domains* is UCPOP's , all the rest are mine
(defvar *objects* '()
  "list of all objects in the world")

(defvar *fluents* '()
  "list of two-element lists, each (fluent-name arity-without-time)")

;; Note that these lists do not include types.
(defvar *true-fluents* '()
  "list of lists of the form (<fluent-name> (<obj1> <obj2>) ...) where
\(FLUENT obj1 obj2\) is necessarily true")
(defvar *false-fluents* '()
  "list of lists of the form (<fluent-name> (<obj1> <obj2>) ...) where
\(FLUENT obj1 obj2\) is necessarily false")
;; Not a great name.
(defvar *variant-fluents* '()
  "list of lists of the form (<fluent-name> (<obj1> <obj2>) ...) where
\(FLUENT obj1 obj2\) may be true or false")
(defvar *possible-fluents* '()
  "Hash table containing all state fluents (positive or negative) that may be true.")

(defvar *unary-types* '()
  "list of all ONE-PLACE static fluent names")
(defvar *unary-types-list* '()
  "list of lists of the form (<type-name> <obj1> <obj2>...) where
obj1 and obj2 are of type TYPE")
;; DANGER! The non-unary-types... variables do not necessarily include eq, neq!
(defvar *non-unary-types* '()
  "list of types that have more than one argument")
(defvar *non-unary-types-list* '()
  "list of lists of the form (<type-name> (<obj1> <obj2>) ...) where
\(TYPE obj1 obj2\) is always true, but TYPE is true of no other arguments")
(defvar *removable-types* '()
  "Types that can be eliminated by type elimination.")

(defvar *ground-actions* '()
  "list of ground actions ONLY; these are non-negated sprops with time :DURING.
See also *ground-effects*")
(defvar *ground-effects* '()
  "list of ground effects of actions in *ground-actions*")
;; Apparently only used by sequential exclusion.
(defvar *operators* '()
  "list of operators, each an op structure" )
(defvar *initial-state* '()
  "list of all possible ground fluents (some negated)")
(defvar *fluent-effect-lst* '()
  "list of lists, where each inner list is (<fluent-name> <op1> <op2>...<op_n>),
where op_i is an operator whose effect has a corresponding (<fluent ?x)" )
(defvar *fluent-precond-lst* '()
  "list of lists, where each inner list is (<fluent-name> <op1> <op2>...<op_n>),
where op_i is the name of an operator whose precond contains (<fluent ?x)
[[This appears to be only for positive occurrences of fluent -- negative
  occurrences are ignored??]]
[[And, it occurs to be [unary] types only, not all fluents??]]")
(defvar *infer-lst* '()
  "list of lists, where each inner list is (<fluent-name> <t1> <t2>...),
where t_i is the type of ?x whenver the effect has (<fluent-name> ?x)" )
(defvar *ground-action-precond-lits* '()
  "list of pairs, each a ground action and a list of all precondition literals")
(defvar *ground-action-effect-lits* '()
  "list of pairs, each a ground action and a list of all effect literals")
(defvar *unioned-obj-lsts* '()
  "for each argument position, list of all objects that can occur in that
position for any operator")

(defvar *fluents-used-as-types* '()
  "List of fluent names that appear in quantifier bindings.")
(defvar *vacuous-fluents-used-as-types* '()
  "Fluents used in quantifier bindings that don't appear elsewhere in the problem.")
(defvar *fluent-appearances-in-effects* '()
  "list of lists, each a fluent-name, the number of action effects it
appears in (negated or positive), and a flag which is true if the counter
hasn't yet been incremented for the current action (when the counter is
incremented, the flag is set to false, so that each action is counted just
once even if the fluent appears multiple times in the effect).
This is used when writing .percent files (only) and shouldn't be depended
on, as the list of operators may change after it is computed.")
(defvar *action-fluents* '()
  "list of lists, each an action-name followed by a list of those fluents
mentioned (positively or negatively) in the action's effect")

(defvar *skolemized-parameters* '()
  "An assoc list of (op-name . number) pairs, where the number is how
many of the parameters were added by skolemization.")

(defconstant *clone-op-suffix* "!!!"
  "String to insert between name and gensym number when cloning an operator.")
;; (defvar *cloned-ops* '()
;;   "List of names of operators that were cloned.")

(defvar *orig-problem-init* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Debugging variables
;;;

(defvar *debug-consistency* nil)
(defvar *debug-consistency-verbose* nil)
(defvar *debug-operator-prune* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Operator structure
;;;


(defun op-action (op)
  (let ((time (op-time op))
	(sans-time (cons (op-name op) (op-formals op))))
    (if time
	(append sans-time (list time))
      sans-time)))
(defun op-time (op)
   (let ((last-param (car (last (op-parameters op)))))
     (and (or (keywordp last-param) (numberp last-param))
	  last-param)))
(defun op-arity (op)
  (assert (= (length (op-formals op)) (length (op-obj-lst op))))
  (length (op-formals op)))
(defun max-arity (ops)
  (apply #'max (mapcar #'op-arity ops)))


; create a new op struct
(defun create-op (operator)
  (let ((new-op (make-operator :name (second operator)
			       :parameters (fourth operator)
			       :formals (fourth operator)
			       :formal-types nil
			       :precondition (sixth operator)
			       :effect (eighth operator))))
    new-op))

;create bindings for an operator
(defun create-bindings (operator)
  (declare (type operator operator))
  (assert (operator-p operator))
  (setf (op-bindings operator) (all-ground (op-formals operator)
					   (op-obj-lst operator))))

;; A simpler, better implementation than the below!  sublis is built in.
;; It even works when FORM is a simple variable.
; instantiate parms with the given bindings.
(defun instantiate (binding form)
  (sublis binding form))
;; destructive version
(defun ninstantiate (binding form)
  (nsublis binding form))

;; ;; This could be a call to subst or nsubst.
;; ; instantiate parms with the given bindings.
;; (defun instantiate (binding form)
;;   (cond
;;    ((null form) '())
;;    ;; Calling varp is inefficient in current implementation; assoc is cheap.
;;    ;;    ((varp form)
;;    ;;     (let ((match (assoc form binding)))
;;    ;;       (if (null match) form (cdr match))))
;;    ;;    ((atom form) form)
;;    ((atom form)
;;     (let ((match (assoc form binding)))
;;       (if (null match) form (cdr match))))
;;    (t (cons (car form) (mapcar #'(lambda (a)
;; 				   (instantiate binding a))
;; 			       (cdr form))))))
;; (defun instantiate-variable (var binding)
;;   (let ((match (assoc var binding)))
;;     (if match (cdr match) var)))
;; ;; Calling varp is less efficient than doing the assoc.
;; ;; (defun instantiate-variable (var binding)
;; ;;   (if (varp var)
;; ;;       (cdr (assoc var b))
;; ;;     var))
;; (defun instantiate-variables (vars binding)
;;   (mapcar #'(lambda (var) (instantiate-variable var binding))
;; 	  vars))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The main routine
;;;

;; For testing.
;; (top-level-convert <filename>) returns a CNF-formula which includes
;; conjunctions of the following things:
;;  1) the AND of all initial conditions, constrained to happen at :BEGINNING
;;  2) the goal formula, with each fluent in it constrained to happen at :END
;;  3) ground versions of (action :DURING) => (preconds :BEFORE) and
;;     (action :DURING) => (effects :AFTER) for all ground instantiations of
;;     all actions in the world.
;;(defun top-level-convert (filename)
;;  (multiple-value-bind (dom prob) (read-problem-file filename)
;;  (multiple-value-bind (init-clauses action-clauses goal-clauses)
;;	(init-action-and-goal-clauses dom prob filename)
;;      (convert-to-cnf (list :and init-clauses goal-clauses action-clauses)))))

;; Return three values:
;;  1) the AND of all initial conditions, constrained to happen at :BEGINNING
;;  2) ground versions of (action :DURING) => (preconds :BEFORE) and
;;     (action :DURING) => (effects :AFTER) for all ground instantiations of
;;     all actions in the world.
;;  3) the goal formula, with each fluent in it constrained to happen at :END
;; Also modifies some global variables.
;; [Might this need to specify substitutions, perhaps for simplicity?
;;  Apparently not so far...]
(defun init-action-and-goal-clauses (prob dom)
  (declare (type problem prob) (type domain dom))
  (assert (problem-p prob))
  (assert (domain-p dom))
  (assert (every #'operator-p (domain-operators dom)))
  (if *show-stages* (format t "~&Entering init-action-and-goal-clauses~%"))


  ;; this stuff used to be done in read-problem-file, but that function
  ;; is no longer used
  (setq *no-op-operator* (make-operator :name (car *no-op-action*)
					:parameters (cdr *no-op-action*)
					:formals (cdr *no-op-action*)
					:formal-types nil
					:precondition (cnf-true)
					:effect (cnf-true)))
  (setq *no-op-sprop* (make-sprop-from-list *no-op-action* nil))
  (if (or (eq 'classical *frame-strategy*)
	  (eq 'bitwise *action-representation*))
      (push *no-op-operator* (domain-operators dom)))

  ;; These should all be lists; I set them to a symbol to catch any attempts
  ;; to use them before they're set.
  (setq *fluents* 'not-yet-set)
  (setq *fluent-appearances-in-effects* 'not-yet-set)
  (setq *fluents-used-as-types* 'not-yet-set)
  (setq *vacuous-fluents-used-as-types* 'not-yet-set)
  (setq *unary-types* 'not-yet-set)
  (setq *unary-types-list* 'not-yet-set)
  (setq *non-unary-types* 'not-yet-set)
  (setq *non-unary-types-list* 'not-yet-set)
  (setq *removable-types* 'not-yet-set)
  (setq *true-fluents* 'not-yet-set)
  (setq *false-fluents* 'not-yet-set)
  (setq *variant-fluents* 'not-yet-set)
  (setq *possible-fluents* 'not-yet-set)
  (setq *skolemized-parameters* 'not-yet-set)

  ;; (setq *initial-state* '())

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; First, precondition/simplify the problem.

  (setq *orig-problem-init* (copy-list (problem-init *problem*)))
  (assert (every #'op-precondition (domain-operators dom)))

  (get-all-fluents-and-types prob dom)

  ;; Push not into formulas
  (setf (problem-init prob) (push-not-into-formula (problem-init prob)))
  (setf (problem-goal prob) (push-not-into-formula (problem-goal prob)))
  (dolist (o (domain-operators dom))
    (setf (op-precondition o) (push-not-into-formula (op-precondition o)))
    (setf (op-effect o) (push-not-into-formula (op-effect o))))

  ;; Set *objects*
  (let* ((objects-in-init (non-function-position (problem-init prob)))
	 (objects-in-goal (non-function-position (problem-goal prob)))
	 ;; Are there allowed to be objects mentioned only in actions?
	 (objects-in-operators
	  (reduce #'union
		  (mapcar #'op-objects (domain-operators dom))))
	 (objects-outside-init (union objects-in-goal objects-in-operators)))
    (setq *objects* (union objects-in-init objects-outside-init))
    (if (> (length *objects*) (length objects-in-init))
	(format t "Warning: not all objects appear in init:~%   ~{ ~S~}~%"
		(set-difference objects-outside-init objects-in-init))))
  ;; (format t "set *objects* to ~s~%" *objects*)

  ;; Eliminate :FORALL
  (setf (problem-init prob) (flatten-ands (forall->and (problem-init prob) prob)))
  (setf (problem-goal prob) (forall->and (problem-goal prob) prob))
  (dolist (o (domain-operators dom))
    (setf (op-precondition o) (forall->and (op-precondition o) prob))
    (setf (op-effect o) (forall->and (op-effect o) prob)))

  (setf (problem-init prob) (cons :and
				  ;; In case the init repeats a proposition.
				  (remove-duplicates
				   (conjunctive-clauses (problem-init prob))
				   :test #'equal)))
  (if (not (simple-init-p (problem-init prob)))
      (error "Init is not a simple conjunction of fluents:~% ~s" (problem-init prob)))

  ;; Eliminate :WHEN
  ;; (setq *cloned-ops* '())
  (setf (domain-operators dom)
	(mapcan #'spawn-op-for-each-when
		(domain-operators dom)))

  ;; Eliminate :EXISTS
  (if (eq *exists-strategy* 'exists-skolemize)
      (progn
	(setq *skolemized-parameters* '())
	(dolist (o (domain-operators dom))
	  (setf (op-precondition o) (exists-skolemize
				     (alpha-rename  (op-precondition o))
				     o))))
    ;; else
    (dolist (o (domain-operators dom))
      (setf (op-precondition o) (exists->or (alpha-rename (op-precondition o))))))

  (setf (problem-goal prob) (exists->or (problem-goal prob)))
  ;; Note that we don't eliminate the exists from operator effects.
  ;; Either of our exists eleiminating strategies would be incorrect
  ;; for removing exists from effects.


  ;; What is the purpose of this?
  ;; Make the precond a (possibly empty or one-element) conjunction
  (dolist (o (domain-operators dom))
    (setf (op-precondition o)
      (cons :and (conjunctive-clauses (flatten-ands (op-precondition o))))))
  (dolist (o (domain-operators dom))
    (setf (op-effect o)
      (cons :and (conjunctive-clauses (flatten-ands (op-effect o))))))


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Done preconditioning/simplifying.
  ;; Compute tables and statistics describing about the domain and problem.

  (setf *fluent-effect-lst* '())
  (setf *fluent-precond-lst* '())
  (dolist (o (domain-operators dom))
    (create-fluent-effect-lst (op-effect o) (op-name o)))
  (dolist (o (domain-operators dom))
    (create-fluent-precond-lst (op-precondition o) (op-name o)))

  ;; More preconditioning; add extra clauses to specify types.
  (if *infer-types*
      (progn
	(setf *infer-lst* (all-inferred-types (problem-init prob)))
	(dolist (op (domain-operators dom))
	  (dolist (fluent+types *infer-lst*)
	    (update-precond-with-types op fluent+types)))))

  ;; It seems that *fluent-effect-lst* is not used below here.
  ;; Why isn't is passed into the above instead of being a global variable?

  ;;   (create-infer-lst-first-pass *fluent-effect-lst*)
  ;;   (setf *fluent-effect-lst* (mapcar #'(lambda (e)
  ;; 					(cons (car e) (cddr e)))
  ;; 				    *fluent-effect-lst*))

  ;;; binding-prune requires that we always compute the types lists -- or
  ;;; rather, that *non-unary-types* (computed in get-fluents) and
  ;;; *non-unary-types-list* are consistent.  Should clean this up!!
  ;; (if (or (eq *types-strategy* 'eliminate) (eq *types-strategy* 'infer))
  ;;     (build-types-list (problem-init prob)))

  (build-types-list (problem-init prob))

  (fill-out-inits prob)

  ;;  (print "Creating and pruning Obj-lsts")

  (dolist (o (domain-operators dom))
    (create-obj-lst o prob)
    ;; (print-operator-long o)
    ;; "some" used to be "every" here.  why?
    (if (some #'null (op-obj-lst o))
	(format t "Warning: empty initial obj-lst for operator ~s~%" (op-name o))
      ;; This used to side-effect *objects*.
      (if (eq *types-strategy* 'eliminate)
	  (progn
	    (prune-obj-lst o (problem-init prob))
	    ;; (print-operator-long o)
	    ;; "some" used to be "every" here.  why?
	    (if (some #'null (op-obj-lst o))
		(format t "Warning: empty obj-lst for operator ~s~%" (op-name o)))))))
  ;; (print (domain-operators dom))

  ;; [sprop change should probably be below here.]


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Create and prune ground actions

  ;;  (print "Create and prune ground actions")

  (dolist (o (domain-operators dom))
    (create-bindings o))

  (dolist (o (domain-operators dom))
    (binding-prune o))

  (if *debug-operator-prune*
      (format t "~d operators: ~s~%" (length (domain-operators dom)) (domain-operators dom)))

  ;; flatten-formula isn't required, but pretties up the result
  (setf (problem-goal prob) (flatten-formula (prune-goal (problem-goal prob))))

  (let ((unpruned-operators
	 (delete-if #'(lambda (op) (trivial-op-p op *debug-operator-prune*))
		    (domain-operators dom))))
     (if (null unpruned-operators)
	 (error "About to eliminate all operators"))
     (setf (domain-operators dom) unpruned-operators))

  (if *debug-operator-prune*
      (format t "~d operators: ~s~%" (length (domain-operators dom)) (domain-operators dom)))

  ;; If we're doing Factoring/Explanatory frames, add NEQ/EQ constraints to
  ;; the initial state.
  ;; NOTE:  It would be better to convert every NEQ appearance to an
  ;; equivalent EQ appearance in the operators and then only add
  ;; EQ constraints to the initial state.
  (if (or (and (factored-splitting-p)
	       (eq *frame-strategy* 'explanatory))
	  (fluent-appears ':eq prob dom)
	  (fluent-appears ':neq prob dom))
      (setf (problem-init prob) (append (problem-init prob)
					(generate-neq-eq-constraints))))

  ;; (print (problem-init prob))

  ;; prune-ground-unary-types takes the place of this
  ;;   ;NOTE: I'M ASSUMING THAT THERE ARE NO TYPES IN ORs IN THE GOAL
  ;;   (if (eq *types-strategy* 'eliminate)
  ;;       (setf (problem-goal prob)
  ;; 	(if (member (car (problem-goal prob)) *unary-types* :test #'eq)
  ;; 	    (cnf-true)
  ;; 	  (rec-remove-if #'(lambda (p)
  ;; 			     (and (listp p)
  ;; 				  (member (car p) *unary-types* :test #'eq)))
  ;; 			 (problem-goal prob)))))

  (prune-ground-unary-types (problem-goal prob))

  (setf (problem-goal prob) (flatten-formula (problem-goal prob)))

  (if (eq *types-strategy* 'eliminate)
      (setf (problem-init prob)
	    ;; apparently this is careful not to eliminate types which still appear
	    (let ((non-appearing-types (eliminated-types prob dom)))
	      ;; (print "inits") (print (problem-init prob))
	      ;; (print "non-appearing-types") (print non-appearing-types)
	      (remove-if #'(lambda (term)
			     (and (listp term)
				  (member
				   (if (eq :not (first term))
				       (first (second term))
				     (first term))
				   non-appearing-types
				   :test #'eq)))
			 (problem-init prob)))))

  ;; Eliminate inconsistent bindings

  ;; Quite possibly this should be run only if inferring types.
  ;; (But sometimes type inference is illegal when this is still safe, I think.)
  ;; If you think pair consistency is a problem, set *no-pair-consistency-when-factoring*.
  (if (not (or (and *no-pair-consistency-when-factoring*
		    (factored-splitting-p))
	       *no-pair-consistency-ever*))
      (pair-consistency-2 prob dom)
    (progn
      (setq *true-fluents* nil)
      (setq *false-fluents* nil)
      (setq *variant-fluents*
	(mapcar #'(lambda (fl)
		    (cons (car fl)
			  (if (zerop (cadr fl))
			      (list ())
			    (combinations (cadr fl) *objects*))))
		*fluents*))
      (setq *removable-types* nil)))

  (if *debug-operator-prune*
      (format t "~d operators: ~s~%" (length (domain-operators dom)) (domain-operators dom)))

  (setf (domain-operators dom)
    (delete-if #'(lambda (op) (trivial-op-p op *debug-operator-prune*))
	       (domain-operators dom)))

  (if *debug-operator-prune*
      (format t "~d operators: ~s~%" (length (domain-operators dom)) (domain-operators dom)))

  ;; I think we're done with bindings, so recompute the obj-lst field
  (dolist (op (domain-operators dom))
    (op-set-obj-lst-from-bindings op))

  (setq *unioned-obj-lsts* (unioned-obj-lsts (domain-operators dom)))

  ;; (print "inits afterward") (print (problem-init prob))

  ;; Side-effecting the operators changes the contents of this list.
  (setq *operators* (domain-operators dom))

  ;; [Maybe this is the place to turn into sprops.]

  (dolist (o (domain-operators dom))
    (setf (op-precondition o) (add-time-to-formula :BEFORE (op-precondition o)))
    (setf (op-effect o) (add-time-to-formula :AFTER (op-effect o)))
    ;; not (add-time-to-formula :DURING (op-parameters o)) as this isn't
    ;; a formula and may be nil.  This is equally gross as testing for
    ;; whether the parameters are null.
    (setf (op-parameters o) (append (op-parameters o) '(:DURING))))
  (setf (problem-init prob)
	(cons :and (mapcar #'(lambda (i) (add-time-to-formula :BEGINNING i))
			   (cdr (problem-init prob)))))
  (setf (problem-goal prob) (add-time-to-formula :END (problem-goal prob)))
  ;; (format t "domain-operators: ~s~%" (domain-operators dom))


  ;; ***** at present this does the sprop conversion *****
  ;; Perhaps convenient not to be sprops yet, for instantiate; but is
  ;; convenient to be sprops, for globals.
  (setf *ground-actions* '())
  (setf *ground-effects* '())
  (setf *ground-action-precond-lits* '())
  (setf *ground-action-effect-lits* '())

  (dolist (o (domain-operators dom))
    (build-globals o))
  (setq *initial-state* (spropify-formulas (cdr (problem-init prob))))
  ;; Not necessary; this should be an ordinary list.
  ;; (setq *no-op-action* (spropify-formula *no-op-action*))

  ;; [Everything should be sprops by now.]

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Actually create the clauses

  (let* ((act-clauses (action-clauses dom))
	 (init-clauses (eliminate-instantiated-non-removable-types-clauses-sprop
			(spropify-formulas (cdr (problem-init prob)))))
	 (cnf-goal (eliminate-instantiated-non-removable-types-sprop
		    (convert-to-cnf (spropify-formula (problem-goal prob)))))
	 (goal-clauses (cond
			((cnf-true-p cnf-goal) ())
			((cnf-false-p cnf-goal)
			 (if (domain-operators dom)
			     (error "Goal is a FALSE formula: ~s" cnf-goal)
			   (error "Goal is a FALSE formula (and no operators could be instantiated): ~s" cnf-goal)))
			(t
			 (if (not (domain-operators dom))
			     (error "No operators could be instantiated, so can't satisfy non-trivial goal:~%~s" cnf-goal))
			 (conjunctive-clauses cnf-goal))))
	 (partial-action-clauses
	  (if (factored-splitting-p)
	      (if (eq *partial-action-method* 'extra-vars)
		  (partial-action-elim-extra-vars (domain-operators dom))
		(partial-action-elim-ordinary (domain-operators dom)))
	    '()))
	 ;; Todd says:
	 ;; If the partial-action-method is 3, then at-least-one clauses are not
	 ;; generated unless we're doing simple-splitting (the (if (eq
	 ;; *action-representation* 'simple-split-factored) above)).  The comment
	 ;; is supposed to only apply to *partial-action-method* Q.
	 (at-least-one-clauses
	  ; NOTE:  Don't need at-least-one when doing overloaded splitting
	  ; (the partial-actions stuff takes care of it)
	  (if (and (eq *frame-strategy* 'classical)
		   (factored-splitting-p))
	      (if (eq *partial-action-method* 'extra-vars)
		  (if (eq *action-representation* 'simple-split-factored)
		      (list (factored-at-least-one-extra-vars (domain-operators dom))))
		(list (factored-at-least-one-ordinary (domain-operators dom))))
	    '())))

    ;; (format t "partial-action-clauses: ~s" partial-action-clauses)

    (assert (cnf-clauses-p init-clauses))
    (assert (cnf-clauses-p act-clauses))
    (assert (cnf-clauses-p goal-clauses))
;;  (print "init-clauses") (print init-clauses)
;;  (print "act-clauses") (print act-clauses)
;;  (print "goal-clauses") (print goal-clauses)
;    (print "ground actions")
;    (print *ground-actions*)
;    (print "ground effects")
;    (print *ground-effects*)
;    (print *ground-action-precond-lits*)
;    (print "here come the effects")
;    (print *ground-action-effect-lits*)

    (assert (every #'spropified-formula-p init-clauses))
    (assert (every #'spropified-formula-p act-clauses))
    (assert (every #'spropified-formula-p goal-clauses))

;;     ;; [hack: convert everything to using sprops here, instead of
;;     ;; doing a smaller number of conversion earlier.]
;;     (setq init-clauses (spropify-formulas init-clauses)
;; 	  act-clauses (spropify-formulas act-clauses)
;; 	  goal-clauses (spropify-formulas goal-clauses)
;; 	  *initial-state* (spropify-formulas *initial-state*)
;; 	  *no-op-action* (spropify-formula *no-op-action*)
;; 	  *ground-actions* (spropify-formulas *ground-actions*)
;; 	  *ground-effects* (spropify-formulas *ground-effects*)
;; 	  *ground-action-precond-lits* (mapcar #'spropify-formulas *ground-action-precond-lits*)
;; 	  *ground-action-effect-lits* (mapcar #'spropify-formulas *ground-action-effect-lits*))

    (if *show-stages* (format t "~&Exiting init-action-and-goal-clauses~%"))


    (values init-clauses
	    ;; act-clauses
	    (append act-clauses partial-action-clauses at-least-one-clauses)
	    goal-clauses)))


(defun simple-init-p (init)
  (and (listp init)
       (eq :and (first init))
       (every #'(lambda (x) (and (listp x)
				 (or (and (eq :not (first x))
					  (listp (second x))
					  (every #'symbolp (second x)))
				     (every #'symbolp x))))
	      (cdr init))))

(defun trivial-op-p (op &optional verbose)
  ;; Don't count the no-op-operator as trivial, as it wouldn't be inserted
  ;; unless we're using classical frame axioms and it is needed for correctness.
  (and (not (eq op *no-op-operator*))
       (or (and (cnf-false-p (op-precondition op))
		(or (not verbose)
		    (progn (format t "Operator ~s has false precondition~%" op)
			   t)))
	   (and (cnf-false-p (op-effect op))
		(or (not verbose)
		    (progn (format t "Operator ~s has false effect~%" op)
			   t)))
	   (and (cnf-true-p (op-effect op))
		(or (not verbose)
		    (progn (format t "Operator ~s has true effect~%" op)
			   t)))
	   (and (null (op-bindings op))
		(or (not verbose)
		    (progn (format t "Operator ~s has no bindings~%" op)
			   t))))))


;;; I think we can now handle these situations.
;; ;; make sure no NEQ/EQ occur within OR, unless there is an EXISTS
;; ;; in between.  Also checks for NOT(exists) and (NOT ...(neq/eq)...).
;; ;; Errs if there is a problem, otherwise returns.
;; (defun caveat-check (form &optional in-not in-or)
;;   (let ((head (first form)))
;;     (cond
;;      ((eq head :and)
;;       (dolist (a (cdr form))
;; 	(caveat-check a in-not in-or)))
;;      ((eq head :or)
;;       (dolist (a (cdr form))
;; 	(caveat-check a in-not t)))
;;      ((eq head :not)
;;       (caveat-check (second form) t in-or))
;;      ((eq head :exists)
;;       (if in-not
;; 	  (error "ERROR: FOUND (NOT (EXISTS...)): ~s" form)
;; 	(dolist (a (cdr form))
;; 	  (caveat-check a in-not nil))))
;;      ((or (eq head :eq) (eq head :neq))
;;       (if (or in-or in-not)
;; 	  (error "ERROR: FOUND ~s WITHIN ~s: ~s" head (if in-or :or :not) form)))
;;      (t nil))))


;; Doesn't test preconditions, only goal and operators
(defun fluent-appears (fluent problem domain)
  (flet ((tester (x) (and (consp x) (eq (car x) fluent))))
    (or (rec-find-if #'tester (problem-goal problem))
	(some #'(lambda (op)
		  (or (rec-find-if #'tester (op-precondition op))
		      (rec-find-if #'tester (op-effect op))))
	      (domain-operators domain)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Functions
;;;

(defun create-obj-lst (op prob)
  (setf (op-obj-lst op)
    (let ((formal-types (op-formal-types op)))
      (mapcar #'(lambda (f)
		  ;; May be modified by destructive operations like type-filter.
		  (objects-of-type (second (assoc f formal-types)) prob))
	      (op-formals op)))))

(defun op-set-obj-lst-from-bindings (op)
  (setf (op-obj-lst op)
    (reduce #'(lambda (lsts b) (mapcar #'adjoin (mapcar #'cdr b) lsts))
	    (op-bindings op)
	    :initial-value (make-list (op-arity op)))))


(defun unioned-obj-lsts (ops)
  (if (null ops)
      '()
    (let* ((n (max-arity ops))
	   (obj-lsts (mapcar #'(lambda (op)
				 (pad-list (op-obj-lst op) n))
			     ops)))
      (apply #'mapcar (cons #'union-of-lists obj-lsts)))))

(defun create-fluent-effect-lst (effect name)
  (let ((fn (car effect)))
    (declare (symbol fn))
    (cond
     ((eq fn :and)
      (mapcar #'(lambda (o)
		  (create-fluent-effect-lst o name))
	      (cdr effect)))
     ((eq fn :not)
      '())
     (t (let ((match (assoc fn *fluent-effect-lst*)))
	  (if (null match)
	      (if (length-two effect)
		  (setq *fluent-effect-lst* (cons (list (car effect)
							(cons name (cdr effect)))
						  *fluent-effect-lst*))
		'())
	    (rplacd match (cons (cons name (cdr effect)) (cdr match)))))))))

;; PRECOND may not be a precond, but any subpart thereof.
(defun create-fluent-precond-lst (precond name)
  (let ((fn (car precond)))
    (declare (symbol fn))
    (cond
     ((eq fn :and)
      (mapcar #'(lambda (o)
		  (create-fluent-precond-lst o name))
	      (cdr precond)))
     ((not (or (eq fn :not) (eq fn :or)))
      (let ((match (assoc fn *fluent-precond-lst*)))
	(if (null match)
	    (if (member fn *unary-types* :test #'eq)
		(push (list fn (cons name (cdr precond)))
		      *fluent-precond-lst*))
	  ;; This looks like (push (cons name (cdr precond)) (cdr match)).  Yes?
	  (rplacd match (cons (cons name (cdr precond)) (cdr match)))))))))


(defun ground-aux (parm objs bindings)
  (mapcar #'(lambda (obj)
	      (cons (cons parm obj) bindings))
	  objs))


; create a list of all possible bindings of the parms with objects
; each binding is a (var . value) pair.
; This should be a fresh list, so it can be modified by side effect.
;; see also combinations-multiple
(defun all-ground (formals obj-lst)
  (declare (list formals) (list obj-lst))
  (assert (= (length formals) (length obj-lst)))
  (assert (every #'varp formals))
  (labels ((all-ground-helper (formals obj-lst current)
	     (if (null formals)
		 current
	       (all-ground-helper (cdr formals) (cdr obj-lst)
				  (mapcan #'(lambda (b)
					      (ground-aux (car formals)
							  (car obj-lst) b))
					  current)))))
    (all-ground-helper (reverse formals) (reverse obj-lst) '(()))))

; tests whether an atom is a variable
;; (would it be faster to memoize the answers in a hash table?)
(defun varp (a)
  (and (atom a)
       (not (numberp a))		; numbers can't be coerced to string
       (let ((astring (string a)))
	 (declare (type simple-base-string astring))
	 (char= (char astring 0) #\?))))


;; return a list of all objects noted in the inits of a problem --
;; that is, all symbols used in non-function-position
(defun get-objects (inits)
  (if (null inits)
      '()
    (reduce #'union (mapcar #'non-function-position inits))))

;; The non-variable symbols used in non-function-position in the sexp
(defun non-function-position (sexp)
  (let ((fn (car sexp)))
    (cond ((or (formula-combiner-p fn) (eq :when fn))
	   ;; union shouldn't be vcalled with zero arguments
	   (if (null (cdr sexp))
	       '()
	     (reduce #'union (mapcar #'non-function-position (cdr sexp)))))
	  ((or (eq :forall fn) (eq :exists fn))
	   (non-function-position (third sexp)))
	  ((or (not (keywordp fn)) (eq :eq fn) (eq :neq fn))
	   ;; DO NOT use delete-if here -- it modifies the expression
	   (delete-duplicates (remove-if #'varp (cdr sexp))))
	  (t
	   (error "What formula combiner? ~s" fn)))))
;; Simpler version only for possibly negated propositions
;; (defun non-function-position (sexp)
;;   (if (eq (car sexp) :not)
;;       (non-function-position (cadr sexp))
;;     (cdr sexp)))
;; Testing
;; (non-function-position '(:and (:or (foo bar baz) (bar baz)) (bim bik (1 2))))

;; List of objects mentioned in the operator.
(defun op-objects (op)
  (union (remove-if #'varp (non-function-position (op-precondition op)))
	 (remove-if #'varp (non-function-position (op-effect op)))))


;; This does not appear to be called any longer
;; ; compile a list of all objects satisfying a certain static predicate
;; (defun get-universe (name inits)
;;   (cond ((null inits) '())
;; 	((equal name (caar inits))
;; 	 (cons (cadar inits) (get-universe name (cdr inits))))
;; 	(t (get-universe name (cdr inits)))))




;; This function sets *fluents*, *unary-types*, *non-unary-types*.
;; *fluent-appearances-in-effects*, *action-fluents*,
;; and writes the .percent file.
;; It determines what the fluents and types are, but not all the objects of
;; those types.

;; This should not be used to determine information which depends on
;; specific operators, only information that is true about fluents,
;; because operators can change (for instance, due to when splitting)
;; after this runs.

(defun get-all-fluents-and-types (prob dom)
  (declare (type problem prob) (type domain dom))
  (setq *action-fluents* (mapcar #'(lambda (o)
				     (list (op-name o)))
				 (remove *no-op-operator* (domain-operators dom))))
  (assert (not (member *no-op-action* *action-fluents* :test #'equal)))
  (setq *fluents* '())
  (setq *fluent-appearances-in-effects* '())
  (setq *fluents-used-as-types* '())
  (setq *vacuous-fluents-used-as-types* '())

  (dolist (o (domain-operators dom))
    (get-fluents (op-effect o) (op-name o)))
  (dolist (o (domain-operators dom))
    (get-fluents (op-precondition o) nil))
  (get-fluents (problem-init prob) nil)
  (get-fluents (problem-goal prob) nil)

  (dolist (type-fluent *fluents-used-as-types*)
    (if (not (assoc type-fluent *fluents*))
	(progn
	  (format t "Warning: quantifier using fluent ~s is vacuous (no such fluent)~%" type-fluent)
	  (push type-fluent *vacuous-fluents-used-as-types*))))

  ;; Now set *unary-types* and *non-unary-types*.
  ;; After this, we are done with *fluents-used-as-types*.
  (setq *unary-types* '())
  (setq *non-unary-types* '())
  (dolist (fluent+arity *fluents*)
    (let* ((fluent-name (car fluent+arity))
	   (appearances-assoc (assoc fluent-name *fluent-appearances-in-effects*))
	   (effects-appearing-in (cdr appearances-assoc)))
      (assert appearances-assoc)
      (if (null effects-appearing-in)
	  ;; Appears in no effects, so it's a type
	  (if (= 1 (second fluent+arity))
	      (push fluent-name *unary-types*)
	    (push fluent-name *non-unary-types*))
	(if (member fluent-name *fluents-used-as-types*)
	    ;; This actually wouldn't be so hard to correct:  when doing
	    ;; forall (and maybe exists) conversion, use all possible objects
	    ;; but add the non-type condition in the body.
	    (error "Fluent ~s used in quantifier binding, but modified by actions:  ~s" fluent-name effects-appearing-in)))))

  ;; This seems wrong; perhaps just guard uses of *-types* if overloaded factoring.
  ;; Make sure nothing is considered a type if we're doing overloaded factoring.
  (if (eq *action-representations* 'overloaded-split-factored)
      (progn
	(setq *unary-types* '())
	(setq *non-unary-types* '())))

  ;; add NEQ/EQ to *fluents* if they are not already in there but should be
  (if (and (factored-splitting-p)
	   (eq *frame-strategy* 'explanatory))
      (progn
	(if (not (member '(:neq 2) *fluents* :test #'equal))
	    (push '(:neq 2) *fluents*))
	(if (not (member '(:eq 2) *fluents* :test #'equal))
	    (push '(:eq 2) *fluents*))))

  ;; Output domain-name.stat file which has info obtained in
  ;; *fluent-appearances-in-effects* and *action-fluents*
  (if (member 'percent *output-files*)
      (write-domain-percent-stat-file
       (domain-name dom) (length *fluents*)
       (let ((num-ops (length (domain-operators dom))))
	 (if (some #'(lambda (o)
		       (eq (op-name o)
			   (car *no-op-action*)))
		   (domain-operators dom))
	     (- num-ops 1) num-ops))
       (append *unary-types* *non-unary-types*)
       (problem-name prob)))
  )



;; Determine names and arities of all fluents in the formula.
;; OP-NAME is non-nil if we are examining effects, NIL otherwise (which
;; includes preconditions, when guards, and initial and goal formulas).

;; This function modifies *fluents*,
;; *fluent-appearances-in-effects*, *action-fluents*, *fluents-used-as-types*.

(defun get-fluents (form op-name)
  (declare (list form) (symbol op-name))
  (let ((fn (car form))
	(arity (1- (length form))))	; faster than (length (cdr form)) due to types
    (declare (symbol fn))
    (cond
     ((formula-combiner-p fn)
      (dolist (subform (cdr form))
	(get-fluents subform op-name)))
     ((or (eq fn :forall) (eq fn :exists))
      (assert (= 2 arity))
      (let ((types (rec-retrieve-atom-if #'(lambda (x) (not (varp x)))
					 (second form))))
	(setq *fluents-used-as-types* (union types *fluents-used-as-types*)))
      (get-fluents (third form) op-name))
     ((eq fn :when)
      (assert (= 2 arity))
      (get-fluents (second form) nil)
      (get-fluents (third form) op-name))
     ((and (or (eq fn :neq) (eq fn :eq))
	   (not (and (factored-splitting-p)
		     (eq *frame-strategy* 'explanatory))))
      (assert (= 2 arity))
      '())
     ((keywordp form)
      (error "What keyword in function position? ~s" form))
     (t
      (let ((fluents-assoc (assoc fn *fluents*)))
	(if fluents-assoc
	    (progn
	      ;; this fluent has already been seen
	      (if (not (eq arity (second fluents-assoc)))
		  (error "fluent ~s used with multiple arities ~s and ~s"
			 fn (second fluents-assoc) arity))
	      (if op-name
		  ;; this occurrence appears in an effect
		  (let ((lookup1 (assoc fn *fluent-appearances-in-effects*))
			(lookup2 (assoc op-name *action-fluents*)))
		    (pushnew op-name (cdr lookup1))
		    (pushnew fn (cdr lookup2)))))
	  ;; this fluent hasn't yet been seen
	  (progn
	    (push (list fn arity) *fluents*)
	    (if op-name
		;; this occurrence appears in an effect
		(progn
		  (push (list fn op-name) *fluent-appearances-in-effects*)
		  (let ((lookup (assoc op-name *action-fluents*)))
		    ;; (rplacd lookup (adjoin fn (cdr lookup)))
		    (pushnew fn (cdr lookup))))
	      (push (list fn) *fluent-appearances-in-effects*)))))))))


;; Old version
;; (defun get-fluents (form in-effect op-name)
;;   (declare (list form) (type bool in-effect) (symbol op-name))
;;   (let ((fn (car form)))
;;     (declare (symbol fn))
;;     (cond
;;      ((formula-combiner-p fn)
;;       (mapcar #'(lambda (x) (get-fluents x in-effect op-name))
;; 	      (cdr form)))
;;      ((and (or (eq fn :neq) (eq fn :eq))
;; 	   (not (and (factored-splitting-p)
;; 		     (eq *frame-strategy* 'explanatory))))
;;       '())
;;      ((assoc fn *fluents*)	; fn is first element of a list in *fluents*
;;       (if (and op-name in-effect)
;; 	(let ((lookup1 (cdr (assoc fn *fluent-appearances-in-effects*)))
;; 	      (lookup2 (assoc op-name *action-fluents*)))
;; 	  (pushnew fn (cdr lookup2))
;; 	  (if (second lookup1)
;; 	      (progn (incf (first lookup1))
;; 		     (setf (second lookup1) nil))))))
;;      ((or (not in-effect) (not op-name))
;;       (if (length-two form)
;; 	  (push fn *unary-types*)
;; 	(push fn *non-unary-types*))
;;       (push (list fn (1- (length form))) ; faster than (length (cdr form)) due to types
;; 	    *fluents*))
;;      (t
;;       (push (list fn (length (cdr form)))
;; 	    *fluents*)
;;       (push (list fn 1 nil) *fluent-appearances-in-effects*)
;;       (let ((lookup (assoc op-name *action-fluents*)))
;; 	(rplacd lookup (adjoin fn (cdr lookup))))))))



; Add time parameter TIME to each predicate in FORMULA.
;; FORMULA is not in sprops form.
(defun add-time-to-formula (time formula)
  (declare (cons formula))
  (assert (not (null formula)))
  (cond ((formula-combiner-p (car formula))
	 (cons (car formula) (mapcar #'(lambda (e) (add-time-to-formula time e))
				     (cdr formula))))
	((and (member *types-strategy* '(time-only eliminate))
	      ;; only substitute :none for fluents, not actions of same name
	      (not (eq time :during))
	      (or (member (car formula) *unary-types* :test #'eq)
		  (member (car formula) *non-unary-types* :test #'eq)))
	 (append formula '(:none)))
	(t
	 (append formula (list time)))))


;; Build up the global variables *ground-actions*, *ground-effects*,
;; *ground-action-precond-lits*, and *ground-action-effect-lits*
(defun build-globals (o)
  (let ((names (mapcar #'(lambda (b)
			   (spropify-formula (instantiate b (op-action o))))
		       (op-bindings o)))
	(pres (mapcar #'(lambda (b)
			  (spropify-formula (instantiate b (op-precondition o))))
		      (op-bindings o)))
	(effs (mapcar #'(lambda (b)
			  (spropify-formula (instantiate b (op-effect o))))
		      (op-bindings o))))
    (dolist (n names)
      (push n *ground-actions*))
    (dolist (e effs)
      (push e *ground-effects*))
    (mapcar #'(lambda (n p)
		(push (cons n (mapcar #'sprop-negate
				      (all-lits-no-time p)))
		      *ground-action-precond-lits*))
	    names pres)
    (mapcar #'(lambda (n e)
		(push (cons n (all-lits-no-time e))
		      *ground-action-effect-lits*))
	    names effs)))

; make the operator clauses
(defun op-action-clauses-nonground (op)
  (list :and
	(make-implies (op-action op)
		      ;; could simplify this, but result is converted to CNF
		      (op-precondition op))
	(make-implies (op-action op)
		      (op-effect op))))

(defun action-clauses (domain)
  (mapcan #'action-clauses-for-op (domain-operators domain)))


;; Returns a list of CNF clauses, not a formula
;; Side-effects *ground-actions* and *ground-effects*.
(defun action-clauses-for-op (o)
  (declare (type operator o))
  (assert (operator-p o))
  (let ((cnf-formula (factor-cnf-clauses
		      o (convert-to-cnf (op-action-clauses-nonground o)))))
    (cond
     ((cnf-true-p cnf-formula) ())
     ((cnf-false-p cnf-formula)
      (error "Found a false clause during CNF-conversion in ACTION-CLAUSES"))
     (t
      (let ((cnf-grounds (mapcar #'(lambda (b)
				     (let ((instantiated (instantiate b cnf-formula)))
				       (eliminate-instantiated-non-removable-types-sprop
					(spropify-formula
					 instantiated))))
				 (if (and *use-unioned-obj-list-naively-2*
					  (factored-splitting-p))
				     ;; not necessarily the original, unpruned
				     ;; bindings, but a fully orthogonal,
				     ;; regular set of bindings (ie, every
				     ;; possible set of combinations)
				     (if (and (overloaded-splitting-p)
					      *use-unioned-obj-list-naively*)
					 (all-ground (op-formals o)
						     (firstn
						      (length (op-formals o))
						      *unioned-obj-lsts*))
				       (all-ground (op-formals o) (op-obj-lst o)))
				   (op-bindings o)))))

	(if (and (not *use-unioned-obj-list-naively-2*)
		 (factored-splitting-p))
	    (progn
	      ;; For att-log1, 70 bindings in all are omitted.
	      ;;  (format t "omitted bindings: ~d~%"
;; 	            (length (set-difference (all-ground (op-formals o) (op-obj-lst o))
;; 	      			      (op-bindings o)
;; 	       			      :test #'equal)))
 	      (setq cnf-grounds
		(append (mapcar #'(lambda (b)
				    ;; instantiation is trivial since the bindings
				    ;; are in order and we're only trying to
				    ;; instantiate the action, not an arbitrary
				    ;; formula
				    (cons :or
					  (mapcar
					   #'(lambda (l)
					       (make-sprop-from-list l t))
					   (split-action-from-args
					    (op-name o) (mapcar #'cdr b) :during))))
				(set-difference
				 (all-ground (op-formals o) (op-obj-lst o))
				 (op-bindings o)
				 :test #'equal))
			cnf-grounds))))

	;; This is in leiu of, for overloaded splitting, mapping over
	;; (all-ground (op-formals o) *unioned-obj-lsts*) just above.
	;; (It's also smaller.)
	;; It would We could do this more aggressively by checking
	;; above whether any precond/effect was not satisfied; if so, then don't
	;; generate all the action => precond/effect clauses.  This can reduce
	;; formula size quite a bit.
	(if (not *use-unioned-obj-list-naively*)
	(if (and (factored-splitting-p) (overloaded-splitting-p))
	    (let* ((actionat (make-actionat-sprop (op-name o) :during))
		   (not-actionat (sprop-negate actionat)))
	      (loop for impossible-args in (mapcar #'set-difference
						   *unioned-obj-lsts*
						   (op-obj-lst o))
		  as i = 1 then (1+ i)
		  do (dolist (arg impossible-args)
		       (declare (fixnum i))
		       (push (list :or not-actionat
				   (sprop-negate (overloaded-split-argnum-sprop
						  i arg :during)))
			     cnf-grounds))))))
	(assert (every #'cnf-p cnf-grounds))
	(apply #'append
	       (mapcar #'conjunctive-clauses cnf-grounds)))))))

;; NOTE: CNF-CLAUSES IS ACTUALLY A CNF FORMULA, NOT A SET OF CLAUSES!
;; Every element of cnf-clauses is a disjunction one of whose elements is
;; the negation of the specified action (and the other elements of which
;; are possibly negated literals using the action's variables).
;; These are non-ground CNF clauses in list (not sprop) form.
(defun factor-cnf-clauses (op cnf-clauses)
  (if (not (factored-splitting-p))
      cnf-clauses
    (let ((split-action (split-action-from-args
			 (op-name op) (op-formals op) (op-time op))))
      (if (not split-action)
	  cnf-clauses
	(let ((var-alist (if (overloaded-splitting-p)
			     ;; No variable is associated with "actionis", but
			     ;; we choose the first element of the list to
			     ;; include if no other element is, so that at least
			     ;; one part of the action is required to occur
			     ;; in order for the action to occur.  (Eg, suppose
			     ;; sawing always makes the floor dusty, independent
			     ;; of what is being sawed.)
			     (cons (cons (gensym "no-var") (car split-action))
				   (mapcar #'(lambda (split)
					       (assert (varp (second split)))
					       (cons (second split) split))
					   (cdr split-action)))
			   (mapcar #'(lambda (split)
				       (assert (varp (second split)))
				       (cons (second split) split))
				   split-action)))
	      (negated-action (list :not (op-action op))))
	  ;; (format t "split-action: ~s~%" split-action)
	  ;; (format t "var-alist: ~s~%" var-alist)
	  (assert (every #'(lambda (disjunction) (member negated-action disjunction :test #'equal)) (conjunctive-clauses cnf-clauses)))
	  (cons
	   :and
	   (delete-duplicates
	    (mapcar #'(lambda (disjunction)
		       (add-factored-action-to-cnf-disjunction
			;; This can't happen earlier because delete removes
			;; elements at the top level, not recursively.
			(delete negated-action disjunction :test #'equal)
			var-alist))
		    (conjunctive-clauses cnf-clauses))
	    :test #'equal)))))))

;; The action's negation has already been deleted; reinsert the disjunction
;; of the negations of the appropriate parts.
(defun add-factored-action-to-cnf-disjunction (disjunction var-alist)
  (declare (cons disjunction))
  (assert (and (consp disjunction) (eq :or (car disjunction))))
  (let ((used-vars '()))
    (dolist (disjunct (cdr disjunction))
      (dolist (arg (butlast (cdr (if (eq :not (car disjunct))
				     (second disjunct)
				   disjunct))))
	(pushnew arg used-vars)))
    ;; (format t "used-vars = ~s for ~s~%" used-vars disjunction)
    (let ((added nil))
      (dolist (var used-vars)
	(declare (symbol var))
	(let ((var-assoc (assoc var var-alist)))
	  (if var-assoc
	      (progn
		(setq added t)
		(push (list :not (cdr var-assoc))
		      ;; put it after the :or
		      (cdr disjunction))))))
      ;; If not dependent on anything, then arbitrarily add the first element.
      ;; This is actionis in the case of overloaded splitting.
      ;; [Should I be using make-actionis here instead?]
      (if (or (not added) (overloaded-splitting-p))
	  (push (list :not (cdr (first var-alist)))
		(cdr disjunction)))))
  disjunction)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Fluent truth tests
;;;

;; The non-sprop versions aren't currently in use.
;; There is a big danger with using nsubst-if anyway, as it can compare
;; against the cdr of a fluent, mistaking it for another fluent.
;; ;; Instantiated non-removable types would be eliminated by DPLL anyway;
;; ;; this does make action clauses a tad smaller.
;; (defun eliminate-instantiated-non-removable-types (cnf-formula)
;;   (assert (cnf-p cnf-formula))
;;   (if (or (factored-splitting-p)
;; 	  (not (eq *types-strategy* 'eliminate)))
;;       cnf-formula
;;     (progn
;;       (setq cnf-formula
;; 	(nsubst-if (cnf-true) #'true-fluent-or-negation-p cnf-formula))
;;       (setq cnf-formula
;; 	(nsubst-if (cnf-false) #'false-fluent-or-negation-p cnf-formula))
;;       ;; Ouch, expensive!  Is there an easier way?
;;       ;; (Maybe the formulas are small enough that this isn't so horrible.)
;;       (convert-to-cnf cnf-formula))))
;; 
;; (defun eliminate-instantiated-non-removable-types-clauses (cnf-clauses)
;;   (conjunctive-clauses (eliminate-instantiated-non-removable-types
;; 			(cons :and cnf-clauses))))

;; WARNING!  This operates on clauses, the above on a formula
(defun eliminate-instantiated-non-removable-types-sprop (cnf-formula)
  (assert (cnf-p cnf-formula))
  (if (or (factored-splitting-p)
	  (not (eq *types-strategy* 'eliminate)))
      cnf-formula
    (progn
      (setq cnf-formula (nsubst-if (cnf-true) #'true-sprop-p cnf-formula))
      (setq cnf-formula (nsubst-if (cnf-false) #'false-sprop-p cnf-formula))
      ;; Ouch, expensive!  Is there an easier way?
      ;; (Maybe the formulas are small enough that this isn't so horrible.)
      (convert-to-cnf cnf-formula))))

;; WARNING!  This operates on clauses, the above on a formula
(defun eliminate-instantiated-non-removable-types-clauses-sprop (cnf-clauses)
  ;; (assert (not (and (consp cnf-clauses) (member (car cnf-clauses) '(:not :and :or)))))
  (assert (cnf-clauses-p cnf-clauses))
  (conjunctive-clauses (eliminate-instantiated-non-removable-types-sprop
			(cons :and cnf-clauses))))

(defun contains-true-or-false-sprop-p (f)
  (and (not (factored-splitting-p))
       (eq *types-strategy* 'eliminate)
       (rec-find-if #'(lambda (x) (and (sprop-p x)
				       (or (true-sprop-p x)
					   (false-sprop-p x))))
		    f)))

(defun false-timeless-formula-p (f)
  (let ((fn (car f)))
    (cond ((eq :and fn)
	   (some #'false-timeless-formula-p (cdr f)))
	  ((eq :or fn)
	   (every #'false-timeless-formula-p (cdr f)))
	  ((eq :not fn)
	   (true-timeless-formula-p (second f)))
	  ((eq :eq fn)
	   (and (not (eq (second f) (third f)))
		(not (varp (second f)))
		(not (varp (third f)))))
	  ((eq :neq fn)
	   (and (eq (second f) (third f))
		(not (varp (second f)))
		(not (varp (third f)))))
	  (t
	   nil
	   ;; when this is called, it's too early for this test
	   ;; (false-timeless-fluent-or-negation-p f)
	   ))))

(defun true-timeless-formula-p (f)
  (let ((fn (car f)))
    (cond ((eq :and fn)
	   (every #'true-timeless-formula-p (cdr f)))
	  ((eq :or fn)
	   (some #'true-timeless-formula-p (cdr f)))
	  ((eq :not fn)
	   (false-timeless-formula-p (second f)))
	  ((eq :eq fn)
	   (and (eq (second f) (third f))
		(not (varp (second f)))
		(not (varp (third f)))))
	  ((eq :neq fn)
	   (and (not (eq (second f) (third f)))
		(not (varp (second f)))
		(not (varp (third f)))))
	  (t
	   nil
	   ;; when this is called, it's too early for this test
	   ;; (true-timeless-fluent-or-negation-p f)
	   ))))


;; (defun cnf-clauses-eliminate-false-fluents-sprop (cnf-clauses)
;;   (dolist (disjunction cnf-clauses)
;;     (if (eq :or (car disjunction))
;; 	(setf (cdr disjunction)
;; 	  ;; Delete only examines top-level members of the list.
;; 	  (delete-if #'false-sprop-p
;; 		     (cdr disjunction)))))
;;   cnf-clauses)

;;; This code is so old it should almost certainly all be deleted.
;; (defun cnf-clauses-eliminate-positive-impossible-fluents (cnf-clauses)
;;   (dolist (disjunction cnf-clauses)
;;     (if (eq :or (car disjunction))
;; 	(setf (cdr disjunction)
;; 	  ;; Delete only examines top-level members of the list.
;; 	  (delete-if #'(lambda (x)
;;                       (declare (cons x))
;; 			 (assert (consp x) (x) "expected a cons: ~s" x)
;; 			 (impossible-fluent x))
;; 		     (cdr disjunction)))))
;;   cnf-clauses)
;; 
;; (defun cnf-eliminate-positive-impossible-fluents (cnf-formula)
;;   (assert (eq :and (car cnf-formula)))
;;   (cnf-clauses-eliminate-positive-impossible-fluents (cdr cnf-formula))
;;   cnf-formula)
;; 
;; (defun cnf-clauses-eliminate-positive-impossible-fluents-sprop (cnf-clauses)
;;   (dolist (disjunction cnf-clauses)
;;     (if (eq :or (car disjunction))
;; 	(setf (cdr disjunction)
;; 	  ;; Delete only examines top-level members of the list.
;; 	  (delete-if #'(lambda (x)
;;                       (declare (type sprop x))
;; 			 (assert (sprop-p x))
;; 			 (and (not (sprop-negated-p x))
;; 			      (let ((fluents-assoc (assoc (sprop-name x) *possibly-true-fluents*)))
;; 				(and fluents-assoc
;; 				     (not (member (sprop-args x)
;; 						  (cdr fluents-assoc)
;; 						  :test #'equal))))))
;; 		     (cdr disjunction)))))
;;   cnf-clauses)
;; 
;; (defun cnf-eliminate-positive-impossible-fluents-sprop (cnf-formula)
;;   (assert (eq :and (car cnf-formula)))
;;   (cnf-clauses-eliminate-positive-impossible-fluents-sprop (cdr cnf-formula))
;;   cnf-formula)

;; For every fluent or negated fluent, one of these is true:
;;   {variant,true,false}-fluent-or-negation-p

(defun variant-fluent-or-negation-p (f)
  (and (consp f)
       (variant-timeless-fluent-or-negation-p (butlast f))))
(defun variant-timeless-fluent-or-negation-p (f)
  (and (consp f)
       (variant-timeless-nonnegated-fluent-p (if (eq :not (car f)) (second f) f))))
(defun variant-nonnegated-fluent-p (f)
  (variant-timeless-nonnegated-fluent-p (butlast f)))
(defun variant-timeless-nonnegated-fluent-p (f)
  (variant-fluent-name-args-p (car f) (cdr f)))
(defun variant-fluent-name-args-p (name args)
  (declare (symbol name) (list args))
  (member-equal args (cdr (assoc name *variant-fluents*))))

;; A non-unary fluent or its negation that must be true.
;; Not the negation of false-fluent-or-negation-p!
(defun true-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (false-nonnegated-fluent-p (second lit))
	 (true-nonnegated-fluent-p lit))))

;; A non-unary fluent or its negation that must be false.
;; Not the negation of true-fluent-or-negation-p!
(defun false-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (true-nonnegated-fluent-p (second lit))
	 (false-nonnegated-fluent-p lit))))

;; YUCK!
(defun true-timeless-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (false-nonnegated-fluent-p (append (second lit) '(:none)))
	 (true-nonnegated-fluent-p (append lit '(:none))))))
(defun false-timeless-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (true-nonnegated-fluent-p (append (second lit) '(:none)))
	 (false-nonnegated-fluent-p (append lit '(:none))))))

(defun true-sprop-p (prop)
  ;; (assert (sprop-p prop))
  (and (sprop-p prop)
       (not (eq :during (sprop-time prop))) ; don't return true for an action!
       (if (sprop-negated-p prop)
	   (false-nonnegated-sprop-p (sprop-negate prop))
	 (true-nonnegated-sprop-p prop))))

(defun false-sprop-p (prop)
  ;; (assert (sprop-p prop))
  (and (sprop-p prop)
       (not (eq :during (sprop-time prop))) ; don't return true for an action!
       (if (sprop-negated-p prop)
	   (true-nonnegated-sprop-p (sprop-negate prop))
	 (false-nonnegated-sprop-p prop))))

;; Returns true if fluent is certainly false.
;; Returns false if the fluent might or might not be false.
;; Not a great name; it makes it sound like the entry point.
(defun false-nonnegated-fluent-p (f)
  (and (consp f)
       (not (eq :during (car (last f)))) ; don't return true for an action!
       (or (false-non-unary-type-p f)
	   ;; Test whether this is a fluent known to be true
	   (let ((fluents-assoc (assoc (car f) *false-fluents* :test #'eq)))
	     (and fluents-assoc
		  (member-equal (butlast (cdr f)) (cdr fluents-assoc)))))))

(defun true-nonnegated-fluent-p (f)
  (and (consp f)
       (not (eq :during (car (last f)))) ; don't return true for an action!
       (or (true-non-unary-type-p f)
	   ;; Test whether this is a fluent known to be true
	   (let ((fluents-assoc (assoc (car f) *true-fluents* :test #'eq)))
	     (and fluents-assoc
		  (member-equal (butlast (cdr f)) (cdr fluents-assoc)))))))

;; Returns true if fluent is certainly false.
;; Returns false if the fluent might or might not be false.
(defun false-nonnegated-sprop-p (prop)
  (declare (type sprop prop))
  (assert (sprop-p prop))
  (assert (not (sprop-negated-p prop)))
  (or (false-non-unary-type-nonnegated-sprop-p prop)
      ;; Test whether this is a fluent, but can't be true.
      (let ((fluents-assoc (assoc (sprop-name prop) *false-fluents*)))
	(and fluents-assoc
	     (member-equal (sprop-args prop) (cdr fluents-assoc))))))

(defun true-nonnegated-sprop-p (prop)
  (declare (type sprop prop))
  (assert (sprop-p prop))
  (assert (not (sprop-negated-p prop)))
  (or (true-non-unary-type-nonnegated-sprop-p prop)
      ;; Test whether this is a fluent, but can't be false.
      (let ((fluents-assoc (assoc (sprop-name prop) *true-fluents*)))
	(and fluents-assoc
	     (member-equal (sprop-args prop) (cdr fluents-assoc))))))

(defun true-non-unary-type-and-args-p (name args)
  (declare (symbol name))
  (cond ((eq :eq name) (eq (second args) (third args)))
	((eq :neq name) (not (eq (second args) (third args))))
	((member name *non-unary-types* :test #'eq)
	 (member args (cdr (assoc name *non-unary-types-list* :test #'eq))
		 :test #'equal))
	(t nil)))

(defun false-non-unary-type-and-args-p (name args)
  (declare (symbol name))
  (cond ((eq :eq name) (not (eq (second args) (third args))))
	((eq :neq name) (eq (second args) (third args)))
	((member name *non-unary-types* :test #'eq)
	 (not (member args (cdr (assoc name *non-unary-types-list* :test #'eq))
		      :test #'equal)))
	(t nil)))

;; Returns true if fluent is certainly true.
;; Returns false if the fluent might or might not be true.
(defun true-non-unary-type-p (f)
  (assert (keyword-or-number-p (car (last f))))
  (and (consp f)
       (true-non-unary-type-and-args-p (car f) (butlast (cdr f)))))

;; You probably don't want this; use false-nonnegated-fluent-p (which calls it) instead.
(defun false-non-unary-type-p (f)
  (assert (keyword-or-number-p (car (last f))))
  (and (consp f)
       (false-non-unary-type-and-args-p (car f) (butlast (cdr f)))))

(defun true-non-unary-type-nonnegated-sprop-p (prop)
  (declare (type sprop prop))
  (assert (sprop-p prop))
  (assert (not (sprop-negated-p prop)))
  (true-non-unary-type-and-args-p (sprop-name prop) (sprop-args prop)))

(defun false-non-unary-type-nonnegated-sprop-p (prop)
  (declare (type sprop prop))
  (assert (sprop-p prop))
  (assert (not (sprop-negated-p prop)))
  (false-non-unary-type-and-args-p (sprop-name prop) (sprop-args prop)))

(defun true-unary-timeless-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (false-unary-nonnegated-fluent-p (append (second lit) '(:none)))
	 (true-unary-nonnegated-fluent-p (append lit '(:none))))))

(defun false-unary-timeless-fluent-or-negation-p (lit)
  (and (consp lit)
       (if (eq :not (car lit))
	   (true-unary-nonnegated-fluent-p (append (second lit) '(:none)))
	 (false-unary-nonnegated-fluent-p (append lit '(:none))))))

(defun true-unary-nonnegated-fluent-p (f)
  (and (consp f)
       (= 3 (length f))
       (not (eq :during (car (last f))))
       (let ((fluents-assoc (assoc (car f) *true-fluents*)))
	 (and fluents-assoc
	      (member-equal (butlast (cdr f)) (cdr fluents-assoc))))))

(defun false-unary-nonnegated-fluent-p (f)
  (and (consp f)
       (= 3 (length f))
       (not (eq :during (car (last f))))
       (let ((fluents-assoc (assoc (car f) *false-fluents*)))
	 (and fluents-assoc
	      (member-equal (butlast (cdr f)) (cdr fluents-assoc))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; End of fluent truth tests
;;;

;make a list of all possible instantiations of fluents
(defun all-fluents ()
  (mapcan #'(lambda (f)
	      (if (= (cadr f) 0)
		  (list (list (car f)))
		(let ((c (combinations (cadr f) *objects*)))
		  (mapcar #'(lambda (comb)
			      (cons (car f) comb))
			  c))))
	  *fluents*))

;; Completely specify the initial conditions of a problem
;; by adding negations of everything that doesn't already appear.
(defun fill-out-inits (prob)
  (declare (type problem prob))
  (assert (problem-p prob))
  (let ((negated-inits '()))
    (dolist (fluent (all-fluents))
      (if (not (or (eq (car fluent) :eq)
		   (eq (car fluent) :neq)
		   (member-if #'(lambda (i)
				  ;; equal, not sprop-equal, because lists, not sprops
				  (or (equal i fluent)
				      (equal i (list :not fluent))))
			      (problem-init prob))))
	  ;; This isn't an equality contraint, nor is it or its negation already
	  ;; in problem-init, so add its negation.
	  (push (list :not fluent) negated-inits)))
    (setf (problem-init prob) (append (problem-init prob) negated-inits))))


;; Push nots into formulas; 
;;   convert (not (neq...)) to (eq ...) and (not (eq ...)) to (neq ...)
;; Also simplify other negations:
;;   use DeMorgan's rule for (not (or ...)) and (not (and ...))
;;   convert (not (forall ...)) to (exists (not ...)); likewise for exists
;; Also pushes nots through foralls, changing ~A to E~
;; and ~E to A~
(defun push-not-into-formula (form)
  (if (symbolp form)
      form
    (let ((fn (car form)))
      (case fn
       ((:and :or :when)
	(cons fn (mapcar #'push-not-into-formula (cdr form))))
       ((:forall :exists)
	(list fn (second form) (push-not-into-formula (third form))))
       ((:not)
	(let ((arg (second form)))
	  (if (symbolp arg)
	      form
	    (let ((fn2 (first arg)))
	      (cond ((eq fn2 :eq) (cons :neq (cdr arg)))
		    ((eq fn2 :neq) (cons :eq (cdr arg)))
		    ((eq fn2 :forall)
		     (list :exists (second arg) (push-not-into-formula
						 (list :not (third arg)))))
		    ((eq fn2 :exists)
		     (list :forall (second arg) (push-not-into-formula
						 (list :not (third arg)))))
		    ;; These cases can fire due to :or, :and inside when.
		    ;; Not sure this is the right way to fix this problem.
		    ((eq fn2 :or) (cons :and (mapcar #'(lambda (x)
							 (push-not-into-formula
							  (list :not x)))
						     (cdr arg))))
		    ((eq fn2 :and) (cons :or (mapcar #'(lambda (x)
							 (push-not-into-formula
							  (list :not x)))
						     (cdr arg))))
		    ((eq fn2 :not) (push-not-into-formula (second arg)))
		    ((eq fn2 :when)
		     (error "Can't cope with :when inside :not ~s" form))
		    ((keywordp fn2)
		     (error "What formula head nested in :not? ~s" form))
		    (t (list :not (push-not-into-formula arg))))))))
       ((:eq :neq)
	form)
       (otherwise
	(if (keywordp fn)
	    (error "What formula head? ~s" form))
	form)))))
;; Tests
;; (push-not-into-formula '(:not (:not (:not (:eq a b)))))
;; (push-not-into-formula '(:not (:forall (object ?O) (at ?O home))))
;; (push-not-into-formula '(:not (:forall (person ?P) (:and ?P Q ))))
;; (PUSH-NOT-INTO-FORMULA '(:NOT (:EXISTS (SPICEGIRL ?S) (:EQ ?S LAME))))
;; (push-not-into-formula '(:not (:OR A B )))


; FOR PRECONDS--convert exists to and, put new variable in parm list
(defun exists-skolemize (form o)
  (declare (type list form))
  (declare (type operator o))
  (assert (listp form))
  (assert (operator-p o))
  (cond ((eq (car form) :exists)
	 (let* ((exists-bindings (second form))
		(parsed-bindings (parse-bindings exists-bindings))
		(varnames (mapcar #'first parsed-bindings)))
	   (if (intersection varnames (op-formals o))
	       (error "Trying to add existing formals; alpha-rename"))
	   (let ((skolem-assoc (assoc (op-name o) *skolemized-parameters*)))
	     (if skolem-assoc
		 (incf (cdr skolem-assoc))
	       (push (cons (op-name o) 1) *skolemized-parameters*)))
	   (setf (op-parameters o) (append (op-parameters o)
					   exists-bindings)
		 (op-formals o) (append (op-formals o) varnames)
		 (op-formal-types o) (append (op-formal-types o) parsed-bindings)))
	 ;; since the info is in the formal-types slot, no need to add the :and
	 ;; (list :and (second form) (exists-skolemize (third form) o))
	 (exists-skolemize (third form) o))
	((and (keywordp (car form)) (not (member (car form) '(:eq :neq))))
	 (cons (car form) (mapcar #'(lambda (arg) (exists-skolemize arg o))
				  (cdr form))))
	(t form)))

(defun unskolemize-plan (plan)
  (if (eq *exists-strategy* 'exists-skolemize)
      (mapcar #'unskolemize-sprop plan)
    plan))

(defun unskolemize-sprop (sprop)
  (let ((num-skolem-parameters
	 (cdr (assoc (sprop-name sprop) *skolemized-parameters*))))
    (if (not num-skolem-parameters)
	sprop
      (make-sprop (sprop-name sprop)
		  (butlast (sprop-args sprop) num-skolem-parameters)
		  (sprop-time sprop)
		  (sprop-negated-p sprop)))))





; turn GOAL's (EXISTS <X> <Y>) to (OR (AND <> <>) (AND <> <>)...)
;;  TYPE is a unary predicate, such as (FOO ?X).
(defun exist-aux-type (type form)
  (assert (length-two type))
  (if (member (first type) *vacuous-fluents-used-as-types*)
      (cnf-false)
    (let ((var-name (second type)))
      (cons :or (mapcar #'(lambda (o)
			    (instantiate (list (cons var-name o))
					 (list :and type form)))
			*objects*)))))

;; This version is wrong because *unary-types-list* isn't yet set when
;; exists->or (which calls this) is run.  This should be fixed...
;; ; turn GOAL's (EXISTS <X> <Y>) to (OR (AND <> <>) (AND <> <>)...)
;; ;; FORM is a unary predicate, such as (FOO ?X).
;; (defun exist-aux-type (type form)
;;   ;; Is this comment still true?
;;   ;; *unary-types-list* is null unless *types-strategy* is eliminate.
;;   (assert (length-two type))
;;   (let* ((predicate-name (first type))
;; 	 (var-name (second type))
;; 	 (types-assoc (assoc predicate-name *unary-types-list*)))
;;     ;; (format t "types-assoc: ~s~%" types-assoc)
;;     (if types-assoc
;; 	(cons :or (mapcar #'(lambda (o)
;; 			      (instantiate (list (cons var-name o))
;; 					   form))
;; 			  (cdr types-assoc)))
;;       (cons :or (mapcar #'(lambda (o)
;; 			    (instantiate (list (cons var-name o))
;; 					 (list :and type form)))
;; 			*objects*)))))

(defun exist-aux-var (var form)
  (cons :or (mapcar #'(lambda (o)
			(instantiate (list (cons var o))
				     form))
		    *objects*)))

; rid exists from GOAL using exist-aux
(defun exists->or (form)
  (cond
   ((or (eq (car form) :and) (eq (car form) :or))
    (cons (car form) (mapcar #'exists->or (cdr form))))
   ((eq (car form) :not)
    (list :not (exists->or (cadr form))))
   ((eq (car form) :exists)
    (let ((quant (second form))
	  (body (third form)))
      (if (length-one quant)		; was (cdr quant); why???
	  (exist-aux-var (car quant) (exists->or body))
	(exist-aux-type quant (exists->or body)))))
   (t
    form)))



(defun unclone-sprop (act)
  "Return a new sprop with with orginal name if this one was cloned.
Otherwise, return argument unchanged."
  (let* ((op-as-string (string (sprop-name act)))
	 (sep-pos (search *clone-op-suffix* op-as-string)))
    (if sep-pos
	(make-sprop (intern (subseq op-as-string 0 sep-pos)
			    (symbol-package (sprop-name act)))
		    (sprop-args act) (sprop-time act)
		    (sprop-negated-p act))
      act)))

; given an operator and a WHEN expression, recursively create all the
; various possible operators by removing WHENs.
(defun clone-op-removing-when (oper when)
  (let* ((when-condition (second when))
	 (when-body (third when))
	 (condition-true (true-timeless-formula-p when-condition))
	 (condition-false (false-timeless-formula-p when-condition))
	 (new-op (and (not condition-false)
		      (let ((new-name (gentemp (format nil "~a~a"
						       (op-name oper) *clone-op-suffix*)
					       (symbol-package (op-name oper)))))
			(make-operator
			 :name new-name
			 :parameters (op-parameters oper)
			 :formals (op-formals oper)
			 :formal-types (op-formal-types oper)
			 :precondition (if condition-true
					   (op-precondition oper)
					 (list :and (op-precondition oper)
					       when-condition))
			 :effect (list :and (op-effect oper) when-body))))))
    (if (or condition-true (cnf-true-p (op-effect oper)))
	(setq oper nil)
      (setf (op-precondition oper)
	    (if condition-false
		(op-precondition oper)
	      (list :and (op-precondition oper)
		    (list :not when-condition)))))
    ;; (push (op-name oper) *cloned-ops*)
    (mapcan #'spawn-op-for-each-when (delete nil (list new-op oper)))))


; find a when in the operator's effect, and use it to clone the operator
(defun spawn-op-for-each-when (oper)
  (let ((when (rec-find-if #'(lambda (x) (and (listp x) (eq (car x) :when)))
			   (op-effect oper))))
    (if (null when)
	(list oper)
      (progn (setf (op-effect oper)
		   (if (eq when (op-effect oper))
		       '(:and)
		     (rec-remove-if #'(lambda (x)
					;; Don't use EQ in rec-remove-if
					(equal x when))
				    (op-effect oper))))
	     (clone-op-removing-when oper when)))))


; build the types list, using the initial conditions to find the objects of
; each type.

(defun build-types-list (prob-init)
  (assert (and (listp prob-init) (eq :and (first prob-init))))
  (setq *unary-types-list* (mapcar #'list *unary-types*))
  (setq *non-unary-types-list* (mapcar #'list *non-unary-types*))
  (dolist (this-init (cdr prob-init))
    (let ((fluent-name (first this-init)))
      (if (not (eq fluent-name :not))
	  (let ((unary-assoc (assoc fluent-name *unary-types-list*)))
	    (if unary-assoc
		(push (second this-init) (cdr unary-assoc))
	      (let ((non-unary-assoc (assoc fluent-name *non-unary-types-list*)))
		(if non-unary-assoc
		    (push (cdr this-init) (cdr non-unary-assoc)))))))))
  (let ((non-unary-types-list-by-position
	 ;; list of lists of the form (<type-name> (<obj1> <obj2>) ...)
	 ;; where obj1 and obj2 can appear as the first argument to TYPE.
	 (mapcar #'(lambda (one-non-unary-types-list)
		     ;; (format t "one-non-unary-types-list: ~s~%" one-non-unary-types-list)
		     (cons (car one-non-unary-types-list)
			   (if (length-one one-non-unary-types-list)
			       ;; zero-arity type
			       '()
			     (mapcar #'delete-duplicates
				     ;; shares list structure with *non-unary-types-list*
				     (apply #'mapcar #'list (cdr one-non-unary-types-list))))))
		 *non-unary-types-list*)))
    ;; (format t "non-unary-types-list-by-position: ~s~%" non-unary-types-list-by-position)
    (setq *removable-types* *unary-types*)

    ;; Add some new unary types, one for each position in a non-unary type.
    ;; This lets the simpler unary type code do its work before we bring in the
    ;; big guns to process non-unary types.
    (dolist (name+obj-lists non-unary-types-list-by-position)
      (let ((name (car name+obj-lists)))
	(declare (symbol name))
	(assert (symbolp name))
	;; eq and neq aren't in the initial conditions, so we would mistakenly
	;; deduce they have no successful bindings.
	(if (not (member name '(:eq :neq)))
	    (progn
	      (if (>= (count-if #'length-one (cdr name+obj-lists))
		      (1- (length (cdr name+obj-lists))))
		  (push name *removable-types*))
	      (loop for obj-list in (cdr name+obj-lists)
		    as i = 1 then (+ i 1)
		    do (let ((single-arg-name (intern (format nil "~s###~d" name i))))
			 (push single-arg-name *unary-types*)
			 (push (cons single-arg-name obj-list) *unary-types-list*)))))))
    ;; (format t "removable-types: ~s~%" *removable-types*)
    ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Type inference
;;;

;; EFFECT is a lists of (fluent (action argname) ...)
(defun inferred-types (fluent-effect)
  (cons (car fluent-effect)
	(reduce #'intersection (mapcar #'inferred-type-for-action-argname
				       (cdr fluent-effect)))))
;; To test:  run (fridge-fixa), C-c to break, then
;; (inferred-types (car *fluent-effect-lst*))
;; (mapcar #'inferred-types *fluent-effect-lst*)

;; (setq *fluent-effect-lst*
;;       '((SCREWED (SCREW ?X) (foo ?y)) (IN-PLACE (ATTACH-BACKPLANE ?X))
;; 	(FRIDGE-ON (START-FRIDGE ?F)) (ATTACHED (CHANGE-COMPRESSOR ?Y))))
;; (setq *fluent-precond-lst*
;;       '((SCREW (UNSCREW ?X) (SCREW ?X) (foo ?y))
;; 	(COMPRESSOR (screw ?x) (CHANGE-COMPRESSOR ?Y) (CHANGE-COMPRESSOR ?X))
;; 	(BACKPLANE (UNSCREW ?Y) (SCREW ?Y) (REMOVE-BACKPLANE ?X)
;; 		   (ATTACH-BACKPLANE ?X) (START-FRIDGE ?X) (CHANGE-COMPRESSOR ?A))))


;; Returns a list of types mentioning the arg in the action preconditions
;; Argument is a two-element list like (pick-up ?object).  We want to know
;; the type of ?object in the pick-up action.
(defun inferred-type-for-action-argname (action-argname)
  (mapcan #'(lambda (fluent-precond)
	      ;; EQUAL: Not sure why here.
	      (if (member action-argname (cdr fluent-precond) :test #'equal)
		  (list (car fluent-precond))
		'()))
	  *fluent-precond-lst*))


;; Maps inferred-types over *fluent-effect-lst* and check results with inits
(defun all-inferred-types (problem-init)
  (assert (eq :and (car problem-init)))
  (let* ((init-list (cdr problem-init))
	 (inferred-types (mapcar #'inferred-types *fluent-effect-lst*))
	 (real-inferred-types
	  (mapcar #'(lambda (infer)
		      (let* ((fluent (car infer))
			     (types (cdr infer))
			     (matches
			      (remove-if-not #'(lambda (i)
						 (eq (car i) fluent))
					     init-list))
;; 			     (winners (mapcan
;; 				       #'(lambda (typ)
;; 					   (if (every #'(lambda (m)
;; 							  (member (cons typ (cdr m))
;; 								  init-list :test #'equal))
;; 						      matches)
;; 					       (list typ)))
;; 				       types))
			     (winners (remove-if-not
				       #'(lambda (typ)
					   (every #'(lambda (m)
						      (member (cons typ (cdr m))
							      init-list :test #'equal))
						  matches))
				       types)))

			    (cons fluent winners)))
		  inferred-types)))
    real-inferred-types))


;; take a operator and a list (fluent <type> <type>...) and adds types to the
;; preconditions as appropriate.

(defun update-precond-with-types (op fluent-types-lst)
  (let ((precond (op-precondition op))
	(fluent (car fluent-types-lst))
	(types (cdr fluent-types-lst)))
    (assert (and (listp precond) (eq :and (car precond))))
    (flet ((instantiated-types (arg)
	    (mapcar #'(lambda (type)
			(list type arg))
		    types)))
      (let* ((conjuncts (cdr precond))
	     (fluent-instatiations (remove-if-not #'(lambda (conjunct)
						      (eq (car conjunct) fluent))
						  conjuncts))
	     (inst-types (let ((unreduced
				(mapcar #'(lambda (inst-fluent)
					    (instantiated-types (second inst-fluent)))
					fluent-instatiations)))
			   (if (not (null unreduced))
			       (reduce #'union-equal unreduced)
			     '()))))
	(if inst-types
	    (progn
	      ;; (print "unioning")
	      ;; (print (op-name op))
	      ;; (print inst-types)
	      ;; (print conjuncts)
	      ;; (print (cons :and (union-equal conjuncts inst-types)))
	      (setf (op-precondition op)
		(cons :and (union-equal conjuncts inst-types)))))))))

;; This does not appear to be called any longer
;; It was called only for side effect
;; (defun create-infer-lst-first-pass (effects)
;;   (if (not (null effects))
;;       (let* ((cadar-effects (cadar effects))
;; 	     (types (remove-if-not
;; 		     #'(lambda (p)
;; 			 ;; P is a list of operator names, so o is a symbol
;; 			 (member-if #'(lambda (o)
;; 					(eq o cadar-effects))
;; 				    (cdr p)))
;; 		     *fluent-precond-lst*)))
;; 	(if (null types)
;; 	    (create-infer-lst-first-pass (cdr effects))
;; 	  (progn
;; 	    (setq *infer-lst* (cons (cons (caar effects)
;; 					  (mapcar #'car types))
;; 				    *infer-lst*))
;; 	    (create-infer-lst-first-pass (cdr effects)))))))

;; (defun prune-infer-lst (fluent effects)
;;   (cond ((null effects) '())
;; 	(t (let ((types (remove-if-not
;; 			 #'(lambda (p)
;; 			     (member-if #'(lambda (o)
;; 					    (equal (cadar effects)
;; 						   o))
;; 					(cdr p)))
;; 			 *fluent-precond-lst*)))))))


;; If the preconditions or goals demand conflicting properties of two
;; arguments -- e.g., (on ?x ?y) and (not (on ?x ?z)) -- then add (neq ?y ?z)
;; to the preconditions.
(defun op-add-neq-constraints (op)
  ;; The call to remove-duplicates won't catch reversed order of arguments.
  (let ((conflicts (remove-duplicates
		    (append (conflicting-variables
			     (conjunctive-clauses (op-precondition op)))
			    (conflicting-variables
			     (conjunctive-clauses (op-effect op))))
		    :test #'equal)))
    ;; (format t "Conflicts ~s for ~s~%" conflicts op)
    (setf (op-precondition op)
	  (if (some #'stringp conflicts)
	      ;; a conflict is a string if it would be (:neq x x)
	      (cnf-false)
	    (cons :and
		  (remove-duplicates
		   (append (mapcar #'(lambda (vars) (cons :neq vars)) conflicts)
			   (conjunctive-clauses (op-precondition op)))
		   :test #'equal))))))

;; Takes a list of conjuncts, returns a list of two-element lists of (var1
;; var2) such that the two variables must not have the same value.  One of
;; the variables may be a constant.  The conjunction has elements in list,
;; not sprop, form.
(defun conflicting-variables (conjuncts)
  ;; Not the most efficient possible implementation, but arglist should be
  ;; short.
  (if (null conjuncts)
      '()
    (nconc (delete nil (mapcan #'(lambda (conj)
				   (conflicting-variables-from-pair
				    (car conjuncts) conj))
			       (cdr conjuncts)))
	   (conflicting-variables (cdr conjuncts)))))
;; (conflicting-variables '((foo a b) (foo a c) (:not (foo a ?d)) (bar a) (:not (bar ?q))))


;; Returns a list of two-element lists of the form (var1 var2).
;; (At present the result contains either zero or one such lists.)
(defun conflicting-variables-from-pair (prop1 prop2)
  (let ((head1 (car prop1))
	(head2 (car prop2)))
    (if (or (and (eq :not head1) (eq :not head2))
	    (and (not (eq :not head1)) (not (eq :not head2))))
	'()
      (progn
	(if (eq :not head1)
	    (setq prop1 (second prop1)
		  head1 (first prop1)))
	(if (eq :not head2)
	    (setq prop2 (second prop2)
		  head2 (first prop2)))
	(cond ((or (not (eq head1 head2))
		   (eq head1 :or)
		   (not (= (length prop1) (length prop2))))
	       '())
	      ((equal prop1 prop2)
	       (list (format nil "Can't satisfy both ~s and its inverse" prop1)))
	      (t
	       (let ((pairs (delete nil (mapcar #'(lambda (var1 var2)
						    (if (eq var1 var2)
							nil
						      (list var1 var2)))
						(cdr prop1) (cdr prop2)))))
		 (declare (list pairs))
		 ;; A conflict between two (different) constants can't be
		 ;; satisfied, so add no constraints.
		 (if (member-if #'(lambda (x)
				    (and (not (varp (first x)))
					 (not (varp (second x)))))
				pairs)
		     (setq pairs '()))
		 (if (> (length pairs) 1)
		     ()
		     ;;(error "Teach me how to cope with arities greater than 1: ~s ~s" prop1 prop2)
		   pairs))))))))
;; Tests -- need some tests with constants, too!
;; (conflicting-variables-from-pair '(:not (foo ?a ?b)) '(foo ?a ?c))
;; (conflicting-variables-from-pair '(foo ?a ?c) '(:not (foo ?a ?b)))
;; (conflicting-variables-from-pair '(bar ?a) '(:not (bar ?b)))
;; (conflicting-variables-from-pair '(bar ?a) '(:not (bar ?a))) ; error
;; (conflicting-variables-from-pair '(:not (foo ?a ?b)) '(foo ?c ?d)) ; error
;; (conflicting-variables-from-pair '(foo ?a ?b) '(foo ?a ?c))
;; (conflicting-variables-from-pair '(:not (foo ?a ?b)) '(bar ?a ?c))
;; (conflicting-variables-from-pair '(:not (foo ?a ?b)) '(foo ?a))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Eliminate some ground instantiations
;;;

;; NOTE:  NOT USING SPROPS
;; A better name would be "eliminate-types".
(defun prune-obj-lst (op prob-init)
  (declare (type operator op))
  (assert (operator-p op))
  (assert (and (listp (op-precondition op)) (eq :and (car (op-precondition op)))))
  (assert (and (listp prob-init) (eq :and (first prob-init))))

  (let ((init-list (cdr prob-init)))
    ;; (print-operator-long op)
    (prune-obj-lst-aux op (op-precondition op) init-list)
    ;; (print-operator-long op)
    (prune-obj-lst-aux op (op-effect op) init-list)
    ;; (print-operator-long op)
    (prune-obj-lst-aux op (copy-tree (op-formal-types op)) init-list)
    ;; (print-operator-long op)

    ;; Don't do this:
    ;;  1. don't want to remove all unary types
    ;;  2. prune-obj-lst-aux does the right thing by side effect
    ;; ;; Since we've already reflected their effects, remove unary types.
    ;; ;; Also remove non-unary types if all but one of their argument lists
    ;; ;; have length one.
    ;; (flet ((removable-type-or-negation-p (p)
    ;;	   ;; Equivalent, but not using sprops at this point.
    ;;	   ;; (and (sprop-p p)
    ;;	   ;;    (member (sprop-name *removable-types*)))
    ;;	   (and (listp p)
    ;;		(member (car (if (eq :not (car p)) (second p) p))
    ;;			*removable-types* :test #'eq))))
    ;;   (if (null (op-formals op))
    ;;	(progn
    ;;	  ;; Bad idea, because *false-fluents* and friends aren't yet set.
    ;;	  ;; I'll hope that the true and false fluents are removed from zero-ary
    ;;	  ;; operators later.
    ;;	  ;; 	  ;; Could also check whether car is in *removable-types*, but what's the point?
    ;;	  ;; 	  (if (some #'(lambda (x) (and (not (formula-combiner-p (car x)))
    ;;	  ;; 				       (false-fluent-or-negation-p
    ;;	  ;; 					(add-time-to-formula :foo x))))
    ;;	  ;; 		    (conjunctive-clauses (op-precondition op)))
    ;;	  ;; 	      (progn
    ;;	  ;; 		(format t "Preconds of ~s contain false fluent~%" (op-name op))
    ;;	  ;; 		(break)
    ;;	  ;; 		(setf (op-precondition op) (list :and (cnf-false)))))
    ;;	  ;; 	  (setf (op-precondition op)
    ;;	  ;; 	    (cons :and (remove-if #'(lambda (x) (and (not (formula-combiner-p (car x)))
    ;;	  ;; 						     (true-fluent-or-negation-p
    ;;	  ;; 						      (add-time-to-formula :foo x))))
    ;;	  ;; 				  (conjunctive-clauses (op-precondition op)))))
    ;;	  ;; 	  (if (some #'(lambda (x) (and (not (formula-combiner-p (car x)))
    ;;	  ;; 				       (false-fluent-or-negation-p
    ;;	  ;; 					(add-time-to-formula :foo x))))
    ;;	  ;; 		    (conjunctive-clauses (op-effect op)))
    ;;	  ;; 	      (progn
    ;;	  ;; 		(format t "Effects of ~s contain false fluent~%" (op-name op))
    ;;	  ;; 		(break)
    ;;	  ;; 		(setf (op-effect op) (list :and (cnf-false)))))
    ;;	  ;; 	  (setf (op-effect op)
    ;;	  ;; 	    (cons :and (remove-if #'(lambda (x)
    ;;	  ;; 				      (and (not (formula-combiner-p (car x)))
    ;;	  ;; 					   (true-fluent-or-negation-p
    ;;	  ;; 					    (add-time-to-formula :foo x))))
    ;;	  ;; 				  (conjunctive-clauses (op-effect op)))))
    ;;	  )
    ;;     ;; This works because we know this is a conjunction, there are no nested ands,
    ;;     ;; and we only need to substitute for ands not in :or or :not.
    ;;     (progn
    ;;	(setf (op-precondition op)
    ;;	  (cons :and (remove-if #'removable-type-or-negation-p
    ;;				(conjunctive-clauses (op-precondition op)))))
    ;;	(setf (op-effect op)
    ;;	  (cons :and (remove-if #'removable-type-or-negation-p
    ;;				(conjunctive-clauses (op-effect op))))))))
    ))


;; Prune the obj lsts of actions that have types as preconditions.
;; Side-effects the formula.
;; Checks in ANDs and NOTs (it would be wrong to do so in ORs).
(defun prune-obj-lst-aux (op formula init-list)
  (declare (type operator op))
  (assert (operator-p op))
  (cond ((cnf-true-p formula)
	 ;; nothing to do
	 )
	((eq :and (car formula))
	 (dolist (p (cdr formula))
	   (prune-obj-lst-aux op p init-list)))
	(t
	 (let* ((complemented (eq :not (car formula)))
		(subformula (if complemented (second formula) formula))
		(fluent-name (car subformula)))
	     (cond ((member fluent-name *unary-types* :test #'eq)
		    (setf (op-obj-lst op)
		      (type-filter subformula op complemented init-list))
		    (make-formula-true formula))
		   ((and (member fluent-name *non-unary-types* :test #'eq)
			 (not (member fluent-name '(:eq :neq) :test #'eq))
			 (not (null (cdr subformula))))
		    (loop for arg in (cdr subformula)
			as i = 1 then (+ i 1)
			do (let ((single-arg-name
				  (intern (format nil "~s###~d" fluent-name i))))
			     (setf (op-obj-lst op)
			       (type-filter (list single-arg-name arg)
					    op complemented init-list))))
		    ;; do not set the formula to be true (yet); will happen
		    ;; in prune-aux
		    ))))))


;; Partition the arguments of op into groups such that all those in each
;; group are indistinguishable but must be different.
;; The operator has no time yet.
(defun equivalent-args (op)
  (let ((precond-var-alist (mapcar #'list (op-formals op)))
	(effect-var-alist (mapcar #'list (op-formals op)))
	(result '()))
    (dolist (pre (conjunctive-clauses (op-precondition op)))
      (dolist (arg (cdr (if (eq :not (car pre)) (second pre) pre)))
	(let ((arg-assoc (assoc arg precond-var-alist)))
	  (if arg-assoc
	      (push pre (cdr arg-assoc))
	    (if (varp arg)
		(error "~s should have an entry" arg))))))
    (dolist (eff (conjunctive-clauses (op-effect op)))
      (dolist (arg (cdr (if (eq :not (car eff)) (second eff) eff)))
	(let ((arg-assoc (assoc arg effect-var-alist)))
	  (if arg-assoc
	      (push eff (cdr arg-assoc))
	    (if (varp arg)
		(error "~s should have an entry" arg))))))
    ;; (format t "pre: ~s~%eff: ~s~%" precond-var-alist effect-var-alist)
    (flet ((equiv-args (a1 a2)
	     (declare (symbol a1) (symbol a2))
	     (and
	      (eq (cdr (assoc a1 (op-formal-types op)))
		  (cdr (assoc a2 (op-formal-types op))))
	      (flet ((canonicalize-eq-neq (f)
		      ;; only one of these substitutions can fire, but
		      ;; we don't bother to keep track of which one.
		      (sublis (list (cons a1 "other") (cons a2 "other"))
			      (if (and (or (eq :eq (car f)) (eq :neq (car f)))
				       (equal (third f) "arg"))
				  (list (first f) (third f) (second f))
				f))))
		(let ((pre1 (mapcar #'canonicalize-eq-neq
				    (subst "arg" a1
					   (cdr (assoc a1 precond-var-alist)))))
		      (pre2 (mapcar #'canonicalize-eq-neq
				    (subst "arg" a2
					   (cdr (assoc a2 precond-var-alist)))))
		      (eff1 (subst "arg" a1 (cdr (assoc a1 effect-var-alist))))
		      (eff2 (subst "arg" a2 (cdr (assoc a2 effect-var-alist)))))
		  ;; (format t "comparing args ~s ~s~%" a1 a2)
		  ;; (format t "   ~s~%   ~s~%   ~s~%   ~s~%" pre1 pre2 eff1 eff2)
		  (and (member '(:NEQ "arg" "other") pre1 :test #'equal)
		       (null (set-exclusive-or pre1 pre2 :test #'equal))
		       (null (set-exclusive-or eff1 eff2 :test #'equal))))))))
      (dolist (arg (op-formals op))
	(declare (symbol arg))
	(let ((done nil))
	  (dolist (group result)
	    (if (and (not done) (equiv-args arg (car group)))
		(progn
		  ;; (format t "equivalent: ~s ~s~%" arg (car group))
		  ;; (push arg group) doesn't work here
		  (rplacd group (cons arg (cdr group)))
		  (setq done t))))
	  (if (not done)
	      (push (list arg) result))))
      result)))

;; Since we have done no pruning yet, it should be the case that the bindings
;; are symmetric with respect to the equivalent args.
;; It might be better to do this when creating bindings.
(defun binding-prune-equivalent-args (op)
  (let ((formals (op-formals op)))
    (dolist (equiv (equivalent-args op))
      (declare (cons equiv))
      (if (not (length-one equiv))
	  ;; The list of objects must be the same for all the variables
	  (let ((pos (position (car equiv) formals)))
	    (if (not pos) (error "didn't find ~s in ~s" (car equiv) formals))
	    (let ((objects (nth pos (op-obj-lst op))))
	      (declare (list objects))
	      (if (< (length objects) (length equiv))
		  (error "More objects than variables: ~s ~s~%" objects equiv))
	      (let ((choice-bindings (mapcar #'(lambda (choice)
						 (mapcar #'cons equiv choice))
					     (choose (length equiv) objects))))
		(setf (op-bindings op)
		      (remove-if
		       ;; returns non-nil if some chosen bindings isn't in op-bindings
		       #'(lambda (op-binding)
			   (not (some #'(lambda (choice-binding)
					  ;; (format t "differencing ~s ~s~%" choice-binding op-binding)
					  (null (set-difference
						 choice-binding op-binding
						 :test #'equal)))
				      choice-bindings)))
		       (op-bindings op))))))))))


;; Not using sprops yet.
(defun binding-prune (op)
  (declare (type operator op))
  (assert (operator-p op))
  ;; (print-operator-long op)
  ;; Add neq for conflicting arguments (perhaps should be considered
  ;;   a preconditioning step).
  (op-add-neq-constraints op)
  ;; (format t "op 1: ~d, ~s~%" (length (op-bindings op)) (mapcar #'length (op-obj-lst op))) (print-operator-long op)
  (binding-prune-equivalent-args op)
  ;; (format t "op 2: ~d, ~s~%" (length (op-bindings op)) (mapcar #'length (op-obj-lst op))) (print-operator-long op)
  (prune-aux op (op-precondition op))
  ;; (format t "op 3: ~d, ~s~%" (length (op-bindings op)) (mapcar #'length (op-obj-lst op))) (print-operator-long op)
  (prune-aux op (op-effect op))
  ;; (format t "op 4: ~d, ~s~%" (length (op-bindings op)) (mapcar #'length (op-obj-lst op))) (print-operator-long op)

  ;; Squeeze out what prune-aux added/removed.
  (setf (op-precondition op) (flatten-formula (op-precondition op)))

  ;; Don't do this any longer; now, prune-aux side-effects its formula argument
  ;; by inserting true or false fluents.
  ;;   ;; Only remove the NEQs/EQs, if we're not doing Factoring/Explanatory frames
  ;;   (if (not (and (factored-splitting-p)
  ;; 		(eq *frame-strategy* 'explanatory)))
  ;;       (setf (op-precondition op)
  ;; 	(rec-remove-if #'(lambda (p)
  ;; 			   (and (listp p)
  ;; 				(or (eq (car p) :eq)
  ;; 				    (eq (car p) :neq))))
  ;; 		       (op-precondition op))))
  )

;; Side-effects the bindings and the formula (a precondition or effect).
; Remove neq/eq constraints, and other non-unary types, from op preconds,
; and prune the bindings appropriately.  Looks only within AND (or for type
; immediately within NOT), because it's not right to do this optimization
; inside an OR (or quantifier, though those shouldn't exist at this point).
(defun prune-aux (op formula)
  (declare (type operator op))
  (declare (type (not sprop) formula))
  (assert (operator-p op))
  (assert (not (sprop-p formula)))
  ;; (format t "prune-aux (~d) <= ~s~%" (length (op-bindings op)) formula)
  (cond ((cnf-true-p formula)
	 ;; nothing to do if trivially true
	 )
	((eq :and (car formula))
	 ;; (format t "  combiner AND~%")
	 (dolist (p (cdr formula))
	   (prune-aux op p)))
	((eq :eq (car formula))
	 (setf (op-bindings op) (eq-filter (second formula) (third formula)
					   (op-bindings op)))
	 (make-formula-true formula))
	((eq :neq (car formula))
	 (setf (op-bindings op) (neq-filter (second formula) (third formula)
					    (op-bindings op)))
	 (make-formula-true formula))
	(t
	 (let* ((negated (eq :not (car formula)))
		(subformula (if negated (second formula) formula)))
	   (if (member (car subformula) *non-unary-types* :test #'eq)
	       (progn
		 (setf (op-bindings op)
		       (non-unary-type-filter subformula (op-bindings op) negated))
		 (make-formula-true formula))))))
  ;; (format t "prune-aux => ~d~%" (length (op-bindings op)))
  )



;; I think this means "retain bindings which satisfy the neq constraint".
;; Never use "filter": it's ambiguous in English!
; filter out bindings which satisfy the neq constraint
(defun neq-filter (arg1 arg2 bindings)
  (cond ((and (varp arg1) (varp arg2))
	 (delete-if-not #'(lambda (b)
			    (not (eq (cdr (assoc arg1 b :test #'eq))
				     (cdr (assoc arg2 b :test #'eq)))))
			bindings))
	((varp arg1)
	 (delete-if-not #'(lambda (b)
			    (not (eq (cdr (assoc arg1 b :test #'eq)) arg2)))
			bindings))
	((varp arg2)
	 (delete-if-not #'(lambda (b)
			    (not (eq (cdr (assoc arg2 b :test #'eq)) arg1)))
			bindings))
	;; neither argument is a var
	((eq arg1 arg2)
	 ;; args to neq are the same, so delete all bindings
	 nil)
	(t
	 ;; args to neq are different, so retain all bindings
	 bindings)))

;; I think this means "retain bindings which satisfy the eq constraint".
;; Never use "filter": it's ambiguous in English!
; filter out bindings which satisfy the eq constraint
(defun eq-filter (arg1 arg2 bindings)
  (cond ((and (varp arg1) (varp arg2))
	 (delete-if-not #'(lambda (b)
			    (eq (cdr (assoc arg1 b))
				(cdr (assoc arg2 b))))
			bindings))
	((varp arg1)
	 (delete-if-not #'(lambda (b)
			    (eq (cdr (assoc arg1 b)) arg2))
			bindings))
	((varp arg2)
	 (delete-if-not #'(lambda (b)
			    (eq (cdr (assoc arg2 b)) arg1))
			bindings))
	((eq arg1 arg2)
	 ;; args to eq are the same, so retain all bindings
	 bindings)
	(t
	 ;; args to eq are different, so delete all bindings
	 nil)))


;; creates a list of neq/eq constraints to add to the initial state (in order
;; that we can treat neq/eq as ordinary fluents)
(defun generate-neq-eq-constraints ()
  (append (generate-neq-constraints) (generate-eq-constraints)))

;; create (NEQ X Y) for all distinct X, Y;  (NOT (NEQ X X)) for all X
(defun generate-neq-constraints ()
  (let ((args1 *objects*)
	(args2 *objects*))
    (mapcan #'(lambda (a1)
		(mapcar #'(lambda (a2)
			    (if (eq a1 a2)
				(list :not (list :neq a1 a2))
			      (list :neq a1 a2)))
			args2))
	    args1)))


;; create (NOT (EQ X Y)) for all distinct X, Y;  (EQ X X) for all X
(defun generate-eq-constraints ()
  (let ((args1 *objects*)
	(args2 *objects*))
    (mapcan #'(lambda (a1)
		(mapcar #'(lambda (a2)
			    (if (eq a1 a2)
				(list :eq a1 a2)
			      (list :not (list :eq a1 a2))))
			args2))
	    args1)))

;; I don't know that this works; it hasn't been tested yet.
;; Called for side effect.
(defun non-unary-type-filter (prop bindings negated)
  (declare (list prop))
  (assert (not (sprop-p prop)))
  ;; ARGLISTS is a list of proper-arity lists of constants satisfying the type.
  (assert (or (not (member (car prop) *non-unary-types* :test #'eq)) (assoc (car prop) *non-unary-types-list*)))
  (let ((arglists (cdr (assoc (car prop) *non-unary-types-list*))))
    ;; (format t "arglists: ~s~%" arglists)
    (funcall
     (if negated #'delete-if #'delete-if-not)
     ;; Return true if this sprop matches one of the arglists.
     #'(lambda (b)
	 ;; B is a binding -- a proper-arity list of (varname . value) pairs.
	 (let ((arg-values (instantiate b (cdr prop))))
	   ;; (format t "  testing ~s in ~s: ~s~%" arg-values arglists (member arg-values arglists :test #'equal))
	   (member arg-values arglists :test #'equal)))
     bindings)))

;; Destructive version side-effected *objects*, before we used copy-list in
;; create-obj-lst.
;; Returns the new object list.
(defun type-filter (constraint op complement init-list)
  (declare (list constraint))
  (assert (not (sprop-p constraint)))
  (assert (not (eq :not (car constraint))))
  (cond ((varp (second constraint))
	 (type-filter-destructive constraint op complement))
	(t
	 ;; The argument is not a variable, but a constant.  If it's true, just
	 ;; prune.  If it's false, then the obj-lst should be set empty and
	 ;; the pre/post conditions set to false; the action can't occur.

	 ;; This code doesn't assume the closed-world assumption has been
	 ;; implemented yet.
	 (let ((constraint-true (member constraint init-list :test #'equal)))
	   (if (xor constraint-true complement)
	       ;; This (ground) constraint is always satisfied.
	       ;; That is, the constraint is in the init-list and this
	       ;; appearance is not complemented, or the constraint isn't in
	       ;; the init-list and this appearance is complemented.
	       (let ((subs (list (cons (list :not constraint)
				       (if constraint-true (cnf-false) (cnf-true)))
				 (cons constraint
				       (if constraint-true (cnf-true) (cnf-false))))))
		 (if *debug-consistency*
		     (format t "subs: ~s~%" subs))
		 (setf (op-precondition op)
		   (flatten-ands (nsublis subs (op-precondition op))))
		 (setf (op-effect op)
		   (flatten-ands (nsublis subs (op-effect op)))))
	     ;; This (ground) constraint is never satisified
	     (progn
	       (setf (op-precondition op) (list :and (cnf-false)))
	       (setf (op-effect op) (list :and (cnf-false)))
	       (setf (op-obj-lst op)
		 (mapcar #'(lambda (x)
			     (declare (ignore x))
			     '())
			 (op-obj-lst op)))))
	   (if *debug-consistency*
	       (format t "new precond: ~s~%new effects: ~s~%"
		       (op-precondition op) (op-effect op)))
	   (op-obj-lst op)))))

;; Filter out objs in obj-lst which don't satisfy the type constraint.
;; CONSTRAINT is an sprop whose name appears in *unary-types*
(defun type-filter-nondestructive (constraint op complemented)
  (let ((index (position (second constraint) (op-formals op)))
	(obj-lsts (op-obj-lst op)))
    (assert index)
    ;; Perform type-filter-one on just the indexth (zero-based) element
    (splice-at index
	       (type-filter-one
		(car constraint) (nth index obj-lsts) complemented)
	       obj-lsts)))

;; Filter out objs in obj-lst which don't satisfy the type constraint.
;; CONSTRAINT is an action whose name appears in *unary-types*
(defun type-filter-destructive (constraint op complemented)
  (let ((index (position (second constraint) (op-formals op)))
	(obj-lsts (op-obj-lst op)))
    (declare (type (or fixnum (member nil)) index))
    (assert index)
    ;; Perform type-filter-one on just the indexth (zero-based) element
    (setf (nth index obj-lsts)
	  (type-filter-one
	   (car constraint) (nth index obj-lsts) complemented))
    obj-lsts))

;; Destructive implementation (but use the returned value).
;; Filter out objs in obj-lst which don't satisfy the type constraint.
;; This is basically doing either a set intersection or set difference.
(defun type-filter-one (type-name obj-lst complemented)
  (let ((lst (cdr (assoc type-name *unary-types-list*))))
    (funcall (if complemented #'delete-if #'delete-if-not)
	     #'(lambda (o) (member o lst))
	     obj-lst)))

;;; Non-destructive implementation.
;; ;; filter out objs in obj-lst which don't satisfy the type constraint
;; (defun type-filter-one (constraint obj-lst complemented)
;;   (funcall (if complemented #'remove-if #'remove-if-not)
;; 	   #'(lambda (o)
;; 	       (let ((lst (assoc type-name *unary-types-list*)))
;; 		 (member o lst)))
;; 	   obj-lst))



;;; Old implementation
;; ;; filter out objs in obj-lst which don't satisfy the type constraint
;; (defun type-filter-aux (constraint obj-lst complemented index counter)
;;   (cond ((= index counter)
;; 	 (let ((new (type-filter-one constraint (car obj-lst) complemented)))
;; 	   (cons new (cdr obj-lst))))
;; 	(t (cons (car obj-lst) (type-filter-aux constraint (cdr obj-lst)
;; 						complemented index
;; 						(+ counter 1))))))
;; 
;; (defun type-filter (constraint op complement)
;;   (let ((index (position (cadr constraint) (op-formals op))))
;;     (type-filter-aux (car constraint) (op-obj-lst op) complement index 0)))



;; substitute true or false for types and eq/neq in the goal
(defun prune-goal (form)
  (let ((fn (car form)))
    (case fn
      ((:eq)  (cnf-true-or-false (eq (second form) (third form))))
      ((:neq) (cnf-true-or-false (not (eq (second form) (third form)))))
      ((:and :or :not)  (cons (car form) (mapcar #'prune-goal (cdr form))))
      (otherwise
       (if (and (eq *types-strategy* 'eliminate)
		(member fn *unary-types* :test #'eq))
	   (let ((lst (assoc fn *unary-types-list*)))
	     (cnf-true-or-false (member (second form) (cdr lst))))
	 form)))))


(defun eliminated-types (prob dom)
  (let* ((all-preconds (cons (problem-goal prob)
			     (mapcar #'op-precondition (domain-operators dom))))
	 (non-appearing-types
	  (remove-if #'(lambda (type)
			 (rec-find-if #'(lambda (term)
					  (and (listp term)
					       (eq type (first term))))
				      all-preconds))
		     *unary-types*)))
    non-appearing-types))


(defun prune-ground-unary-types (formula)
  (let ((fn (car formula)))
    (cond ((member fn *unary-types* :test #'eq)
	   (setf (car formula)
		 (if (member (second formula)
			     (cdr (assoc fn *unary-types-list* :test #'eq))
			     :test #'eq)
		     :and
		   :or))
	   (setf (cdr formula) nil))
	  ((formula-combiner-p fn)
	   (cons fn (mapcar #'prune-ground-unary-types (cdr formula)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Consistency
;;;

;; Formulas are in list (not sprop) form and are timeless.

;; Don't call this; obsolete.
;; ;; Sets *possibly-{true,false}-fluents* (which are lists, for each fluent,
;; ;; of possible args) by examining initial conditions and action postconditions.
;; (defun create-fluents-list-from-inits (problem domain)
;;   ;; to prime lists with a gensym, use #'(lambda (fname) (list fname (gensym)))
;;   (setq *possibly-true-fluents* (mapcar #'list (mapcar #'first *fluents*)))
;;   (setq *possibly-false-fluents* (mapcar #'list (mapcar #'first *fluents*)))
;;   ;; Perhaps omit all types (unary and non-unary) here as well -- haven't
;;   ;; they already been dealt with?
;;   (dolist (init (problem-init problem))
;;     (let ((negated (eq :not (car init))))
;;       (if negated (setq init (second init)))
;;       (let ((assoc (assoc (car init) (if negated
;; 					 *possibly-false-fluents*
;; 				       *possibly-true-fluents*))))
;; 	(if *debug-consistency*
;; 	    (format t "Pushing ~a~s for ~s onto ~s~%"
;; 		    (if negated "NOT:" "") (cdr init) (car init) (cdr assoc)))
;; 	(pushnew (cdr init) (cdr assoc) :test #'equal))))
;; 
;;   ;;   ;; This does a pessimistic analysis (insert all possibilities, then see
;;   ;;   ;; what can be removed) rather than an optimistic one (assuming only the
;;   ;;   ;; inits, what must be removed to achieve consistency).  Both reach a
;;   ;;   ;; fixed point, but the optimistic one reaches a better (lower) fixed point.
;;   ;;   (dolist (op (domain-operators domain))
;;   ;;     (dolist (effect (conjunctive-clauses (op-effect op)))
;;   ;;       (if (not (eq :not (car effect)))
;;   ;; 	  ;; It's a bit wasteful to iterate over all bindings rather than just
;;   ;; 	  ;; the relevant ones; but this shouldn't be so bad.
;;   ;; 	  (dolist (b (op-bindings op))
;;   ;; 	    (progn
;;   ;; 	      (if *debug-consistency*
;;   ;; 		  (format t "Pushing ~s for ~s with binding ~s~%"
;;   ;; 			  (instantiate b (cdr effect)) (op-name op) b))
;;   ;; 	      (pushnew (instantiate b (cdr effect))
;;   ;; 		       (cdr (assoc (car effect) *possibly-true-fluents*))
;;   ;; 		       :test #'equal))))))
;; 
;;   ;; Use when priming lists with a gensym
;;   ;; (setq *possibly-true-fluents* (mapcar #'butlast *possibly-true-fluents*))
;;   ;; (setq *possibly-false-fluents* (mapcar #'butlast *possibly-false-fluents*))
;;   )

;;; [What does "positive" mean in this comment?
;; For a problem every operator of which is positive (eg att-log1), it might
;; be nice to just skip over negative things in the worklist.  This probably
;; isn't such a widely-applicable and important optimization.

;;; Not currently used
;; Sets *true-fluents*, *false-fluents*, and *variant-fluents*.
(defun pair-consistency (problem domain)
  ;; now done in pair-consistency-once-optimistic
  ;; (create-fluents-list-from-inits problem domain)
  (pair-consistency-once-optimistic problem domain)
  (if (not (eq *action-representation* 'overloaded-split-factored))
      (prune-from-fluents-list problem domain))
  ;; This looping shouldn't be necessary and I don't see why it's right.
  ;;   (loop while (prog1
  ;; 		  (prune-from-fluents-list problem domain)
  ;; 		(pair-consistency-once-optimistic problem domain)))
  )

;; Remove bindings inconsistent with *{true,false}-fluents*.
;; This should probably happen only once, after adding all possible bindings.
;; Returns a boolean indicating whether any change occurred.
(defun prune-from-fluents-list (problem domain)
  ;; goal may contain arbitrary disjunctions (exists turns into disjunction).

  ;; Don't do this; nsubst can confuse the cdr of a fluent for another fluent.
  ;;   (setf (problem-goal problem)
  ;;     (nsubst-if (cnf-true)
  ;; 	       (if (factored-splitting-p)
  ;; 		   #'true-unary-timeless-fluent-or-negation-p
  ;; 		 #'true-timeless-fluent-or-negation-p)
  ;; 	       (problem-goal problem)))
  ;;   (setf (problem-goal problem)
  ;;     (nsubst-if (cnf-false)
  ;; 	       (if (factored-splitting-p)
  ;; 		   #'false-unary-timeless-fluent-or-negation-p
  ;; 		 #'false-timeless-fluent-or-negation-p)
  ;; 	       (problem-goal problem)))

  ;; Perhaps this is so trivial that it doesn't even matter.  (Not sure.)
  (setf (problem-goal problem)
    ;; convert it to cnf so we can avoid using nsubst to traverse it.
    (let* ((cnf-goal (convert-to-cnf (problem-goal problem)))
	   (false-p (if (factored-splitting-p)
			#'false-unary-timeless-fluent-or-negation-p
		      #'false-timeless-fluent-or-negation-p))
	   (true-p (if (factored-splitting-p)
		       #'true-unary-timeless-fluent-or-negation-p
		     #'true-timeless-fluent-or-negation-p)))
      (convert-to-cnf
       (cons :and
	     (mapcar
	      #'(lambda (conjunct)
		  (declare (cons conjunct))
		  (assert (consp conjunct))
		  (cond ((formula-combiner-p (car conjunct))
			 (let* ((sans-true (nsubstitute-if
					    (cnf-true) true-p (cdr conjunct)))
				(sans-false (nsubstitute-if
					     (cnf-false) false-p sans-true)))
			   (cons (car conjunct) sans-false)))
			((funcall true-p conjunct)
			 (cnf-true))
			((funcall false-p conjunct)
			 (cnf-false))
			(t
			 conjunct)))
	      (conjunctive-clauses cnf-goal))))))

  ;; Perhaps simplify at this point.

  ;; Don't remove fluents known to be true from the initial conditions.
  ;; We could probably do so, but it's safer not to at this point.
  ;;  1.  If they're not used anywhere else, simplification removes them.
  ;;  2.  If they are used somewhere else, it would be an error to remove them.
  ;;   ;; Need to make sure these aren't needed at any intermediate time.
  ;;   (setf (problem-init problem)
  ;;     (delete-if (if (factored-splitting-p)
  ;; 		   #'true-unary-timeless-fluent-or-negation-p
  ;; 		 #'true-timeless-fluent-or-negation-p)
  ;; 	       (problem-init problem)))
  ;;   (assert (or (factored-splitting-p) (every #'variant-timeless-fluent-or-negation-p (problem-init problem))))

  (assert (simple-init-p (problem-init problem)))
  (assert (or (factored-splitting-p) (notany #'false-timeless-fluent-or-negation-p (cdr (problem-init problem)))))

  ;; should this use false-p and true-p?
  (flet ((false-instantiation-p (fluent binding)
	   (and (not (cnf-true-p fluent))
		(let ((instantiated (instantiate binding fluent)))
		  ;; On one line because of automatic tools for one-line assertions.
		  (assert (or (false-timeless-fluent-or-negation-p instantiated) (true-timeless-fluent-or-negation-p instantiated) (variant-timeless-fluent-or-negation-p instantiated)))
		  (if *debug-consistency*
		      (format t "    ~aremoving instantiation ~s~%        binding ~s~%"
			      (if (false-timeless-fluent-or-negation-p instantiated) "" "not ")
			      instantiated binding))
		  (false-timeless-fluent-or-negation-p instantiated)))))
    (dolist (op (domain-operators domain))
      (assert (eq :and (car (op-precondition op))))
      (if *debug-consistency*
	  (format t "operator ~s~%" (op-name op)))
      (dolist (pre (conjunctive-clauses (op-precondition op)))
	(if *debug-consistency*
	    (format t "  precondition ~s~%" pre))
	(if *debug-consistency*
	    (format t "    initial bindings ~s~%" (op-bindings op)))
	(setf (op-bindings op)
	  (delete-if #'(lambda (b) (false-instantiation-p pre b))
		     (op-bindings op)))
	(if (null (op-bindings op))
	    (format t "WARNING: no remaining bindings for ~s after precond ~s~%"
		    (op-name op) pre)))
      (dolist (eff (conjunctive-clauses (op-effect op)))
	(if *debug-consistency*
	    (format t "  effect ~s~%" eff))
	(setf (op-bindings op)
	  (delete-if #'(lambda (b) (false-instantiation-p eff b))
		     (op-bindings op)))
	(if (null (op-bindings op))
	    (format t "WARNING: no remaining bindings for ~s after effect ~s~%"
		    (op-name op) eff))))))



;; Sets *true-fluents*, *false-fluents*, and *variant-fluents*.  (Actually,
;; this seems to set these values incorrectly, since times aren't stripped.)
;; Also removes bindings incompatible with those lists.
(defun pair-consistency-once-optimistic (problem domain)
  (assert (simple-init-p (problem-init problem)))
  ;; Perhaps omit all types (unary and non-unary) here as well -- haven't
  ;; they already been dealt with?

  ;; possibly-{true,false}-fluents determine what the worklist has already seen.
  ;; Each is a list of lists of the form (<fluent-name> (<obj1> <obj2>) ...),
  ;; where (FLUENT obj1 obj2) may be, but is not necessarily, true (false).
  (let ((possibly-true-fluents (mapcar #'list (mapcar #'first *fluents*)))
	(possibly-false-fluents (mapcar #'list (mapcar #'first *fluents*)))
	(worklist (cdr (problem-init problem)))
	(binding-marks '()))

    (dolist (op (domain-operators domain))
      (assert (not (cnf-false-p (op-precondition op))))
      (dolist (b (op-bindings op))
	(push (list (cons (op-name op) b)) binding-marks))

      (if (cnf-true-p (op-precondition op))
	  (dolist (b (op-bindings op))
	    (setq worklist
	      (append (conjunctive-clauses (instantiate b (op-effect op)))
		      worklist)))))

    ;; I'll mark each binding once for each precondition; if a binding is
    ;; marked by every precondition, then it's OK; otherwise, it will later
    ;; be removed.
    (labels ((binding-contents (op b)
	       (declare (type operator op))
	       (assert (operator-p op))
	       (cdr (assoc (cons (op-name op) b) binding-marks :test #'equal)))
	     (binding-ok (op b i)
	       (declare (type operator op) (fixnum i))
	       (assert (operator-p op))
	       (member i (binding-contents op b)))
	     (binding-set-ok (op b i)
	       (declare (type operator op) (fixnum i))
	       (assert (operator-p op))
	       (assert (not (binding-ok op b i)))
	       (push i (cdr (assoc (cons (op-name op) b) binding-marks :test #'equal)))
	       (if *debug-consistency*
		   (format t "    set-ok ~s binding ~s: ~s~%"
			   (op-name op) b (binding-contents op b))))
	     (binding-fully-ok (op b)
	       (declare (type operator op))
	       (assert (operator-p op))
	       (= (length (cdr (op-precondition op)))
		  (length (binding-contents op b))))
	     (binding-pretty-print (op b)
	       (declare (type operator op))
	       (assert (operator-p op))
	       (format t "~aOK op ~s binding ~s~%"
		       (if (binding-fully-ok op b) "" "NOT ")
		       (op-name op) b)
	       (loop for pre in (cdr (op-precondition op))
		   as i = 1 then (the fixnum (1+ i))
		   do (format t "  precondition #~d: ~aok: ~s~%"
			      i (if (binding-ok op b i) "" "NOT ") pre)))
	     (binding-explain-all ()
	       (dolist (op (domain-operators domain))
		 (dolist (b (op-bindings op))
		   (binding-pretty-print op b)))))
      (if *debug-consistency*
	  (format t "initial worklist: ~s~%" worklist))
      (loop while worklist
	  do (let* ((fluent (car worklist))
		    (fluent-negated-p (eq :not (car fluent)))
		    (fluent-name (car (if fluent-negated-p (second fluent) fluent)))
		    (fluent-args (cdr (if fluent-negated-p (second fluent) fluent)))
		    (possible-assoc (assoc fluent-name
					   (if fluent-negated-p
					       possibly-false-fluents
					     possibly-true-fluents))))
	       (assert possible-assoc)
	       (if *debug-consistency*
		   (format t "processing ~a from worklist: ~s~%"
			   (if (member fluent-args (cdr possible-assoc) :test #'equal) "old" "new")
			   fluent))
	       (setq worklist (cdr worklist))

	       (if (not (member fluent-args (cdr possible-assoc) :test #'equal))
		   (progn
		     (push fluent-args (cdr possible-assoc))
		     (dolist (op (domain-operators domain))
		       (assert (eq :and (car (op-precondition op))))
		       (loop for pre in (cdr (op-precondition op))
			   as i = 1 then (the fixnum (1+ i))
			   do (let ((pre-negated-p (eq :not (car pre))))
				(if *debug-consistency*
				    (format t " op ~s pre #~d ~s~%"
					    (op-name op) i pre))
				(if (and (eq fluent-negated-p pre-negated-p)
					 (eq fluent-name (car (if pre-negated-p
								  (second pre)
								pre))))
				    (dolist (b (op-bindings op))
				      (if (and *debug-consistency*
					       *debug-consistency-verbose*)
					  (format t "  binding (~aok) ~s~%"
						  (if (binding-ok op b i) "" "not ") b))
				      ;; Do nothing if binding is already OK.
				      (if (not (binding-ok op b i))
					  (progn
					    (if (and *debug-consistency*
						     *debug-consistency-verbose*)
						(format t "   fluent test: ~s ~s~%"
							fluent (instantiate b pre)))
					    (if (equal fluent-args
						       (instantiate b (cdr (if pre-negated-p
									       (second pre)
									     pre))))
						(progn
						  (binding-set-ok op b i)
						  (if (binding-fully-ok op b)
						      (progn
							(if *debug-consistency*
							    (format t "fully OK: ~s ~s~%" (op-name op) b))
							(dolist (effect (conjunctive-clauses (op-effect op)))
							  (if *debug-consistency*
							      (format t "Add to worklist: ~s~%"
								      (instantiate b effect)))
							  (push (instantiate b effect)
								worklist)))))))))))))))))
      ;; Lots of output here
      (if *debug-consistency*
	  (format t "~s~%" binding-marks))
      (if *debug-consistency*
	  (binding-explain-all))
      (if *debug-consistency*
	  (progn
	    (format t "~s~%" (domain-operators domain))
	    (format t "CHECKING ================~%")
	    (dolist (op (domain-operators domain))
	      (format t "op ~s pre ~s~%" (op-name op) (op-precondition op))
	      (dolist (b (op-bindings op))
		(if (binding-fully-ok op b)
		    (format t "   fully ok: ~s~%" b)
		  (format t "   okayed by pre ~s: ~s~%" (binding-contents op b) b))))))

      (setq *true-fluents* '())
      (setq *false-fluents* '())
      (setq *variant-fluents* '())
      ;; not dolist because multiple arguments
      (mapc #'(lambda (op+true-args-list op+false-args-list)
		(assert (eq (car op+true-args-list) (car op+false-args-list)))
		(let ((fname (car op+true-args-list))
		      (true-args (cdr op+true-args-list))
		      (false-args (cdr op+false-args-list)))
		  (push (cons fname (set-difference-equal true-args false-args))
			*true-fluents*)
		  (push (cons fname (intersection-equal true-args false-args))
			*variant-fluents*)
		  (push (cons fname (set-difference-equal false-args true-args))
			*false-fluents*)))
	    possibly-true-fluents possibly-false-fluents)


      (dolist (op (domain-operators domain))
 	(setf (op-bindings op)
 	  (delete-if-not #'(lambda (b) (binding-fully-ok op b))
 			 (op-bindings op))))
      )))

;; Sets *true-fluents*, *false-fluents*, and *variant-fluents*.
;; Also sets *possible-fluents*.
;; Also removes bindings incompatible with those lists.
(defun pair-consistency-2 (problem domain)
  (assert (simple-init-p (problem-init problem)))

  ;; Perhaps omit all types (unary and non-unary) here as well -- haven't
  ;; they already been dealt with?

  ;; possible-fluents, a hash of sprops, determines what the worklist has
  ;; already seen.  Those fluents may be, but are not necessarily, true
  ;; (false, if they appear negated).
  (let ((possible-fluents (make-hash-table :test #'eq))
	(possible-bindings (mapcar #'(lambda (op) (list (op-name op)))
				   (domain-operators domain)))
	(worklist (mapcar #'make-timeless-sprop-from-list-or-negation
			  (cdr (problem-init problem))))
	;; action-hash maps from an action sprop to a list of
	;; (preconds effects binding).  Preconds are removed as they are
	;; satisfied; when no more preconds, the entire entry is removed.
	(action-hash (make-hash-table :test #'eq))
	(fluent-hash (make-hash-table :test #'eq)))
    (assert (every #'sprop-p worklist))

    ;; Initialize action-hash, fluent-hash, worklist
    (dolist (op (domain-operators domain))
      (setf (op-precondition op) (convert-to-cnf (op-precondition op)))
      (if (cnf-false-p (op-precondition op))
	  (format t "Warning: operator ~s has false precondition in pair-consistency~%" (op-name op))
	(dolist (b (op-bindings op))
	  (if *debug-consistency* (format t "~s binding ~s~%" (op-name op) b))
	  (if (and (and *debug-consistency* *debug-consistency-verbose*)
		   (eq (op-name op) 'open))
	      (print-operator-long op))
	  (let ((precond-clauses (conjunctive-clauses
				  (timeless-spropify-formula
				   (instantiate b (op-precondition op)))))
		(effect-clauses (conjunctive-clauses
				 (timeless-spropify-formula
				  (instantiate b (op-effect op)))))
		(op-sprop (make-sprop (op-name op) (mapcar #'cdr b) nil nil)))
	    (if (and *debug-consistency* *debug-consistency-verbose*)
		(format t "Checking for all sprop-p...~%"))
	    (assert (every #'sprop-p effect-clauses))
	    (if (and *debug-consistency* *debug-consistency-verbose*)
		(format t "Done checking for all sprop-p.~%"))
	    (if (null precond-clauses)
		(progn
		  (if *debug-consistency-verbose*
		      (format t "No real precond clauses for ~s~%" op-sprop))
		  (setq worklist (append effect-clauses worklist)))
	      (progn
		(if *debug-consistency*
		    (progn
		      (format t "Real precond clauses exist for ~s~%" op-sprop)
		      (if *debug-consistency-verbose*
			  (format t "Precond clauses are ~s~%" precond-clauses))))
		(setf (gethash op-sprop action-hash)
		      (list precond-clauses effect-clauses b))
		;; (format t "precond-clauses: ~s~%" precond-clauses)
		(dolist (precond-clause precond-clauses)
		  (dolist (precond-sprop (disjunctive-clauses precond-clause))
		    (declare (type sprop precond-sprop))
		    (assert (sprop-p precond-sprop))
		    (pushnew op-sprop (gethash precond-sprop fluent-hash))
		    (assert (member op-sprop (gethash precond-sprop fluent-hash)))))))))))

    (if *debug-consistency*
	(format t "initial worklist: ~s~%" worklist))
    (if *debug-consistency*
	(progn
	  (format t "action-hash:~%")
	  (maphash #'(lambda (key val) (format t "  ~s: ~s~%" key val))
		   action-hash)))
    (if *debug-consistency*
	(progn
	  (format t "fluent-hash:~%")
	  (maphash #'(lambda (key val) (format t "  ~s: ~s~%" key val))
		   fluent-hash)))

    ;; set possible-fluents, possible-bindings
    (loop while worklist
	do (let ((fluent (car worklist)))
	     (declare (type sprop fluent))
	     (assert (sprop-p fluent))
	     (if *debug-consistency*
		 (format t "processing ~a from worklist: ~s~%"
			 (if (gethash fluent possible-fluents) "old" "new")
			 fluent))

	     (setq worklist (cdr worklist))
	     (if (not (gethash fluent possible-fluents))
		 (progn
		   (setf (gethash fluent possible-fluents) t)
		   (if *debug-consistency*
		       (format t "actions relevant to ~s: ~s~%"
			       fluent (gethash fluent fluent-hash)))
		   (dolist (action (gethash fluent fluent-hash))
		     ;; action has fluent as a precondition
		     (declare (type sprop action))
		     (assert (sprop-p action))
		     (let* ((this-act-info (gethash action action-hash))
			    (preconds (delete-if #'(lambda (precond)
						     (if (sprop-p precond)
							 (sprop-equal precond fluent)
						       (progn
							 (assert (eq :or (car precond)))
							 (member fluent (cdr precond)))))
						 (first this-act-info))))
		       (if (and *debug-consistency* *debug-consistency-verbose*)
			   (format t "Remaining preconds for ~s:~%    ~s~%"
				   action preconds))
		       (if (null preconds)
			   (progn
			     (if (and *debug-consistency* *debug-consistency-verbose*)
				 (format t "Adding effects to worklist: ~s~%"
					 (second this-act-info)))
			     ;; should this be adjoin?
			     ;; (setq worklist (nconc (second this-act-info) worklist))
			     (setq worklist (append (second this-act-info) worklist))
			     (remhash action action-hash)
			     (push (third this-act-info) (cdr (assoc (sprop-name action) possible-bindings))))
			 (setf (first this-act-info) preconds))))))))
    (setq *possible-fluents* possible-fluents)

    ;; set *{true,false,variant}-fluents*
    (setq *true-fluents* '())
    (setq *false-fluents* '())
    (setq *variant-fluents* '())
    (maphash #'(lambda (fluent ignore)
		 (declare (type sprop fluent) (ignore ignore))
		 (assert (sprop-p fluent))
		 (if (gethash (sprop-negate fluent) possible-fluents)
		     ;; Do I want to both positive and negative versions
		     ;; in *variant-fluents*?  Unlikely...
		     (if (not (sprop-negated-p fluent))
			 (let* ((name (sprop-name fluent))
				(assoc (assoc name *variant-fluents*)))
			   (if assoc
			       (push (sprop-args fluent) (cdr assoc))
			     (push (cons name (list (sprop-args fluent)))
				   *variant-fluents*))))
;;			 (assoc-push (sprop-name fluent) (sprop-args fluent)
;;				     *variant-fluents*)
		   (if (sprop-negated-p fluent)
		       (let* ((name (sprop-name fluent))
			      (assoc (assoc name *false-fluents*)))
			 (if assoc
			     (push (sprop-args fluent) (cdr assoc))
			   (push (cons name (list (sprop-args fluent)))
				 *false-fluents*)))
;;		       (assoc-push (sprop-name fluent) (sprop-args fluent)
;;				   *false-fluents*)
		     (let* ((name (sprop-name fluent))
			    (assoc (assoc name *true-fluents*)))
		       (if assoc
			   (push (sprop-args fluent) (cdr assoc))
			 (push (cons name (list (sprop-args fluent)))
			       *true-fluents*))))))
;;		     (assoc-push (sprop-name fluent) (sprop-args fluent)
;;				 *true-fluents*)
	     possible-fluents)

    ;; Prune operator bindings.
    (let ((bad-bindings-assoc '()))
      (maphash #'(lambda (action pre+eff+bind)
		   (let* ((name (sprop-name action))
			  (assoc (assoc name bad-bindings-assoc)))
		     (if assoc
			 (push (third pre+eff+bind) (cdr assoc))
		       (push (cons name (list (third pre+eff+bind)))
			     bad-bindings-assoc))))
;;		     (assoc-push (sprop-name action) (third pre+eff+bind)
;;				 bad-bindings-assoc)
	       action-hash)
      (dolist (op (domain-operators domain))
	(setf (op-bindings op)
	  (set-difference-equal (op-bindings op)
				(cdr (assoc (op-name op) bad-bindings-assoc))))))
    ;; This is done later, but do it early to avoid a check below.
    (setf (domain-operators domain)
      (delete-if-not #'op-bindings (domain-operators domain)))
    (if (not (domain-operators domain))
	(format t "Eliminated all operators in pair-consistency-2~%"))

    ;; TO DO: quite possibly we should do a similar thing for preconditions.

    ;; Eliminate effects that always reinforce the status quo.
    (dolist (op (domain-operators domain))
      (assert (op-bindings op))
      (setf (op-effect op)
	(cons
	 :and
	 (delete-if #'(lambda (effect)
			;; This tests whether *any* instantiation of the fluent
			;; is variant.  We really want to look at the bindings
			;; and if none of the instantiations of this effect are
			;; variant for any of the bindings, then return true.
			(let* ((negated-p (eq (car effect) :not))
			       (non-negated (if negated-p (second effect) effect))
			       (fn (car non-negated))
			       (args (cdr non-negated))
			       (can-be-true-p nil)
			       (can-be-false-p nil)
			       (can-be-variant-p nil)
			       (true-instantiations (assoc fn *true-fluents*))
			       (false-instantiations (assoc fn *false-fluents*))
			       (variant-instantiations (assoc fn *variant-fluents*)))
			  ;; I could sometimes avoid the loop: if
			  ;;  (not (assoc fn *variant-fluents*))
			  ;; then *no* instantiation of the fluent is variant.
			  ;; But I will do the more involved thing for checking.
			  (dolist (b (op-bindings op))
			    (let ((instantiated-args (instantiate b args)))
			      (cond ((member-equal instantiated-args true-instantiations)
				     (setq can-be-true-p t))
				    ((member-equal instantiated-args false-instantiations)
				     (setq can-be-false-p t))
				    ((member-equal instantiated-args variant-instantiations)
				     (setq can-be-variant-p t))
				    (t
				     (error "Not false, true, or variant: ~s"
					    (cons fn instantiated-args))))
			      (if (or can-be-variant-p (and can-be-false-p can-be-true-p))
				  ;; exit the dolist early
				  ;; maybe avoid this for assertions/debugging?
				  (return))))
			  (assert (or can-be-true-p can-be-false-p can-be-variant-p))
			  (assert (or can-be-variant-p (xor can-be-true-p can-be-false-p)))
			  (assert (or (assoc (car non-negated) *variant-fluents*) (if negated-p (and can-be-false-p (not can-be-true-p)) (and can-be-true-p (not can-be-false-p)))))
			  ;; If it's not variant, and it is either
			  (and (not can-be-variant-p) (xor can-be-true-p can-be-false-p))))
		    (conjunctive-clauses (op-effect op))))))
    ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;; Setting up stuff for parallel exclusion

; given a formula in list format, return its negation
(defun formula-negate (form)
  (cond ((sprop-p form)
	 (sprop-negate form))
	((eq (car form) :not)
	 (cadr form))
	(t
	 (list :not form))))

; given a formula to be negated, push the negation in one level.
(defun push-negation-once (form)
  (cond
   ((eq (car form) :and) (cons :or (mapcar #'formula-negate (cdr form))))
   ((eq (car form) :or) (cons :and (mapcar #'formula-negate (cdr form))))
   (t (formula-negate form))))

;; Since the not-pushing doesn't happen by side effect, oughtn't we do that
;; first and then call this on the result?  (We must push nots eventually...)
; given a formula of ANDs, ORs and NOTs, push the NOTs all the way in to the
; literals, at the same time creating a list of all literals in the formula.
(defun all-literals (form)
  (cond
   ((sprop-p form)
    (list form))
   ((or (eq (car form) :and) (eq (car form) :or))
    (mapcan #'all-literals (cdr form)))
   ((and (eq (car form) :not) (member (car (second form)) '(:and :or :not)))
    (all-literals (push-negation-once (cadr form))))
   ((and (eq (car form) :not) (sprop-p (second form)))
    (list (sprop-negate (second form))))
   (t
    (error "What formula? ~s" form))))

; get all literals of a formula and remove their time parameters
(defun all-lits-no-time (form)
  (mapcar #'sprop-remove-time (all-literals form)))



;; Converts a forall into a conjunction of all appropriate objects
(defun forall->and (form prob)
  ;; base case: if form is a symbol, return it
  (if (symbolp form)
      form
    (let ((fn (car form)))  ;; The first element in form
      (cond
       ((member fn '(:eq :neq) :test #'eq)
	form)
       ;; if we have anything bu a forall, leave it alone
       ;; and recursively convert the pieces
       ((member fn '(:and :or :not :exists :when) :test #'eq)
	;; recursing on the binding for an :exists is unnecessary, but
	;; should be harmless
	(cons fn (mapcar #'(lambda (f) (forall->and f prob))
			 (cdr form))))
       ;; If we do have a forall, then we find the quantified variable
       ;; and replace it in the expression for each object that
       ;; fits (either all things or all tools, frinstance)
       ;; and conjoin the results into conj-list
       ((eq :forall fn)
	;; A(x)(x is blue)
	;; variable-name is 'x'
	;; things-to-instantiate are all the obejects that could fit x
	(let ((parsed-forall-bindings (parse-bindings (second form))))
	  (if (intersection (mapcar #'second parsed-forall-bindings)
			    *vacuous-fluents-used-as-types*)
	      (cnf-true)
	    ;; work from the inside out to get nested scoping right
	    (let ((forall-body (forall->and (third form) prob)))
	      (dolist (var+type parsed-forall-bindings)
		(let ((variable-name (first var+type))
		      (things-to-instantiate
		       (objects-of-type (second var+type) prob)))
		  (setq forall-body
			;; now we just need to dolist things-to-instantiate, replacing
			;; into what each time, then make all that an :and
			(cons :and (mapcar #'(lambda (item)
					       (subst item variable-name forall-body))
					   things-to-instantiate)))))
	      forall-body))))
       ((keywordp fn)
	(error "What formula combiner? ~s in ~s" fn form))
       (t form)))))


;; Tests
;; (forall->and   '(:and (:not (jacked ?h)) (:forall (domains::?y) (:when (on domains::?y ?h) (on domains::?y ground))) (have jack)) (get-problem-copy 'fixit2))
;; (forall->and '(:forall (DOMAINS::tool ?x) (in ?x boot)) (get-problem-copy 'fixit2))
;; (forall->and '(:AND (:AND (:NOT (JACKED ?H)) (:FORALL (?Y)) (HAVE JACK)) (ON ?Y GROUND)) (get-problem-copy 'fixit2 ))
;; (forall->and '(:forall (domains::person domains::?p) (:exists (object ?c) (:and (written-for domains::?p ?c)))) (get-problem-copy 'office1))
;; (forall->and '(:forall (?x) (:and (eq ?x y) (:forall (domains::person ?p) (foo ?p ?x)))) (get-problem-copy 'office1))



;; Not used any longer -- replaced by parse-bindings and objects-of-type.
;; ;; Inits is the problem-inits of the current problem
;; ;; what-to-replace should be of the form (?X) or (tool ?X)
;; ;; In the former candidates is a list of all the objects in the
;; ;; universe.  In the latter case it is a list of all the tools
;; (defun candidates (inits what-to-replace)
;;   (if (or (symbolp what-to-replace)
;; 	  (= 1 (length what-to-replace)))
;;       *objects*
;;     (let ((return-list '())
;; 	  ;; criterion is the property which  objects in the
;; 	  ;; inits need to have ( e.g. tool, block, person)
;; 	  ;; to be added to the return-list
;; 	  (criterion (car what-to-replace)))
;;       (dolist (an-item inits)
;; 	(when (eq (car an-item) criterion)
;; 	  (setf return-list (adjoin  (second an-item) return-list :test #'equal))))
;;       return-list)))
;; ;; Tests
;; ;;  (candidates '((tree a) (Bush b) (Tree c) (rock d) (green a)) '(tree ?x))


;; Should work whether or not *unary-types-list* is set yet.
;; The result may be freely modified (this is a tad wasteful, but safe).
(defun objects-of-type (type problem)
  (cond ((not type)
	 (copy-list *objects*))
	((eq *vacuous-fluents-used-as-types* 'not-yet-set)
	 (error "*vacuous-fluents-used-as-types* not yet set"))
	((and ;; (not (eq *vacuous-fluents-used-as-types* 'not-yet-set))
	      (member type *vacuous-fluents-used-as-types*))
	 nil)
	((eq *unary-types* 'not-yet-set)
	 (error "*unary-types* not yet set"))
	((and ;; (not (eq *unary-types* 'not-yet-set))
	      (not (member type *unary-types*)))
	 (error "Can't cope with non-type argument to objects-of-type: ~s" type))
	((eq *unary-types-list* 'not-yet-set)
	 (loop for item in (cdr (problem-init problem))
	       when (eq type (car item))
	       collect (second item)))
	(t
	 (cdr (assoc type *unary-types-list*)))))


;; Returns a list of (varname type), or nil if the (single-variable)
;; binding can't be parsed.
(defun parse-binding (binding)
  (cond ((symbolp binding)
	 (and (varp binding)
	      (list binding nil)))
	((and (= 1 (length binding))
	      (varp (car binding)))
	 (list (car binding) nil))
	((and (= 2 (length binding))
	      (symbolp (car binding))
	      (not (varp (car binding)))
	      (varp (second binding)))
	 (list (second binding) (first binding)))
	(t
	 nil)))

;; Returns a list (not a cons) of (varname type) lists, where TYPE may be nil.
;; Will need to be extended for the Nisp-style syntax.
(defun parse-bindings (forall-bindings)
  (let ((as-single-binding (parse-binding forall-bindings)))
    (if as-single-binding
	(list as-single-binding)
      (let ((result (mapcar #'parse-binding forall-bindings)))
	(if (member nil result :test #'eq)
	    (error "Can't parse bindings ~s" forall-bindings))
	result))))
;; Testing
;; (parse-bindings '?x)
;; (parse-bindings '(?x))
;; (parse-bindings '(?x ?y))
;; (parse-bindings '(tool ?x))
;; (parse-bindings '(?x ?y ?z))
;; (parse-bindings '(tool ?x ?y))      ; error
;; (parse-bindings '((tool ?x)))
;; (parse-bindings '(?x (?y) (tool ?z) (boot ?w)))
;; (parse-bindings '(?x (?y) (?z ?z))) ; error
;; (parse-bindings '(?x (?y) ((?z ?z)))) ; error


;; Given ?P, returns a name not in *used-variables*
;; Side effect: Adds variable to *used-variables*
(defun rename (variable)
  (assert (symbolp variable))
  (let ((result (if (member variable *used-variables*)
		    (gentemp (symbol-name variable))
		  variable)))
    (setf *used-variables* (cons result *used-variables*))
    result))


;; Takes a formula and renames any quantified variables that conflict
(defun alpha-rename (form)
  (setf *used-variables* nil)
  (the-real-alpha-rename form))
;; (alpha-rename '(:AND (:EXISTS (?F) (REFRIED-BEANS  ?F)) (:EXISTS (?F) (BEEF  ?F)) (:EXISTS (?F) (TORTILLAS  ?F))))
;; (alpha-rename  '(:EXISTS (ANIMAL ?P) (:AND (:FORALL (?P) (COW ?P)) (BROWN ?P) )))
;; (alpha-rename '(:EXISTS (?P) (:WHEN ( (:EXISTS (ANIMAL ?P) (:AND (:FORALL (?P) (COW ?P)) (BROWN ?P)))  )  (DOG ?P)  )  ))
;; (alpha-rename '(:WHEN ( (:EXISTS (ANIMAL ?P) (:AND (:FORALL (?P) (COW ?P)) (BROWN ?P)))  ) (DOG ?X)))
;; (alpha-rename '(:WHEN ( (:EXISTS (ANIMAL ?P) (:AND (:FORALL (?P) (COW ?P)) (BROWN ?P)))  ) (DOG ?P)))

;; Takes a formula and renames any quantified variables that conflict
;; Side-effects *used-variables*
(defun the-real-alpha-rename (form )
  ;; base case: if form is a symbol, return it
  (if (symbolp form)
      form
    (let ( (fn (the-real-alpha-rename (car form))   ) )  ;; The first element in form
      ;; If the first thing in the form is a quantifier, we check
      ;; out the quantifed variable.
      (if (member fn '(:EXISTS :FORALL))
	  (let ((q-variable nil))
	    (progn
	      ;; the second in the form should be the quantified variable
	      (assert (listp (second form)))
	      (setf q-variable  (car (car (parse-bindings (second form)))))
	      (cons fn (mapcar #'(lambda (f) (the-real-alpha-rename f))
			       (subst (rename q-variable) q-variable  (cdr form))))))
	(cons fn (mapcar #'(lambda (f) (the-real-alpha-rename f))
			 (cdr form)))))))

;; (alpha-rename '(:AND (:EXISTS (?F) (REFRIED-BEANS  ?F)) (:EXISTS (?F) (BEEF  ?F)) (:EXISTS (?F) (TORTILLAS  ?F))))
;; (alpha-rename  '(:EXISTS (ANIMAL ?P) (:AND (:FORALL (?P) (COW ?P)) (BROWN ?P) )))


;; Returns a list (not a cons) of (varname type) lists, where TYPE may be nil
;; for Nisp-style syntax
(defun nisp-parse-bindings (forall-bindings)
  ;; If forall-bindings is not a list it should be a variable
  (if (not (listp forall-bindings))
      (progn
	(assert (varp forall-bindings))
	(cons (cons forall-bindings (cons nil nil)) nil))
    ;; if forall-bindings IS a list, we look through it an item at a time...
    (let ( (var-list nil) (return-list nil) (last-item nil) )
      (dolist (an-item forall-bindings)
	(assert (not (listp an-item)))
	(cond
	 ;; if this is a variable, add it to the list var-list
	 ((varp an-item)
	  (setf var-list (cons an-item var-list))
	  )
	 ;; Do nothind with a '-
	 ( (equal '- an-item)
	   )

	 ;; if this is a symbol, unload all the variable from var-list
	 ;; and add their bindings to return-list
	 ((symbolp an-item)
	  (assert (equal last-item '-))
	  (dolist (a-var var-list)
	    (setf return-list  (cons (cons a-var (cons an-item nil)) return-list)))
	  (setf var-list nil))

	 (t (error "Can't parse bindings ~s" forall-bindings)))
	(setf last-item an-item)	; to make sure '- preceeds a symbol
	)
      (dolist (a-var var-list)
	(setf return-list  (cons (cons a-var (cons nil  nil)) return-list))
	)
      return-list)))

;; Tests                                             expected result

;; (nisp-parse-bindings '?x)                         ((?X NIL))
;; (nisp-parse-bindings '(?x))                       ((?X NIL))
;; (nisp-parse-bindings '(?x ?y - dog ?z) )          ((?X DOG) (?Y DOG) (?Z NIL))
;; (nisp-parse-bindings '(?y - dog) )                ((?Y DOG ))
;; (nisp-parse-bindings '(?x - WORM ?y - CHEESE ?z)) ((?Z NIL) (?Y CHEESE) (?X WORM))
;; (nisp-parse-bindings '(?x ?y DOG ))                error
