;; domains.lisp -- read domain and problem definitions in UCPOP format
(in-package domains)

(export '(define def-domain def-problem *domains* *problems*

	  ;; Domains
	  domain-name domain-requirements domain-operators domain-problem-names
	  problem-name problem-domain problem-init problem-goal problem-length
	  domain-p problem-p
	  print-domain print-domain-short print-domain-long
	  print-problem print-problem-short print-problem-long

	  ;; Operators
	  operator-p make-operator
	  op-name op-parameters op-precondition op-effect op-formals
	  op-formal-types op-obj-lst op-bindings
	  print-operator print-operator-short print-operator-long

	  check-requirements get-domain get-problem
	  copy-problem get-domain-copy get-problem-copy
	  ;; Export the types
	  domain operator problem))

;;; Functions that accept domain and problem definitions in UCPOP format

(defmacro DEFINE ((dtype name) &body body)
  (case dtype
    (domain `(apply #'def-domain '(,name ,@body)))
    (problem `(apply #'def-problem '(,name ,@body)))
    ))

(defvar *domains* nil
  "List of all domain names")
(defvar *problems* nil
  "List of all problem names")

;;; Helper function
(defun required ()
  "If this ever gets called, it means something that was required was not
  supplied.  Use as default value for &key args or defstruct slots."
  (error "A required argument was not supplied."))

(defstruct (domain (:print-function print-domain-short))
  (name (required) :type symbol)
  (requirements nil :type list)
  (operators (required) :type list)
  (problem-names nil :type list) ; list of symbols
  )

(defun print-domain-short (domain &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "<DOMAIN ~S>" (domain-name domain)))

(defun print-domain-long (domain &optional (stream t) depth)
  (declare (ignore depth))
  (if (symbolp domain)
      (setq domain (or (get-domain 'domain)
		       (error "Can't find domain ~s" domain))))
  (format stream "~&<DOMAIN ~S
 Requirements ~S
 Operators:  ~S
 Problems:  ~S>~%"
          (domain-name domain)
          (domain-requirements domain)
          (mapcar #'op-name (domain-operators domain))
          (domain-problem-names domain)))

(defun print-domain (domain &optional (stream t) depth)
  (print-domain-long domain stream depth))

(defun def-domain (name &rest clauses)
  (let ((operators nil)
        (requirements nil)
        (warnings nil)
        (problems (loop for problem in *problems*
                        when (eql name (problem-domain (get problem 'problem)))
                         collect problem)))
    (dolist (clause clauses)
      (case (car clause)
        (:requirements
         (setq requirements (union (cdr clause) requirements)))
        (:operator
         (push (apply #'def-operator (second clause) name (cddr clause))
	       operators))
        (t (pushnew (car clause) warnings))))
    (when warnings
      (format t "~&;~4TWarning: unable to process parameter~p, ~{~S, ~}in domain ~S~%"
	      (length warnings) warnings name))
    (let ((domain (make-domain :name name
                               :requirements requirements
                               :operators operators
                               :problem-names problems)))
      (setf (get name 'domain) domain)))
  ;;(export name)
  (pushnew name *domains*)
  name)

(defun get-domain (name)
  (declare (symbol name))
  (get (find-symbol (symbol-name name) :domains) 'domain))

(defun get-domain-copy (name)
  (declare (symbol name))
  (let ((dom (get-domain name)))
    (if dom
	(make-domain
	 :name (domain-name dom)
	 :requirements (copy-tree (domain-requirements dom))
	 :operators (mapcar #'operator-copy (domain-operators dom))
	 :problem-names (copy-tree (domain-problem-names dom)))
      nil)))

;;;This is the operator description.

(defstruct (operator (:conc-name op-) (:print-function print-operator-short))
  (name (required) :type symbol)
  (parameters (required) :type list)  ; raw form; may include type information
  (formals (required) :type list)     ; list of varp; derived from parameters
  (formal-types (required) :type list)   ; list of (var type) lists, derived from parameters
  (precondition (required) :type list)
  (effect (required) :type list)
  (obj-lst nil :type list)
  (bindings nil :type list))

(defun print-operator-short (operator &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "<Operator: ~S>" (op-name operator)))

(defun print-operator (operator &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "<Operator ~S
 Parameters: ~S
 Precondition:  ~S
 Effect:  ~S>"
          (op-name operator)
	  (op-parameters operator)
	  (op-precondition operator)
	  (op-effect operator)))

(defun print-operator-long (operator &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "~&<Operator ~S
 Parameters: ~S
 Formals: ~S
 Typing: ~S
 Precondition:  ~S
 Effect:  ~S
 Obj-lst: ~S
 Bindings: ~S>~%"
	  (op-name operator)
	  (op-parameters operator)
	  (op-formals operator)
	  (op-formal-types operator)
	  (op-precondition operator)
	  (op-effect operator)
	  (op-obj-lst operator)
	  (op-bindings operator)))

;; PROBLEM is used only to make warning messages more comprehensible
(defun def-operator (name domain &key parameters precondition effect)
  (if (not effect)
      (error "Not sensible to define operator ~a with no effect" name))
  (let (formals typing)
    (dolist (parameter parameters)
      (cond ((symbolp parameter)
	     (push parameter formals))
	    ;; parameter is a cons
	    ((= 2 (length parameter))
	     (push (list (second parameter) (first parameter)) typing)
	     (push (second parameter) formals))
	    ((= 1 (length parameter))
	     ;; no info about type
	     (push (first parameter) formals))
	    (t
	     (format t "Warning: parameters should be symbols or two-element (type varname) lists~%  in operator ~s in domain ~s: ~s" name domain parameter)
	     (setq formals (append (cdr parameter) formals)))))
    (setq formals (nreverse formals)
          typing (nreverse typing))
    (if (< (length (remove-duplicates formals)) (length formals))
	(error "Duplicate formal name in operator ~s in domain ~s:~%  ~s"
	       name domain parameters))
    (let ((unbound-precond-vars (unbound-vars precondition formals)))
      (if unbound-precond-vars
	  (format t "~&;~4TWarning: unbound variable~p ~{~s ~}in precondition of ~s in domain ~s~%"
		  (length unbound-precond-vars) unbound-precond-vars name domain)))
    (let ((unbound-effect-vars (unbound-vars effect formals)))
      (if unbound-effect-vars
	  (format t "~&;~4TWarning: unbound variable~p ~{~s ~}in effect of ~s in domain ~s~%"
		  (length unbound-effect-vars) unbound-effect-vars name domain)))
    (setf (get name 'operator)
	  (make-operator :name name
			 :parameters parameters
			 :formals formals
			 :formal-types typing
			 :precondition (or precondition (list :and))
			 :effect effect))))

(defun operator-copy (op)
  (make-operator
   :name (op-name op)
   :parameters (copy-tree (op-parameters op))
   :formals (copy-tree (op-formals op))
   :formal-types (copy-tree (op-formal-types op))
   :precondition (copy-tree (op-precondition op))
   :effect (copy-tree (op-effect op))))

;;;The problem structure remembers the domain, goals, and initial conditions
;;;associated with a single problem instance.  Don't try to reuse these.  Create
;;;it, use it and throw it away.

(defstruct (problem (:print-function print-problem-short)
		    ;; we define copy-problem below
		    (:copier nil))
  (name (required) :type symbol)
  (domain (required) :type symbol)
  (requirements nil :type list)
  (init (required) :type list)
  (goal (required) :type list)
  (length nil :type  list)
  )

(defun print-problem-short (problem &optional (stream t) depth)
  (declare (ignore depth))
  (format stream "<PROBLEM ~S ~S>"
          (problem-name problem)
          (problem-goal problem)))

(defun print-problem-long (problem &optional (stream t) depth)
  (declare (ignore depth))
  (if (symbolp problem)
      (setq problem (or (get-problem problem)
			(error "Can't find problem ~s" problem))))
  (format stream "~&<PROBLEM ~S (domain ~S)
 Goal:  ~S
 Initial-Condition:  ~S
 Length ~S>~%"
          (problem-name problem)
          (problem-domain problem)
          (problem-goal problem)
          (problem-init problem)
	  (problem-length problem)))

(defun print-problem (problem &optional (stream t) depth)
  (print-problem-long problem stream depth))

(defun def-problem (name &rest all-keys  &key domain requirements init goal length
                         &allow-other-keys)
  (if (not (and (consp init) (not (listp (car init)))))
      (error "Malformed init for problem ~a (should be formula): ~a" name init))
  (let ((unbound-init-vars (unbound-vars init)))
    (if unbound-init-vars
	(error "Unbound variable~p ~a in init of ~a in problem ~a:~%    ~a"
	       (length unbound-init-vars) unbound-init-vars name name init)))
  (if (not (and (consp goal) (not (listp (car goal)))))
      (error "Malformed goal for problem ~a (should be formula): ~a" name goal))
  (let ((unbound-goal-vars (unbound-vars goal)))
    (if unbound-goal-vars
	(error "Unbound variable~p ~a in goal of ~a in problem ~a:~%    ~a"
	       (length unbound-goal-vars) unbound-goal-vars name name goal)))

  (let ((problem (make-problem
                  :name name
                  :domain domain
                  :requirements requirements
                  :init init
                  :goal goal
                  :length length)))
    ;; The problem will be looked up in the :domains package.
    (if (not (eq (symbol-package name) (find-package :domains)))
        (error "problem name ~S is in ~S package, should be in :domains package"
               name (symbol-package name)))
    (setf (get name 'problem) problem)

    ;; Check arguments for unrecognized tags
    (let ((counter 0))
      (dolist (a-key all-keys)
	(setf counter (+ counter 1))
	(when (and (oddp counter)
		   (not (member a-key '(:goal :init :length :domain :requirements))))
	  (format t "~&;~4TWarning: unable to process parameter ~s in problem ~s~%"  a-key  name))))

    ;;(export name)
    (pushnew name *problems*)
    (let ((domain-struct (get domain 'domain)))
      (when (domain-p domain-struct)
        (pushnew name (domain-problem-names domain-struct))))
    name))
;; Test
;; (def-problem 'Runto12
;;  :domian  'mcd-grid-world   ;; misspelled domains
;;  :goal '(at-robot 1 2)
;;  :init `(:and ,@*inits55*))


(defun get-problem (name)
  (get (find-symbol (symbol-name name) :domains) 'problem))

(defun copy-problem (prob)
  (make-problem
   :name (problem-name prob)
   :domain (problem-domain prob)
   :requirements (copy-tree (problem-requirements prob))
   :init (copy-tree (problem-init prob))
   :goal (copy-tree (problem-goal prob))
   :length (copy-tree (problem-length prob))))

(defun get-problem-copy (name)
  (let ((prob (get-problem name)))
    (if prob
	(copy-problem prob)
      nil)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Checking requirments

;; When this or any macro changes, consider doing
;;    (delete-all-compiled) (compile-all)

;; PROBLEM and DOMAIN may also be names (symbols)
(defun check-requirements (capabilities problem &optional domain no-err)
  (if (symbolp problem)
      (setq problem (get-problem problem)))
  (if (not domain)
      (setq domain (problem-domain problem)))
  (if (symbolp domain)
      (setq domain (get-domain domain)))
  (let* ((domain-requirements (domain-requirements domain))
         (problem-requirements (problem-requirements problem))
         (trouble
          (set-difference (reduce-requirements (union problem-requirements
                                                      domain-requirements))
                          (reduce-requirements capabilities))))
    (when (and trouble (not no-err))
      (error "Planner is unlikely to work on ~S because of
~{~S ~}
in the domain or problem description." (problem-name problem) trouble))
    (null trouble)))

(defparameter *requirement-reductions*
  '((:adl :conditional-effects :disjunctive-preconditions
          :quantified-preconditions :quantified-effects :typing)
    (:ucpop :adl :domain-axioms :procedural-attachments :safety-constraints)))

(defun reduce-requirements (reqs)
  (loop for req in reqs
	nconc (let ((reduction (assoc req *requirement-reductions*)))
		(if reduction
		    (reduce-requirements (rest reduction))
		  (list req)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Utilities

;; Don't export varp; other packages have their own copies

; tests whether an atom is a variable
;; (would it be faster to memoize the answers in a hash table?)
(defun varp (a)
  (and (atom a)
       (not (numberp a))		; numbers can't be coerced to string
       (let ((astring (string a)))
	 (declare (type simple-base-string astring))
	 (char= (char astring 0) #\?))))

(defun unbound-vars (formula &optional bound-vars)
  (assert (listp formula))
  (let ((fn (car formula)))
    (declare (symbol fn))
    (check-type fn symbol)
    (cond
     ((eq fn :not)
      (unbound-vars (second formula) bound-vars))
     ((member fn '(:forall :exists))
      (let ((sub-bound-vars
	     (reduce #'union (mapcar #'(lambda (b)
					 (cond ((varp b)
						(list b))
					       ((symbolp b)
						'())
					       ((consp b)
						(remove-if-not #'varp b))
					       (t
						(error "What binding? ~a" b))))
				     (second formula)))))
	(let ((already-bound-vars (intersection sub-bound-vars bound-vars)))
	  (if already-bound-vars
	      (format t "~&;~4TWarning: quantifier uses already-bound variable~p~{ ~s~}:~%;       ~a~%"
		      (length already-bound-vars) already-bound-vars formula)))
	(unbound-vars (third formula) (append sub-bound-vars bound-vars))))
     ((and (keywordp fn) (not (member fn '(:eq :neq))))
      ;; eg, :and :or :not :when :oneof :uncertain
      (let ((subformulas (cdr formula)))
	(if (null subformulas)
	    '()
	  (reduce #'union (mapcar #'(lambda (subformula)
				      (unbound-vars subformula bound-vars))
				  subformulas)))))
     (t
      ;; Ought to give a better error message
      (assert (every #'(lambda (x) (or (symbolp x) (numberp x))) formula)
	      (formula)
	      "Proposition should contain only symbols (and numbers): ~s" formula)
      (set-difference (remove-duplicates (remove-if-not #'varp (cdr formula)))
		      bound-vars)))))
;; Testing
;; (unbound-vars '(foo ?x ?y))
;; (unbound-vars '(foo ?x ?y) '(?x ?z))
;; (unbound-vars '(:and (foo ?x ?y) (:forall (?z) (foo q ?z))))
;; (unbound-vars '(:and (foo ?x n) (:forall (?z) (foo ?y ?z))))
;; (unbound-vars '(:and (foo m n) (:forall (?z) (foo l ?z))))
