;; dpll.lisp -- run Davis-Putnam simplification on CNF formula
(in-package medic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Global variables
;;;

(defvar *dpll-true-actions* ()
  "actions found to be true by DPLL analysis")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Local variables
;;;

;; Debugging
(defvar *debug-dpll* nil)
(defvar *dpll-noisy* t
  "If non-nil, report DPLL results.")

(defvar *dpll-pruned* t
  "Non-nil if the current round of DPLL (either pure or unit literal
elimination) made some change.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Code
;;;

;; NOTE:  FOR CORRECTNESS, MUST RUN THE FIRST UNIT CLAUSES PRUNING BEFORE
;; THE FIRST PURE LITERALS PRUNING.

(defun dpll-one-iteration (clauses)
  (if *debug-dpll* (format t "~&Entered dpll-one-iteration~%"))
  (delete-pure-literals (delete-unit-clauses clauses)))

(defun dpll-fixpoint (clauses)
  (cond (*dpll-pruned*
	 (setq *dpll-pruned* nil)
	 (dpll-fixpoint (dpll-one-iteration clauses)))
	((some #'cnf-false-p clauses)
	 (list (cnf-false)))
	(t clauses)))

;; Poor name, because this doesn't do DPLL.
;; It fixpoints over pure and unit literal resolution (which is one step of DPLL).
(defun dpll (cnf-clauses)
  (assert (cnf-clauses-p cnf-clauses))
  (if *show-stages* (format t "~&Entering DPLL~%"))
  (setq *dpll-true-actions* '())
  (setq *dpll-pruned* t)
  (let* ((size-before (and *dpll-noisy* (formula-size (cons :and cnf-clauses))))
	 (result (dpll-fixpoint cnf-clauses))
	 (size-after (and *dpll-noisy* (formula-size (cons :and result)))))
    (if *dpll-noisy*
	(format t "DPLL: before: ~s, after: ~s~%" size-before size-after))
    ;; (setq foo result)
    (if *debug-dpll* (format t "~&Exited DPLL~%"))
    result))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;; Unit Clauses

;; given a list of clauses, return two values: the list of unit
;; clauses, and the list of non-unit-clauses
(declaim (ftype (function (list) (values list list)) separate-unit-clauses))
(defun separate-unit-clauses (cnf-clauses)
  (let ((unit-clauses '())
	(non-unit-clauses '()))
    (dolist (clause cnf-clauses)
      (cond ((sprop-p clause)
	     (push clause unit-clauses))
	    ((and (eq :or (car clause))
		  (length-two clause))
	     (push (second clause) unit-clauses))
	    (t
	     (push clause non-unit-clauses))))
    (values (delete-duplicates unit-clauses :test #'sprop-equal)
	    non-unit-clauses)))


;; returns a boolean indicating whether any two unit clauses are the
;; negation of each other
(defun unit-conflicts (units)
  (some #'(lambda (u)
	    (member (sprop-negate u) units :test #'sprop-equal))
	units))


(defun delete-unit-clauses (clauses)
  (if *debug-dpll* (format t "~&Entered delete-unit-clauses; "))
  (multiple-value-bind (units rest) (separate-unit-clauses clauses)
    (declare (list units) (list rest))
    (cond
     ((null units) rest)
     ((unit-conflicts units) '((:OR)))
     (t
      (progn
	(setq *dpll-pruned* t)
	(dolist (u units)
	  ;; used to test that time is odd, but this should be more general
	  (if (not (sprop-negated-p u))
	      (push u *dpll-true-actions*)))
	(if *debug-dpll* (format t "~d units; " (length units)))
	(let ((rest (delete-if #'(lambda (c)
				   (declare (list c))
				   (assert (and (listp c) (eq :or (car c))))
				   (let ((disjuncts (cdr c)))
				     (some #'(lambda (u) (member u disjuncts :test #'sprop-equal))
					   units)))
			       rest)))
	  ;; Old implementation seems unduly inefficient.
	  ;;     (setq rest (delete-if #'(lambda (c)
	  ;; 			      (intersection (cdr c) units :test #'sprop-equal))
	  ;; 			  rest))
	  (if *debug-dpll* (format t "clauses (pass 1); "))
	  (dolist (c rest)
	    (assert (eq :or (car c)))
	    (let ((result
		   (delete-if #'(lambda (lit)
				  (member (sprop-negate lit) units :test #'sprop-equal))
			      (cdr c))))
	      ;; Trick to side-effect the small list we're working on; we can't
	      ;; do (setf c result), but delete-if doesn't work *only* by side effect.
	      ;; No need to change the car, because we know it's :or.
	      ;; Could do the side effect unconditionally; any reason not to?
	      ;; (if (not (eq (cdr c) result))
	      ;;     (setf (cdr c) result))
	      (setf (cdr c) result)
	      ))
	  (if *debug-dpll* (format t "clauses (pass 2)~%"))
	  rest))))))


;; (delete-unit-clauses '((a) (b) (:or (c) (d))))
;; (delete-unit-clauses '((a) (b) (:or (a) (b) (c) (d))))
;; (delete-unit-clauses '((a) (b) (:or (:not (a)) (:not (b)) (c) (d))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;; Pure Literals


;; Given a set of clauses, returns sprops which represent pure literals.
(defun get-pure-lits (clauses)
  (let ((pure-lit-table (make-sprop-hash-table)))
    (dolist (clause clauses)
      (assert (eq :or (car clause)))
      (dolist (lit (cdr clause))
	(declare (type sprop lit))
	(assert (sprop-p lit))
	(setf (gethash lit pure-lit-table) t)))
    (let ((result '()))
      (maphash #'(lambda (key value)
		   (declare (ignore value))
		   (if (not (gethash (sprop-negate key) pure-lit-table))
		       (push key result)))
	       pure-lit-table)
      result)))

;; Old implementation puts only positive literals in hash table.
;; ;; given a set of clauses, returns entries from the *pure-lit-table* which
;; ;; represent pure literals
;; (defun get-pure-lits (clauses)
;;   (let ((*pure-lit-table* (make-sprop-hash-table)))
;;     (flet ((update-table (lit)
;; 	     (let* ((pos-lit (sprop-nonnegated lit))
;; 		    ;; perhaps just use the boolean, not bother with the letter
;; 		    (letter (if (sprop-negated-p lit) 'n 'p)))
;; 	       (setf (gethash pos-lit *pure-lit-table*)
;; 		 (adjoin letter (gethash pos-lit *pure-lit-table* ()))))))
;;       (dolist (clause clauses)
;; 	(assert (eq :or (car clause)))
;; 	(dolist (lit (cdr clause))
;; 	  (update-table lit)))
;;       (let ((result ()))
;; 	(maphash #'(lambda (key value)
;; 		     (if (length-one value)
;; 			 (push (cons key value) result)))
;; 		 *pure-lit-table*)
;; 	result))))


(defun delete-pure-literals (clauses)
  (declare (list clauses))
  (if *debug-dpll* (format t "~&Entered delete-pure-literals; "))
  (let ((pure-lits (get-pure-lits clauses)))
    (if *debug-dpll* (format t "found ~d pure literals~%" (length pure-lits)))
    (if (null pure-lits)
	clauses
      ;; hash table not empty: we found something
      (progn
	(setq *dpll-pruned* t)
	(dolist (lit pure-lits)
	  ;; used to test that time is odd, but this should be more general
	  (if (not (sprop-negated-p lit))
	      (push lit *dpll-true-actions*)))
	(delete-if #'(lambda (c)
		       (intersection (cdr c) pure-lits :test #'sprop-equal))
		   clauses)))))
;; (delete-pure-literals '((:or (a) (b)) (:or (:not (a)) (c)) (:or (b) (:not (c)))))


;; Do this only if you are very brave.
;; (trace  delete-pure-literals  get-pure-lits  delete-unit-clauses  separate-unit-clauses dpll-one-iteration dpll-fixpoint  dpll)
