;;; driver.lisp

(in-package medic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User parameters
;;;

;; If non-nil, don't call the solver, just generate .cnf file
(defvar *dont-solve* nil)

;; If non-nil, use slow Lisp implementation of DPLL.
;; Otherwise, don't simplify before writing initial .cnf file.
(defvar *run-dpll* nil)

;; Other possible values, if your path isn't correctly set:
;; SGI: "/projects/ai2/satplan/system/bin/walksat"
;; Solaris: "/projects/ai2/satplan/system/bin/solaris/walksat"
(defvar walksat-program "walksat")
(defvar walksath-program "walksath")
(defvar tableau-program "solve")     ; "solve" is a wrapper script for tableau
(defvar relsat-program "rel_sat")
(defvar compact-cnf-program "compact-cnf")
(defvar *ksh-path* "/usr/bin/ksh")

;; These needn't be surrounded by quotes.  Each is a *single* argument, though.
(defvar ubersolve-simp-options "-cp2 -s")
(defvar ubersolve-solve-options "walksat -solcnf")


(defvar *find-minimal-plan* nil
  "Non-nil if find-plan should return the shortest plan;
nil if it should return the first plan it finds.")

;; faster not to start at 1?
(defvar *start-steps-default* 4)

(defvar *cnf-solver* 'walksat-then-relsat
  "One of 'walksat, 'tableau, 'relsat, 'infer, 'walksat-then-tableau, 'walksat-then-relsat, or 'ubersolve")

;; Perhaps this should be part of *output-files*.
;; If non-nil, run compact-cnf on .cnf file produced by Lisp.
(defvar *external-simplifier* t)


(defparameter *output-files-all*
  '(CNF LITNUM CNF-SIZE DATA INFO PERCENT CLAUSES GOALS))

(defparameter *output-files-default*
  ;; goals is incompatible with simplification, so leave it off by default
  '(CNF LITNUM CNF-SIZE DATA INFO PERCENT CLAUSES))

(defvar *output-files*
  (copy-list *output-files-default*)
  "List of symbols representing possible output files.
Presence on this list means the file will be created.
See *output-files-default* for a reasonable initial value.

symbol  extension       description
INFO    .info           flags, solver, machine
PERCENT .percent        % of actions affecting each fluent; fluents per action
CLAUSES .clauses        number of action, frame, and exclusion clauses
CNF     .cnf            unsimplified CNF file

These are written only if CNF is:
LITNUM  .litnum         mapping between symbolic literals and CNF file numbers
CNF-SIZE .cnf-size      number of clauses, variables, and literals; encode time
DATA    .data           size of the cnf formula, solution time
INFO    .info           flags, solver, machine

PERCENT .percent        % of actions affecting each fluent; fluents per action
CLAUSES .clauses        number of action, frame, and exclusion clauses
GOALS   .goals          goal clauses (from *unsimplified* CNF)

Other files not controlled by this variable:

-simple.cnf	CNF file simplified by external simplifier (compact)
.transl		mapping between unsimplified and simplified CNF variables
-ws.data	info on average time for multiple runs of Walksat
...output files from solvers and simplifiers")


(defvar solve-problem-start-function nil
  "Function to be called at the beginning of solve-problem.
It is called before any computation occurs.
The arguments to solve-problem (prob, max-steps, start-steps, and experimentnum)
are passed to the function.")

;; Perhaps should supply the function more -- such as the plan, etc.
(defvar solve-problem-end-function nil
  "Function to be called at the end of solve-problem.
It is called after all computation completes.
The arguments to solve-problem (prob, max-steps, start-steps, and experimentnum)
are passed to the function.")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Global variables
;;;

;; the subset of :requirements which the Medic planner can handle
(defvar *medic-capabilities* '(:cwa :QUANTIFIED-EFFECTS :QUANTIFIED-PRECONDITIONS  :strips :equality :disjunctive-preconditions :conditional-effects :typing))

;;for timing.  To keep from having to pass around many parameters
(defvar *start-encode-time* 0)
(defvar *end-encode-time* 0)
(defvar *start-decode-time* 0)
(defvar *end-decode-time* 0)

;; the file produced by the external simplifier
(defvar *translation-file* nil)

(defvar *prob-name* 'no-problem
  "The current problem name, as a symbol.")

;; a list of all the information returned from the output of the solver
;; (eg, walksat, relsat, or tableau):
;;   success-flag num-clauses num-vars num-lits elapsed-time mean-flips mean-restarts
;; in that order.  Mean-restarts and mean-flips are 0 if tableau was used.
(defvar *file-info*)


;; Not for debugging; used by frames.
(defvar *domain*)


;; (defvar *machine-info*
;;   (format nil "~a; ~a; ~a; ~a"
;; 	  (machine-type) (machine-version) (machine-instance)
;; 	  (let ((istr (run-program "hostname" :wait nil :output :stream)))
;; 	    (read-line istr))))
;; If defconstant, it gets evaluated every time this compiles, which takes time
;;   and is wrong to boot.
;; Can also get amount of memory from (among other places) the 'top' program.
(defvar *machine-info* (software-version))


(defvar *goal-cnf*)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Debugging variables
;;;

;; For debugging {relsat,tableau,walksat}-solve-cnf
(defvar *vtable*)
(defvar *truth-assignments*)
(defvar *problem*)
(defun dops () (domain-operators *domain*))
(defun pops () (dolist (o (dops)) (print-operator-long o)) (values))
;; Not ideal, because we'd like to use the problem's length rather than 30 and 1
(defun sfp (probname) (solve-find-problem probname 30 1))

(defvar *simplified-file-info*)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Flags
;;;

;;; Relevant global variables
; *frame-strategy*: 'classical, 'explanatory
; *action-representation*: 'regular, 'simple-split, 'simple-split-factored,
;    'overloaded-split, 'overloaded-split-factored, 'bitwise
; *exclusion-strategy*: 'sequential-pairwise, 'parallel
; *types-strategy*: 'no-types, 'time-only, 'eliminate
;; These influence meta-behavior
; *find-minimal-plan*: nil, t
; *cnf-solver*: 'walksat, 'tableau, 'relsat, 'infer, 'ubersolve,
;    'walksat-then-tableau, 'walksat-then-relsat

;; I believe these are in the order judged most likely to succeed.  Right?
(defvar *flags-list-maximal-types*
  '("erpe" "esse" "ecse" "eose" "efst" "ebse"   ;; "erse" is dominated by "erpe"
    "crse" "csse" "ccse" "cose" "cfst" "cbse")
  "List of all flags that aren't dominated by a strictly better encoding.")

(defun flags ()
  (declare (type frame-strategy *frame-strategy*)
	   (type action-representation *action-representation*)
	   (type exclusion-strategy *exclusion-strategy*)
	   (type types-strategy *types-strategy*))
  (assert (member *frame-strategy* *frame-strategies*))
  (assert (member *action-representation* *action-representations*))
  (assert (member *exclusion-strategy* *exclusion-strategies*))
  (assert (member *types-strategy* *types-strategies*))
  (format nil "~a: frame=~a act-rep=~a exclude=~a type=~a minimal=~a"
	  (flags-brief)
	  *frame-strategy* *action-representation* *exclusion-strategy*
	  *types-strategy* *find-minimal-plan*))

(defun brief-flags ()
  (format nil "~a~a~a~a"
	  (case *frame-strategy*
	    ((classical) "c")
	    ((explanatory) "e")
	    (otherwise (error "bad value for *frame-strategy* ~a"
			      *frame-strategy*)))
	  (case *action-representation*
	    ((regular) "r")
	    ((simple-split) "s")
	    ((simple-split-factored) "c")
	    ((overloaded-split) "o")
	    ((overloaded-split-factored) "f")
	    ((bitwise) "b")
	    (otherwise (error "bad value for *action-representation* ~a"
			      *action-representation*)))
	  (case *exclusion-strategy*
	    ((sequential-pairwise) "s")
	    ((parallel) "p")
	    (otherwise (error "bad value for *exclusion-strategy* ~a"
			      *exclusion-strategy*)))
	  (case *types-strategy*
	    ((no-types) "n")
	    ((time-only) "t")
	    ((eliminate) "e")
	    (otherwise (error "bad value for *types-strategy* ~a"
			      *types-strategy*)))))

;; I type this name all the time, so we might as well define it.
(defun flags-brief ()
  (brief-flags))


;; Errs if flags are not a legal combination.
;; If optional argument NO-ERR is non-nil, instead returns a boolean.
(defun check-flags (&optional no-err)
  (cond
   ((and (not (eq *action-representation* 'regular))
	 (not (eq *exclusion-strategy* 'sequential-pairwise)))
    (if no-err
	nil
      (error "incompatible *action-representation* ~a, *exclusion-strategy* ~a"
	     *action-representation* *exclusion-strategy*)))
   ((and (eq *exclusion-strategy* 'parallel)
	 (eq *frame-strategy* 'classical))
    (if no-err
	nil
      (error "incompatible *exclusion-strategy* ~a, *frame-strategy* ~a"
	     *exclusion-strategy* *frame-strategy*)))
   ((and (factored-splitting-p)
	 (not (eq *exclusion-strategy* 'sequential-pairwise)))
    (if no-err
	nil
      (error "incompatible *action-representation* ~a, *exclusion-strategy* ~a"
	     *action-representation* *exclusion-strategy*)))
   ((and (eq *action-representation* 'overloaded-split-factored)
	 (not (member *types-strategy* '(no-types time-only))))
    (if no-err
	nil
      (error "incompatible *action-representation* ~a, *types-strategy* ~a"
	     *action-representation* *types-strategy*)))
   ;; otherwise, return t
   (t
    t)))


;; BRIEF is a string like the output of flags-brief.
;; Each position may optionally be ., which makes no change to that flag.
;; Intended for human, not programmatic, use.
(defun set-flags (brief)
  (declare (type simple-base-string brief))
  (if (not (= 4 (length brief)))
      (error "Bad length for brief, should be 4: ~s" brief))
  (case (aref brief 0)
    (#\c (setq *frame-strategy* 'classical))
    (#\e (setq *frame-strategy* 'explanatory))
    (#\. nil)
    (t (error "bad value for *frame-strategy*, should be in {ce}: ~a" (aref brief 0))))
  (case (aref brief 1)
    (#\r (setq *action-representation* 'regular))
    (#\s (setq *action-representation* 'simple-split))
    (#\c (setq *action-representation* 'simple-split-factored))
    (#\o (setq *action-representation* 'overloaded-split))
    (#\f (setq *action-representation* 'overloaded-split-factored))
    (#\b (setq *action-representation* 'bitwise))
    (#\. nil)
    (t (error "bad value for *action-representation* should be in {rscofb}: ~a" (aref brief 1))))
  (case (aref brief 2)
    (#\s (setq *exclusion-strategy* 'sequential-pairwise))
    (#\p (setq *exclusion-strategy* 'parallel))
    (#\. nil)
    (t (error "bad value for *exclusion-strategy*, should be in {sbp}: ~a" (aref brief 2))))
  (case (aref brief 3)
    (#\n (setq *types-strategy* 'no-types))
    (#\t (setq *types-strategy* 'time-only))
    (#\e (setq *types-strategy* 'eliminate))
    (#\. nil)
    (t (error "bad value for *types-strategy*, should be in {nte}: ~a" (aref brief 3))))
  (check-flags)
  ;; flags is executed for value, not side effect
  (flags))

(defun set-flags-default ()
  (set-flags "esen"))
(defun reset-flags ()
  (set-flags-default))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Solve problem file
;;;

;; Given a problemname (and an optional filename), find the problem, and run
;; it on all combinations of the given sets of flags:
;; FRAMES, ACT-REPS, EXCLUSIONS, and TYPES are lists of strategies.
;; Writes two output files.
(defun multi-solve-problem (problemname
			    &optional (problemfile nil)
				      max-steps
				      start-steps
				      (experimentnum 2)
				      (frames *frame-strategies*)
				      (act-reps *action-representations*)
				      (exclusions *exclusion-strategies*)
				      (types (list 'eliminate))
				      ;; (types *types-strategies*)
				      )
  (declare (symbol problemname))
  (declare (type (or fixnum (member nil)) max-steps))
  (declare (fixnum experimentnum))
  (declare (type (or fixnum (member nil)) start-steps))
  (declare (cons frames))
  (declare (cons act-reps))
  (declare (cons exclusions))
  (declare (cons types))
  (assert (symbolp problemname))
  (assert (or (not max-steps) (integerp max-steps)))
  (assert (integerp experimentnum))
  (assert (or (not start-steps) (integerp start-steps)))
  (assert (every #'consp (list exclusions frames act-reps types)))
  (let ((probname-string (format nil "~a" problemname)))
    ;; This relies on dynamic binding of the global *xxx-strategy* variables
    (dolist (*frame-strategy* frames)
      (dolist (*action-representation* act-reps)
	(dolist (*exclusion-strategy* exclusions)
	  (dolist (*types-strategy* types)
	    ;; Perhaps this ought to be configurable.
	    (let ((new-max (or max-steps (get-plan-length problemname)))
		  (new-start (or start-steps (get-plan-length problemname))))
	      (format t "~&~a~%" (flags-brief))
	     ;; (format t "~%new-max is ~a , new-start is " new-max new-start)
	      (if (probe-file (info-file-name experimentnum probname-string))
		  (format t "~&*** SKIPPING ~s~%    because ~s exists.~%"
			  (flags)
			  (info-file-name experimentnum probname-string))
		(if (check-flags t)
		    (let ((result-plan (solve-find-problem problemname
							   new-max new-start
							   problemfile
							   experimentnum)))
		      (cond (*dont-solve* '())
			    ((not result-plan) (error "No plan found."))
			    (t
			     (format t "~&~a: ~a~%" (flags-brief) result-plan))))))
	      (incf experimentnum)
	      )))))))

;; same as the above, but instead of working with the various global
;; variables, it works from a list of brief flags, i.e. ("csne" ...)
(defun multi-flags-solve-problem (problemname
				  &optional problemfile
					    max-steps
					    start-steps
					    (experimentnum 2)
					    (flag-lst *flags-list-maximal-types*))
  (declare (symbol problemname)
	   (type (or fixnum (member nil)) max-steps)
	   (fixnum experimentnum)
	   (type (or fixnum (member nil)) start-steps))
  (assert (symbolp problemname))
  (assert (or (not max-steps) (integerp max-steps)))
  (assert (integerp experimentnum))
  (assert (or (not start-steps) (integerp start-steps)))
  (let ((probname-string (format nil "~a" problemname)))
    (dolist (curflags flag-lst)
      (set-flags curflags)
      (let ((new-max (or max-steps (get-plan-length problemname)))
	    (new-start (or start-steps (get-plan-length problemname))))
	;; new-max can be less than new-start.  Is this is a problem?
	(format t "~&~a~%" (flags-brief))
	(if (probe-file (size-stat-file-name experimentnum probname-string))
	    (format t "~&*** SKIPPING ~s~%    because ~s exists.~%"
		    (flags)
		    (size-stat-file-name experimentnum probname-string))
	  (if (check-flags t)
	      (let ((result-plan (solve-find-problem problemname
						     new-max new-start
						     problemfile
						     experimentnum)))
		(cond (*dont-solve* '())
		      (*run-program-timeout* '())
		      ((not result-plan) (error "No plan found."))
		      (t
		       (format t "~&~a: ~a~%" (flags-brief) result-plan))))
	      (incf experimentnum)))))))


;; perhaps add independent versions for only fluent and action strategies.
(defun multi-solve-frames (probname problemfile max-steps experimentnum &optional start-steps (frames *frame-strategies*))
  (multi-solve-problem probname
		       problemfile max-steps start-steps experimentnum
   frames
   (list *action-representations*)
   (list *exclusion-strategy*)
   (list *types-strategy*)))

(defun multi-solve-act-reps (probname problemfile max-steps experimentnum &optional start-steps (act-reps *action-representations*))
  (multi-solve-problem probname problemfile max-steps start-steps experimentnum
   (list *frame-strategy*)
   act-reps
   (list *exclusion-strategy*)
   (list *types-strategy*)))

(defun multi-solve-types (probname problemfile max-steps experimentnum &optional start-steps (types *types-strategies*))
  (multi-solve-problem probname problemfile max-steps start-steps experimentnum
   (list *frame-strategy*)
   (list *action-representations*)
   (list *exclusion-strategy*)
   types))

(defun multi-solve-excl (probname problemfile max-steps experimentnum &optional start-steps (excl *exclusion-strategy*))
  (multi-solve-problem probname problemfile max-steps start-steps experimentnum
   (list *frame-strategy*)
   (list *action-representations*)
   excl
   (list *types-strategy*)))



(defun file-name-header (problemname &optional max-steps)
  (if max-steps
      (format nil "~a-~a-~d" problemname (brief-flags) max-steps)
    (format nil "~a-~a" problemname (brief-flags))))

(defun data-file-name (experimentnum problemname)
  (declare (ignore experimentnum))
  (and (member 'data *output-files*)
       (format nil "~a.data" (file-name-header problemname))))

(defun info-file-name (experimentnum problemname)
  (declare (ignore experimentnum))
  (and (member 'info *output-files*)
       (format nil "~a.info" (file-name-header problemname))))

(defun size-stat-file-name (experimentnum problemname)
  (declare (ignore experimentnum))
  (format nil "~a.cnf-size" (file-name-header problemname)))

;; PROB is a problem
;; MAX-STEPS is the longest plan to consider
;; Return a plan, or nil if none no longer than MAX-STEPS exists.
;; No display of the plan is performed.
(defun solve-problem (prob  &optional  max-steps start-steps experimentnum)
  (if (< max-steps start-steps)
      (error "Max-steps ~a is less than start-steps ~a" max-steps start-steps))
  (if solve-problem-start-function
      (funcall solve-problem-start-function
	       prob max-steps start-steps experimentnum))
  (if (symbolp prob)
      (setq prob (or (get-problem prob)
		     (error "Problem ~a not found" prob))))

  (setq prob (copy-problem prob))
  (setq *problem* prob
	*domain* (get-domain-copy (problem-domain prob)))
  ;; Check for length information if none is provided
  (if (or (not max-steps) (not start-steps))
      (let ((length-key (cond ((eq *exclusion-strategy* 'PARALLEL)
			       :parallel)
			      ((eq *exclusion-strategy* 'sequential-pairwise)
			       :serial)
			      (t
			       (error "Can't determine key to look up length")))))
	(setq max-steps (or max-steps
			    (nth 1 (assoc length-key (problem-length prob))))
	      start-steps (or start-steps max-steps 0))
	;; If there was no length info in the problem file and
	;; none was given as an argument, throw an error
	(if (not max-steps)
	    (error "~%No ~s length information available for problem ~s" length-key prob))))

  ;; Convert any numbers into symbols.
  ;; in operators:  (don't need this yet; implement it when I do)
  ;; (dolist (o (domain-operators dom))
  ;;    ...)
  ;; in problem:
  (setf (problem-init prob) (number->symbol (problem-init prob)))
  (setf (problem-goal prob) (number->symbol (problem-goal prob)))


  (let* ((probname (problem-name prob)))
    (setq *prob-name* probname)
    (check-requirements *medic-capabilities* *problem* *domain*)
    (if (not experimentnum)
	(progn
	  (setq experimentnum (+ 10 (random 90)))
	  ;; ~& because def-domain may output text but no newline
	  (format t "~&Selecting experiment number ~s for ~a~%"
		  experimentnum probname)))
    (check-flags)
    (if (member 'info *output-files*)
	(write-info-file
	 experimentnum probname
	 (flags) (list (list '*cnf-solver* *cnf-solver*))
	 (flags) *machine-info* ""))
    (setq *start-encode-time* (get-internal-run-time))

    ;;open up the data file so that we can write lines out to it later
    (let ((datafile (data-file-name experimentnum probname )))
      (with-open-file-maybe (datafile-stream datafile
					     :direction :output :if-does-not-exist :create :if-exists :rename-and-delete)
	(let (
	      ;; clauses and substitutions
	      init-clauses goal-clauses
	      op-clauses ;; op-subs
	      frame-clauses frame-subs
	      act-rep-clauses act-rep-subs
	      excl-clauses excl-subs
	      all-substitutions
	      onestep-clauses
	      ;; init-act-rep-clauses onestep-act-rep-clauses goal-act-rep-clauses
	      init-cnf onestep-cnf goal-cnf)
	  (declare (list excl-clauses) (list goal-clauses)
		   (list op-clauses)
		   (list frame-clauses) (list frame-subs)
		   (list act-rep-clauses) (list act-rep-subs)
		   (list excl-clauses) (list excl-subs)
		   (list all-substitutions)
		   (list onestep-clauses)
		   (list init-cnf) (list onestep-cnf) (list goal-cnf))

	  (clrhash *sprop-hash-table*)
	  (clrhash *sprop-args-hash-table*)
	  ;; init-clauses must precede everything else, as it sets many globals

	  ;; Wrong, since initial encoding should include time to read the file.
	  ;; (setq *start-encode-time* (get-internal-run-time))

	  (multiple-value-setq (init-clauses op-clauses goal-clauses)
	    (init-action-and-goal-clauses *problem* *domain*))
	  ;; Action representation precedes exclusion, which depends on both
	  ;;    *action-representation* and on globals set by (act-rep-clauses).
	  (multiple-value-setq (act-rep-clauses act-rep-subs) (act-rep-clauses))
	  (multiple-value-setq (frame-clauses frame-subs) (frame-clauses))
	  (multiple-value-setq (excl-clauses excl-subs) (exclusion-clauses))

	  (assert (substitutions-p frame-subs))
	  (assert (substitutions-p act-rep-subs))
	  (assert (substitutions-p excl-subs))

	  (assert (cnf-clauses-p init-clauses))
	  (assert (cnf-clauses-p op-clauses))
	  (assert (cnf-clauses-p goal-clauses))
	  (assert (cnf-clauses-p frame-clauses))
	  (assert (cnf-clauses-p act-rep-clauses))
	  (assert (cnf-clauses-p excl-clauses))

	  (assert (not (contains-true-or-false-sprop-p init-clauses)))
	  (assert (not (contains-true-or-false-sprop-p op-clauses)))
	  (assert (not (contains-true-or-false-sprop-p goal-clauses)))
	  (assert (not (contains-true-or-false-sprop-p frame-clauses)))
	  (assert (not (contains-true-or-false-sprop-p act-rep-clauses)))
	  (assert (not (contains-true-or-false-sprop-p excl-clauses)))

	  (if *show-stages* (format t "~&Entering prop-substitute~%"))
	  ;; When does frame-clauses do substitutions?
	  (setq all-substitutions (append ;; op-subs
				   frame-subs act-rep-subs excl-subs))
	  (assert (substitutions-p all-substitutions))
	  ;; (format t "~&all-sub-first: ~S~%" (mapcar #'first all-substitutions))
	  (assert (not (contains-duplicate (mapcar #'first all-substitutions) :test #'sprop-equal)))

	  (setq op-clauses (prop-substitute op-clauses all-substitutions))
	  (setq frame-clauses (prop-substitute frame-clauses all-substitutions))
	  (setq act-rep-clauses (prop-substitute act-rep-clauses all-substitutions))
	  (setq excl-clauses (prop-substitute excl-clauses all-substitutions))

	  (assert (notany #'clause-contains-sprop-and-negation init-clauses))
	  (assert (notany #'clause-contains-sprop-and-negation op-clauses))
	  (assert (notany #'clause-contains-sprop-and-negation frame-clauses))
	  (assert (notany #'clause-contains-sprop-and-negation act-rep-clauses))
	  (assert (notany #'clause-contains-sprop-and-negation excl-clauses))
	  (assert (notany #'clause-contains-sprop-and-negation goal-clauses))

	  (setq onestep-clauses
	    (append op-clauses frame-clauses act-rep-clauses excl-clauses))

	  (setq init-clauses (prop-substitute init-clauses all-substitutions))
	  (setq goal-clauses (prop-substitute goal-clauses all-substitutions))
	  (if *show-stages* (format t "~&Exited prop-substitute~%"))

	  (if (member 'clauses *output-files*)
	      (write-problem-clause-breakdown-file
	       probname (length op-clauses)
	       (length frame-clauses) (length excl-clauses)))

	  (assert (cnf-clauses-p init-clauses))
	  (assert (cnf-clauses-p onestep-clauses))
	  (assert (cnf-clauses-p goal-clauses))

	  ;; used to call eliminate-implies and convert-to-cnf here.
	  (assert (notany #'clause-contains-duplicate-sprop init-clauses))
	  (assert (notany #'clause-contains-duplicate-sprop onestep-clauses))
	  (assert (notany #'clause-contains-duplicate-sprop goal-clauses))

	  (setq init-cnf (cons :and init-clauses))
	  (setq onestep-cnf (cons :and onestep-clauses))
	  (setq goal-cnf (cons :and goal-clauses))

	  (multiple-value-bind (plan plan-steps)
	      (find-plan init-cnf onestep-cnf goal-cnf max-steps probname
			 all-substitutions experimentnum
			 start-steps datafile-stream)
	    (setq plan (symbol->number plan))
	    ;; Since it's outside find-plan, this isn't included in our runtime
	    ;; statistics (which is the right thing).
	    (if plan-steps
		;; Verify the plan (no plan if plan-steps = nil)
		(progn
		  (if *show-stages*
		      (format t "Verifying result~%"))
		  (cond ((eq 'parallel *exclusion-strategy*)
			 ;; it's a parallel plan
			 (multiple-value-bind (successp failing-plan)
			     (verify-clean-parallel probname plan)
			   (if (not successp)
			       (let ((*verify-verbosely* t))
				 (verify-clean probname failing-plan)
				 (error "Linearization ~s~% of plan ~s~% failed to verify for ~s" failing-plan plan probname)))))
			((and plan (listp (first (first plan))))
			 ;; it's an alternate (not linear) plan
			 (multiple-value-bind (successp failing-plan)
			     (verify-clean-alternates probname plan)
			   (if (not successp)
			       (let ((*verify-verbosely* t))
				 (verify-clean probname failing-plan)
				 (error "Linearization ~s~% of plan ~s~% failed to verify for ~s" failing-plan plan probname)))))
			(t
			 ;; it's a linear plan
			 (if (not (verify-clean probname plan))
			       (let ((*verify-verbosely* t))
				 (verify-clean probname plan)
				 (error "Plan ~s~% failed to verify for ~s" plan probname)))))))
	    (if solve-problem-end-function
		(funcall solve-problem-end-function
			 prob max-steps start-steps experimentnum))
	    plan))))))

;; (Timing results put all decoding time on the last plan tested, which
;; gives somewhat strange results if *find-minimal-plan* is non-nil.)
;; (Actually, if *find-minimal-plan* is non-nil, the actual decoding may be
;; completely left out.)

;; Return a minimal-length plan satisfying the time-neutral formulas,
;; or nil if none no longer than max-steps exists.
(defun find-plan (init-cnf onestep-cnf goal-cnf max-steps probname
			   substitutions experimentnum
			   start-steps datafile-stream)
  (declare (list init-cnf) (list onestep-cnf) (list goal-cnf)
	   (fixnum max-steps) (symbol probname)
	   (list substitutions) (fixnum experimentnum)
	   (fixnum start-steps) (type (or stream (member nil)) datafile-stream))
  (assert (or (not datafile-stream) (streamp datafile-stream)))
  (if *show-stages* (format t "~&Entering find-plan~%"))
  ;; This do loop performs a binary search to find the shortest plan (or
  ;; any plan, if *find-minimal-plan* is nil).
  (do ((try-steps start-steps ;; was (or start-steps (min *start-steps-default* max-steps)) -- and should be again?
		  (if success-steps
		      (round (/ (+ failure-steps success-steps) 2))
		    (min max-steps (if (zerop try-steps) 1 (* try-steps 2)))))
       (failure-steps -1)
       (success-steps nil)
       (success-plan nil))
      ;; (end-test . result)
      ;; This end test seemed to be giving problems before (not stopping?)
      ((or (>= failure-steps max-steps)
	   (and success-plan (= failure-steps (1- success-steps)))
	   (and success-plan (not *find-minimal-plan*)))
       ;; Return value (in implicit progn)
       (values
	(and success-plan
	     ;; For efficiency, these substitutions could be done after
	     ;; removing even-time propositions.
	     ;; Why wasn't the "nonnegated" necessary before?
	     (let* ((nonnegated-plan (remove-if #'sprop-negated-p success-plan))
		    (odd-plan (plan-odd-time-only nonnegated-plan))
		    (unsplit-plan (prop-unsubstitute odd-plan *split-table*
						     success-steps))
		    (unsubstituted-plan (prop-unsubstitute unsplit-plan
							   substitutions
							   success-steps))
		    (unskolemized-plan (unskolemize-plan unsubstituted-plan)))
	       (multiple-value-bind (pruned-plan num-actions)
		   (plan-prune unskolemized-plan success-steps)
		 (if datafile-stream
		     (write-data-line-non-ucpop
		      experimentnum *prob-name* try-steps
		      num-actions datafile-stream))
		 pruned-plan)))
	success-steps))
    ;; The call to instantiate-init-step-and-goal takes a long time;
    ;; it may not be necessary, if we don't manipulate the final CNF.
    ;; We could instead just write directly to the .cnf file.
    (let* ((full-cnf (instantiate-init-step-and-goal
		      init-cnf onestep-cnf goal-cnf try-steps))
	   ;; If simplified-full-cnf is empty, we have a plan already.
	   (simplified-full-cnf (cons :and (if *run-dpll*
					       (dpll (cnf-clauses full-cnf))
					     (cnf-clauses full-cnf))))
	   (new-plan
	    (progn
	      (if (member 'cnf-size *output-files*)
		  (write-prob-size-stat-file
		   ;; why is this max-steps instead of try-steps??
		   (size-stat-file-name experimentnum probname)
		   simplified-full-cnf))
	      (if (member 'goals *output-files*)
		  (setq *goal-cnf* (sprop-goal-instantiate goal-cnf try-steps)))
	      (setq *file-info* nil)
	      (cond (*dont-solve*
		     (let ((basename (file-name-header probname try-steps)))
		       (write-cnf-file simplified-full-cnf basename)
		       (assert (or (not (member 'cnf *output-files*)) (probe-file  (format nil "~a.cnf" basename))))
		       )
		     ;; Return the "plan"
		     nil)
		    ((cnf-false-p simplified-full-cnf)
		     (setq *file-info* (list nil 0
					     (length (cdr simplified-full-cnf))
					     0 0 0.0 0.0))
		     nil)
		    ((cnf-true-p simplified-full-cnf)
		     ;; length might be 0 or 1.
		     (setq *file-info* (list t 0
					     (length (cdr simplified-full-cnf))
					     0 0 0.0 0.0))
		     ;; Don't return nil, which indicates failure
		     'dpll-solved-plan)
		    (*external-simplifier*
		     (let ((solution
			    (solve-cnf-external-simplifier
			     ;; it isn't really simplified if we got here
			     simplified-full-cnf
			     (file-name-header probname try-steps))))
		       (setq *simplified-file-info* *file-info*)
		       (let ((cnf-size (formula-size simplified-full-cnf))
			     (elapsed-time (fifth *file-info*)))
			 (setq *file-info* (list (if solution t nil)
						 (first cnf-size)
						 (second cnf-size)
						 (third cnf-size)
						 elapsed-time 0.0 0.0)))
		       solution))
		    (t
		     (solve-cnf
		      simplified-full-cnf
		      (file-name-header probname try-steps)))))))
      (if *dont-solve*
	  (setq failure-steps max-steps) ; end the loop
	(progn
	  (let ((cnf-size (formula-size simplified-full-cnf)))
	    (assert (= 7 (length *file-info*)))
	    (assert (eq (first cnf-size) (second *file-info*))) ; variables
	    (assert (or (eq *cnf-solver* 'tableau) ; tableau doesn't report unit clauses
			(eq (second cnf-size) (third *file-info*)))) ; clauses
	    ;; literals; tableau doesn't report this, only numhorn
	    (if (not (fourth *file-info*))
		(setf (fourth *file-info*) (third cnf-size)))
	    (assert (eq (third cnf-size) (fourth *file-info*))) ; literals
	    )
	  ;; (format t "~&new-plan ~s~%" new-plan)
	  (if new-plan
	      (progn
		(if (eq new-plan 'dpll-solved-plan)
		    (setq new-plan '()))
		(assert (or (not success-steps) (< try-steps success-steps)))
		;; (format t "~&success-plan = ~s~%" (append *dpll-true-actions* new-plan))
		(setq success-plan (append *dpll-true-actions* new-plan)
		      success-steps try-steps))
	    (progn
	      ;; no valid plan, so write out plan time steps as 0
	      (if datafile-stream
		  (write-data-line-non-ucpop
		   experimentnum *prob-name* try-steps 0 datafile-stream))
	      (assert (> try-steps failure-steps))
	      (setq failure-steps try-steps)
	      ;; For the next iteration.
	      (setq *start-encode-time* (get-internal-run-time)))))))))

(defun plan-odd-time-only (plan)
  (remove-if-not #'(lambda (step)
		     (declare (type sprop step))
		     (assert (sprop-p step))
		     (let ((time (sprop-time step)))
		       (and (numberp time)
			    (oddp time))))
		 plan))

;; Remove everything except actions, and sort 'em.
;; Returns two values: the plan and the number of actions in it.
;; The plan may be linear (list of actions) or parallel (list of lists of
;; actions).
(defun plan-prune (plan steps)
  ;; This isn't right, since decode time must include time to read the file.
  ;; (setq *start-decode-time* (get-internal-run-time))

  (let* ((plan-copy (copy-list plan)) ;; clarify debugger argument reports
	 (pruned
	  ;; already pruned non-odd times
	  ;; (remove-if-not #'(lambda (step)
	  ;; 			    (and (listp step)
	  ;; 				 (let ((time (sprop-time step)))
	  ;; 				   (and (numberp time)
	  ;; 					(oddp time)))))
	  ;; 			plan-copy)
	  (cond
	   ((and (eq *action-representation* 'simple-split-factored)
		 (eq *partial-action-method* 'extra-vars))
	    (delete-if #'(lambda (prop)
			   (eq 'action (sprop-name prop)))
		       plan-copy))
	   ((and (eq *action-representation* 'overloaded-split-factored)
		 (eq *partial-action-method* 'extra-vars))
	    (delete-if #'(lambda (prop)
			   (eq 'overloaded (sprop-name prop)))
		       plan-copy))
	   (t plan-copy)))
	 ;; This call to sort is destructive (pruned won't be sensible after)
	 (sorted (sort pruned #'sprop-time-lessp)))
    ;; (format t "~&sorted: ~s~%" sorted)

    ;; destructively change gensymed operators back to their original names
    (setq sorted (mapcar #'unclone-sprop sorted))

    ;; Do final checking and return result (which does not include time).
    (let ((sans-negation (delete-if #'sprop-negated-p sorted)))
      ;; No-ops are removed later (at the end of this function); for instance,
      ;; they might be in alternation with some other action.
      (cond ((eq 'parallel *exclusion-strategy*)
	     ;; Rather than returning a list of actions, return a list of lists
	     ;; of actions.  In each list of actions, the actions can occur in
	     ;; any order but must occur before those in subsequent lists.
	     (let* ((pplan (linear-plan->parallel-plan sans-negation))
		    (as-lists (mapcar #'(lambda (sublist)
					  (mapcar #'sprop->list-no-time sublist))
				      pplan)))
	       (assert (notany #'(lambda (sublist) (member *no-op-action* sublist :test #'equal)) as-lists))
	       (values as-lists (apply #'+ (mapcar #'length as-lists)))))
	    (t
	     (let ((has-dup (contains-duplicate (mapcar #'sprop-time sans-negation))))
	       (if has-dup
		   (case *frame-strategy*
			 (explanatory
			  (error "Multiple actions for single time step in plan ~s" sans-negation))
			 (classical
			  ;; This is permitted if the multiple actions have
			  ;; identical effects -- we're free to choose any
			  ;; one of them.
			  (let* ((aplan (linear-plan->parallel-plan sans-negation 'retain-noops))
				 (as-lists (mapcar #'(lambda (sublist)
						       (mapcar #'sprop->list-no-time sublist))
						   aplan)))
			    (values as-lists (apply #'+ (mapcar #'length as-lists)))))
			 (otherwise
			  (error "Bad value for *frame-strategy* ~s"
				 *frame-strategy*)))
		 (progn
		   ;; ensure right number of time steps
		   (case *frame-strategy*
			 (explanatory
			  (if (and *find-minimal-plan*
				   (< steps (length sorted)))
			      (error "Plan should have length no more than ~s: ~s" steps sorted)))
			 (classical
			  (if (and *find-minimal-plan*
				   (not (= steps (length sorted))))
			      (error "Plan should have length ~s: ~s" steps sorted)))
			 (otherwise
			  (error "Bad value for *frame-strategy* ~s"
				 *frame-strategy*)))
		   (let* ((as-lists (mapcar #'sprop->list-no-time sans-negation))
			  (sans-noops (delete *no-op-action* as-lists :test #'equal)))
		     (values sans-noops (length sans-noops)))))))))))

;; Takes a sorted list of actions.
;; Returns two values: the parallel plan (a list of lists of actions) and the
;; number of actions in it.
;; Set retain-no-ops when this is being used to create an alternate, not
;; parallel, plan; in an alternate plan, one elt is chosen at each time step.
(defun linear-plan->parallel-plan (plan &optional retain-alternate-no-ops)
  (declare (list plan))
  (let ((current-time 1)
	(current-time-actions '())
	(total-actions (length plan))
	(result '()))
    (declare (fixnum current-time) (list current-time-actions)
	     (fixnum total-actions) (list result))
    (dolist (action plan)
      (if (not (sprop-negated-p action))
	  (let ((action-time (sprop-time action))
		(action-timeless (sprop-remove-time action)))
	    (cond ((= current-time action-time)
		   (push action-timeless current-time-actions))
		  (t			; (= action-time (+ 2 current-time))
		   ;; If this timestep contains only a no-op, ignore it.
		   (if (not (equal (list *no-op-sprop*) current-time-actions))
		       (push current-time-actions result))
		   (setq current-time action-time)
		   (setq current-time-actions (list action-timeless)))
		  ;; (t
		  ;;  (error "Bad time ~a; apparently no actions at time ~a"
		  ;; 	  action-time (+ 2 current-time)))
		  ))))
    (push current-time-actions result)
    ;; result contains no times or negation
    (values (nreverse (delete nil
			      (mapcar #'(lambda (actionlist)
					  (if retain-alternate-no-ops
					      actionlist
					    (delete *no-op-sprop* actionlist :test #'sprop-equal)))
				      result)))
	    total-actions)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Solve CNF formula
;;;

;; Merely calls {relsat,tableau,walksat}-solve-cnf and returns its result.
;; Returns nil (if no assignment) or a list of literal truth assignments,
;;    each of which is a [positive] literal possibly wrapped in (NOT ...).
;; Perhaps eliminate optional argument num-clauses.
(defun solve-cnf (cnf-formula file-basename &optional num-clauses vtable)
  ;; This is wrong, since encoding time must include time to write the file.
  ;; (setq *end-encode-time* (get-internal-run-time))
  (if (not file-basename)
      ;; (setq file-basename "input")
      (error "solve-cnf needs a file name"))

  (assert (or (not *external-simplifier*) (probe-file *translation-file*)))

  (setq *vtable* vtable)
  (cond ((or (eq 'tableau *cnf-solver*)
	     (and (eq 'infer *cnf-solver*)
		  ;; I have no idea whether this is the right heuristic.
		  (< (or num-clauses (length cnf-formula)) 2000)))
	 (tableau-solve-cnf cnf-formula file-basename vtable))
        ((eq 'ubersolve *cnf-solver*)
	 ;; (format t "~&Calling ubersolve: ")
	 (ubersolve-solve-cnf cnf-formula file-basename vtable))
	((eq 'relsat *cnf-solver*)
	 (relsat-solve-cnf cnf-formula file-basename vtable))
	((or (eq 'walksat *cnf-solver*)
	     (eq 'infer *cnf-solver*))
	 (walksat-solve-cnf-three-tries cnf-formula file-basename 40 vtable))
	((eq 'walksat-then-tableau *cnf-solver*)
	 (or (walksat-solve-cnf-three-tries cnf-formula file-basename 20 vtable)
	     (let ((tableau-result
		    (tableau-solve-cnf cnf-formula file-basename *vtable*)))
	       (if tableau-result
		   (format t "*** Walksat failed but tableau succeeded for ~a~%"
			   file-basename))
	       tableau-result)))
	((eq 'walksat-then-relsat *cnf-solver*)
	 (or (walksat-solve-cnf-three-tries cnf-formula file-basename 20 vtable)
	     (let ((relsat-result
		    (relsat-solve-cnf cnf-formula file-basename *vtable*)))
	       (if relsat-result
		   (format t "*** Walksat failed but relsat succeeded for ~a~%"
			   file-basename))
	       relsat-result)))
	(t
	 (error "Bad value for *cnf-solver* ~s" *cnf-solver*))))


(defun solve-cnf-external-simplifier (cnf-formula &optional file-basename num-clauses)
  (declare (ignore num-clauses))
  (let* ((unsimple-cnf-file (format nil "~a.cnf" file-basename))
	 (vtable (write-cnf-file cnf-formula file-basename))
	 (simple-basename (format nil "~a-simple" file-basename))
	 (simple-cnf-file (format nil "~a.cnf" simple-basename))
	 (translation-file (format nil "~a.transl" file-basename))
	 (compact-error-file (format nil "/tmp/compact-cnf.~a.external-err"
				     (get-universal-time)))
	 ;; with flags " -p -s -l -b", compact fails.
	 (compact-cnf-args (list (format nil "-v~d" (1+ (hash-table-count vtable))) ; compact demands 1+
				 (format nil "-f~a" translation-file)
				 unsimple-cnf-file
				 simple-cnf-file)))
    (format t "~&running compact-cnf with args ~s~%" compact-cnf-args)
    (assert (probe-file unsimple-cnf-file))
    (time-program compact-cnf-program compact-cnf-args
		  :error-output compact-error-file)
    (let ((error-output-len
	   (with-open-file (error-stream compact-error-file :direction :input)
	     (file-length error-stream))))
      (if (zerop error-output-len)
	  (delete-file compact-error-file)
	(format t "Warning!  Examine error output in ~s~%" compact-error-file)))

    ;; Sometimes it doesn't seem to exist.  Why??  Well,
    ;; sometimes the unsimplified .cnf file contains a contradiction,
    ;; in which case compact-cnf will not produce a simplified .cnf file
    ;; Solution: just use the original, unsimplified .cnf file for solving
    ;; Of course, the solver will not find a solution, so some time is wasted...
    (if (not (and (probe-file translation-file)
		  (probe-file simple-cnf-file)))
	(error "Translation file ~a~% or simplified CNF file ~a is missing.~%See error file ~a"
	       translation-file simple-cnf-file compact-error-file))
    (with-open-file (instream simple-cnf-file :direction :input)
      (cond ((string-equal "p cnf 0 0" (read-line instream))
	     (let ((truth-vector (apply #'vector
					(read-compact-translation translation-file)))
		   (result '()))
	       ;; (declare (type (vector fixnum) truth-vector))
	       (maphash #'(lambda (key value)
			    (let ((truth (= -1 (aref truth-vector (1- value)))))
			      (push (if truth key (sprop-negate key))
				    result)))
			vtable)
	       result))
	    ((string-equal "UNSAT" (read-line instream))
	     nil)
	    (t
	     (setq *translation-file* translation-file)
	     (solve-cnf cnf-formula simple-basename nil vtable))))))


;; What are the right values for tries and cutoff?
;; -tries 20 -cutoff 1000000  takes too long
;; -tries 10 -cutoff 100000 is the default.  It (sometimes) fails on
;;   PRODIGY-BW.PRODIGY-SUSSMAN.lisp-6-ws.cnf

;; Formula must already be in CNF.
;; writes out a line to the appropriate data file
;; Returns nil (if no assignment) or a list of literal truth assignments,
;;    each of which is a proposition (a possibly negated literal).
;; I could omit the negated ones...
(defun walksat-solve-cnf-three-tries (cnf-formula file-basename tries &optional vtable)
  (or (walksat-solve-cnf-one-try cnf-formula file-basename tries
				 '("-noise" "30" "100") vtable)
      (walksat-solve-cnf-one-try cnf-formula file-basename tries
				 '("-noise" "20" "100") *vtable*)
      (walksat-solve-cnf-one-try cnf-formula file-basename tries
				 '("-noise" "50" "100") *vtable*)))

(defun walksat-command-args (tries cutoff noise-flags)
  (declare (fixnum tries) (integer cutoff) (list noise-flags))
  (append (list "-solcnf" "-numsol" "1"
		"-tries" (format nil "~a" tries)
		"-cutoff" (format nil "~a" cutoff))
	  noise-flags))

(defun walksat-solve-cnf-one-try (cnf-formula file-basename tries &optional (noise-flags '("-noise" "30" "100")) vtable)
  (let* ((cnf-file (format nil "~a.cnf" file-basename))
	 (outfile (format nil "~a-ws.out" file-basename)))

    (if (or (not vtable) (not (probe-file cnf-file)))
	(setq vtable (write-cnf-file cnf-formula file-basename)))
    (assert (or (not (member 'cnf *output-files*)) (probe-file cnf-file)))
    (setq *vtable* vtable)

    (setq *end-encode-time* (get-internal-run-time))
    ;; Bart Selman says: "cutoff should be about N^2 (sometimes less),
    ;; where N is the number of vars."
    (let* ((num-vars (hash-table-count vtable))
	   (cutoff (* 2 num-vars num-vars)))
      (let ((command-args (walksat-command-args tries cutoff noise-flags)))
	(format t "~&Calling walksat ~s on ~a~%" command-args cnf-file)
	;; (format t "~&command = ~a~%" command)
	(if (probe-file outfile)
	    (delete-file outfile))
	(time-program
	 ;; walksat allows 100,000 atoms, 500,000 clauses.
	 ;; walksath allows 100,000 atoms, any number of clauses.
	 (if (< num-vars 50000) walksat-program walksath-program)
	 command-args
	 :input cnf-file
	 :output outfile)))
    (setq *start-decode-time* (get-internal-run-time))
    (let* ((walksat-info (read-walksat-file outfile))
	   (truth-assignments (first walksat-info))
	   (num-clauses (second walksat-info))
	   (num-vars (third walksat-info))
	   (num-lits (fourth walksat-info))
	   (elapsed-time (fifth walksat-info))
	   (mean-flip (sixth walksat-info))
	   (mean-restarts (seventh walksat-info)))

      ;; This looks like just (cdr walksat-info).
      ;; Perhaps add the symbol 'walksat to this
      (setq *file-info*
	(list truth-assignments num-vars num-clauses num-lits elapsed-time mean-flip mean-restarts))
      ;; (format t "*file-info* = ~s~%" *file-info*)

      ;;use truth-assignments as bool to tell if we've found soln.
      ;;don't know how to get num-plan-steps just yet
      (and truth-assignments
	   (progn
	     (setq *truth-assignments* truth-assignments)
	     ;; Pragmatically, the original two lists are better than this,
	     ;; but this is more readable.
	     ;; (mapcar #'(lambda (literal value)
	     ;; 		 (if value
	     ;; 		     literal
	     ;; 		   (list :not literal)))
	     ;; vtable truth-assignments)

	     (if *external-simplifier*
		 (let ((simp-vec (apply #'vector truth-assignments)))
		   (setq truth-assignments
		     (mapcar #'(lambda (index)
				 (declare (fixnum index))
				 (cond ((= index -1) t)
				       ((= index -2) nil)
				       (t (aref simp-vec (1- index)))))
			     (read-compact-translation *translation-file*)))))

	     (let ((truth-vector (apply #'vector truth-assignments))
		   (result '()))
	       (maphash #'(lambda (key value)
			    (declare (fixnum value))
			    (let ((truth (aref truth-vector (1- value))))
			      (push (if truth key (sprop-negate key))
				    result)))
			vtable)
	       result))))))


;; Returns a list of literal truth assignments, each of which is a literal
;;    possibly wrapped in (NOT ...).
;; writes out a line to the appropriate data file
(defun tableau-solve-cnf (cnf-formula file-basename &optional vtable)
  (assert (or (not *external-simplifier*) (probe-file *translation-file*)))

  (let* ((cnf-file (format nil "~a.cnf" file-basename))
	 ;; tableau produces plain-outfile, but we move result to outfile
	 (plain-outfile (format nil "~a.out" file-basename))
	 (outfile (format nil "~a-tb.out" file-basename))
	 ;; (contradictionfile (format nil "~a.contr" file-basename))
	 )
    (if (or (not vtable) (not (probe-file cnf-file)))
	(setq vtable (write-cnf-file cnf-formula file-basename)))
    (assert (or (not (member 'cnf *output-files*)) (probe-file cnf-file)))
    (setq *vtable* vtable)

    (setq *end-encode-time* (get-internal-run-time))
    (format t "~&Calling tableau on ~a~%" cnf-file)

    ;; The output contradictionfile contains a "positive literals:" string, but
    ;; what follows it is only the last line of positive literals, not all of
    ;; them.  We don't really need it.
    (time-program tableau-program (list cnf-file "-solver" "tableau"))
    (if (not (probe-file plain-outfile))
	(if (not (probe-file *ksh-path*))
	    (error "can't find ksh, so can't run solve" *ksh-path* plain-outfile)
	  (error "solve produced no output!" plain-outfile)))
    (rename-file plain-outfile outfile)

    (setq *start-decode-time* (get-internal-run-time))
    (let* ((tableau-info (read-tableau-file outfile))
	   (true-truth-assignments (first tableau-info))
	   (numclause (second tableau-info))
	   (numvar (third tableau-info))
	   (numhorn (fourth tableau-info))
	   (t-elapsed-time (fifth tableau-info)))

      ;; Perhaps add the symbol 'tableau to this.
      ;; Problem:  we want (+ numclause numhorn) to be the number of clauses
      ;; in the .cnf file, but tableau doesn't report on unit clauses.
      (setq *file-info* (list true-truth-assignments numvar (+ numclause numhorn) nil t-elapsed-time 0.0 0.0))
      (and true-truth-assignments
	   (progn
             ;; Strip off the leading dummy zero, which distinguishes an
	     ;; all-false truth assignment from failure to find an assignment
	     (assert (zerop (car true-truth-assignments)))
	     (setq true-truth-assignments (cdr true-truth-assignments))
	     (setq *truth-assignments* true-truth-assignments)
	     ;; This doesn't work because we don't want to return nil
	     ;; for truth assignment with no positive literals.
	     ;; (mapcar #'(lambda (lit-num) (nth (1- lit-num) vtable))
	     ;;     (cdr true-truth-assignments))

;;; Incorrect; vtable is a hash table, not a list
;; (loop for literal in vtable
;; 	 as i = 1 then (1+ i)
;; 	 collect
;; 	   ;; slightly inefficient
;; 	   (if (member i true-truth-assignments)
;; 	       literal
;; 	     (list :not literal)))

	     (if *external-simplifier*
		 (setq true-truth-assignments
		   (let ((unsimp-true '()))
		     (loop for val in (read-compact-translation *translation-file*)
			 as i = 1 then (1+ i)
			 do (if (or (= val -1)
				    (member val true-truth-assignments))
				(push i unsimp-true)))
		     unsimp-true)))

	     (let ((result '()))
	       (maphash #'(lambda (key value)
			    (assert (not (sprop-negated-p key)))
			    (if (member value true-truth-assignments)
				(push key result)
			      ;; gratuitous; helps debugging
			      (push (sprop-negate key) result)))
			vtable)
	       result))))))


;; Returns a list of literal truth assignments, each of which is a literal
;;    possibly wrapped in (NOT ...).
;; writes out a line to the appropriate data file
(defun relsat-solve-cnf (cnf-formula file-basename &optional vtable)
  (assert (or (not *external-simplifier*) (probe-file *translation-file*)))

  (let* ((cnf-file (format nil "~a.cnf" file-basename))
	 (outfile (format nil "~a-rs.out" file-basename)))
    (if (or (not vtable) (not (probe-file cnf-file)))
	(setq vtable (write-cnf-file cnf-formula file-basename)))
    (assert (or (not (member 'cnf *output-files*)) (probe-file cnf-file)))
    (setq *vtable* vtable)

    (setq *end-encode-time* (get-internal-run-time))
    (format t "~&Calling relsat on ~a~%" cnf-file)
    (time-program relsat-program (list "4" cnf-file) :output outfile
		  :if-output-exists :supersede)

  (post-relsat-or-ubersolve-decode cnf-file outfile vtable)))


;; slightly modified version of relsat-solve-cnf
(defun ubersolve-solve-cnf (cnf-formula file-basename &optional vtable)
  (assert (or (not *external-simplifier*) (probe-file *translation-file*)))

  (let* ((cnf-file (format nil "~a.cnf" file-basename))
         (outfile (format nil "~a-rs.out" file-basename)))
    (if (or (not vtable) (not (probe-file cnf-file)))
        (setq vtable (write-cnf-file cnf-formula file-basename)))
    (assert (or (not (member 'cnf *output-files*)) (probe-file cnf-file)))
    (setq *vtable* vtable)

    (setq *end-encode-time* (get-internal-run-time))

    (let ((command (format nil "ubersolve  ~a ~a \"~a\" \"~a\"   " cnf-file outfile ubersolve-simp-options ubersolve-solve-options)  ))
      (format t "Calling ubersolve: ~a~%" command))

    (time-program "ubersolve" (list cnf-file outfile
				    ubersolve-simp-options
				    ubersolve-solve-options))

    (post-relsat-or-ubersolve-decode cnf-file outfile vtable)))


;; Abstracted out of relsat-solve-cnf to avoid repeating code  (jim)
(defun post-relsat-or-ubersolve-decode (cnf-file outfile vtable)

    (setq *start-decode-time* (get-internal-run-time))
    (let* ((relsat-info (read-relsat-file outfile))
           (truth-assignments (first relsat-info))
           (elapsed-time (second relsat-info))
           (learn-order (third relsat-info))
           (branch-points (fourth relsat-info))
           (vars-labeled (fifth relsat-info))
           (failed-count (sixth relsat-info))
           ;; Returns 3 values, but we ignore error output and process number.
           ;; This is a bit expensive; oh well.
	   (out-stream (run-program "count-cnf" (list cnf-file)
				    :wait nil :output :stream))
	   (num-vars (read out-stream))
	   (num-clauses (read out-stream))
           (num-lits (read out-stream)))
      (declare (ignore vars-labeled failed-count))
      (close out-stream)

      ;; This looks like just (cdr walksat-info).
      ;; Perhaps add the symbol 'relsat to this
      (setq *file-info*
        (list truth-assignments
              num-vars num-clauses num-lits
              elapsed-time learn-order branch-points))
      ;; (format t "*file-info* = ~s~%" *file-info*)

      ;;use truth-assignments as bool to tell if we've found soln.
      ;;don't know how to get num-plan-steps just yet
      (and truth-assignments
           (progn
             (setq *truth-assignments* truth-assignments)

             (if *external-simplifier*
                 (let ((simp-vec (apply #'vector truth-assignments)))
                   (setq truth-assignments
                     (mapcar #'(lambda (index)
                                 (cond ((= index -1) t)
                                       ((= index -2) nil)
                                       (t (aref simp-vec (1- index)))))
                             (read-compact-translation *translation-file*)))))

             (let ((truth-vector (apply #'vector truth-assignments))
                   (result '()))
               (maphash #'(lambda (key value)
                            (let ((truth (aref truth-vector (1- value))))
                              (push (if truth key (sprop-negate key))
                                    result)))
                        vtable)
     result)))))

;;;
;;; Call walksat multiple times -- for doing testing runs
;;;

;; calls multi-walksat-solve-cnf for all options (that exist) of a given
;; problem.
(defun problem-walksat-solve-cnf (name &optional (n 5) (tries 10))
  (dolist (*frame-strategy* *frame-strategies*)
    (dolist (*action-representation* *action-representations*)
      (dolist (*exclusion-strategy* *exclusion-strategies*)
	(dolist (*types-strategy* *types-strategies*)
	    (if (check-flags t)
		(let* ((steps (get-plan-length name))
		       (filebase (format nil "~a-~a-~a-snum" name
					 (flags-brief)
					 steps)))
		  (if (probe-file (format nil "~a.cnf" filebase))
		      (multi-walksat-solve-cnf n filebase tries)))))))))


;; calls multi-walksat-solve-cnf for the specified flags
(defun flags-walksat-solve-cnf (name &optional
				     (n 5)
				     (tries 10)
				     (flag-lst *flags-list-maximal-types*))

  (dolist (cur-flags flag-lst)
    (set-flags cur-flags)
    (if (check-flags t)
	(let* ((steps (get-plan-length name))
	       (filebase (format nil "~a-~a-~a-renum" name
				 (flags-brief)
				 steps)))
	  (if (probe-file (format nil "~a.cnf" filebase))
	      (multi-walksat-solve-cnf n filebase tries))))))




;; runs N trials of walksat on <basename>.cnf  with tries TRIES and
;; cutoff CUTOFF and writes a data file with the results
(defun multi-walksat-solve-cnf (n basename tries)
  (declare (fixnum n) (fixnum tries))
  (let* ((total-time 0.0)
	 (cnf-file (format nil "~a.cnf" basename))
	 (num-vars (or
		    ;; Under what conditions does this call fail?
		    (number-of-variables cnf-file)
		    (error "When does this happen?  I'm curious...")
		    ;; (progn
;; 		      (with-open-file (str cnf-file :direction :output :if-exists :append)
;; 			(format str "p cnf 1 1~%1 0~%"))
;; 		      0)
 		    ))
	 (cutoff (* 2 num-vars num-vars))
	 (outfile (format nil "~a-ws.out" basename))
	 (datafile (format nil "~a-ws.data" basename))
	 (walksat-program (if (< num-vars 50000) walksat-program walksath-program))
	 (command-args (walksat-command-args tries cutoff '("-noise" "30" "100"))))
    (with-open-file (ostream datafile :direction :output
		     :if-does-not-exist :create :if-exists :rename-and-delete)
      (write-line datafile ostream)
      (write-line "" ostream)
      (do ((i 0 (+ i 1)))
	  ((= i n))
	(declare (fixnum i))
	(format t "~&Calling walksat on ~a.cnf:  ~d~%" basename i)
	(time-program walksat-program command-args
		      :input cnf-file :output outfile)
	(let ((ws-info (read-walksat-file outfile)))
	  (setq total-time (+ total-time (fifth ws-info)))
	  (if (null (car ws-info)) (format t "Walksat failed")
	    (format t "Walksat succeeded!!"))
	  (write ws-info :stream ostream)
	  (fresh-line ostream)))
      (write-line "" ostream)
      (write-string "Average walksat solve time (in milliseconds):  " ostream)
      (write (/ total-time n) :stream ostream)
      (write-line "" ostream))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Instantiate time-neutral formulas
;;;

;; Return a single CNF conjunction
(defun instantiate-init-step-and-goal (init-cnf onestep-cnf goal-cnf num-steps)
  (assert (cnf-p init-cnf))
  (assert (cnf-p onestep-cnf))
  (assert (cnf-p goal-cnf))
  (if *show-stages* (format t "~&Entered instantiate-init-step-and-goal~%"))

  (let ((goal-time-cnf (sprop-goal-instantiate goal-cnf num-steps))
	(init-time-cnf (sprop-start-state-instantiate init-cnf))
	(steps-time-cnfs
	 (loop for this-step from 1 to num-steps
	    collect (sprop-step-instantiate onestep-cnf this-step))))
    (assert (cnf-p goal-time-cnf))
    (assert (cnf-p init-time-cnf))
    (assert (every #'cnf-p steps-time-cnfs))
    (let ((result (apply #'cnf-conjoin
			 init-time-cnf goal-time-cnf steps-time-cnfs)))
      (assert (cnf-p result))
      (if *show-stages* (format t "~&Exited instantiate-init-step-and-goal~%"))
      result)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Utilities
;;;

;; Given tree structure, return copy with numbers changed to symbols
(defun number->symbol (x)
  (cond ((numberp x)
	 (intern (format nil "~d" x)))
	((consp x)
	 (cons (number->symbol (car x)) (number->symbol (cdr x))))
	(t
	 x)))

(defun symbol->number (x)
  (cond ((symbolp x)
	 (let* ((p (symbol-name x))
		;; paranoia; is this test necessary?
		(pr (and (every #'(lambda (c) (or (digit-char-p c)
						  (eq #\/ c)))
				p)
			 (read-from-string p))))
	   (if (and (numberp pr)
		    (string= (format nil "~d" pr) p))
	       pr
	     x)))
	((consp x)
	 (cons (symbol->number (car x)) (symbol->number (cdr x))))
	(t
	 x)))


;; Takes a filename and returns an absolute filename, for a problem file
(defun find-problem-file (file)
  (or (probe-file file)			; try in current directory first
      (dolist (dir *domains-directories* nil)
	(let ((merged (merge-pathnames file dir)))
	  (if (probe-file (merge-pathnames file dir))
	      (return merged))))
      (and (not (equal "lisp" (pathname-type file)))
	   (let ((file-dot-lisp (merge-pathnames *lisp-type-path* file)))
	     (dolist (dir *domains-directories* nil)
	       (let ((merged (merge-pathnames file-dot-lisp dir)))
		 (if (probe-file (merge-pathnames file dir))
		     (return merged))))))
      (error "Can't locate problem file ~s" file)))


;; given a problem name and an optional file name (without the .lisp extension)
;; return the problem
(defun find-problem (problemname &optional filename)
  (or (get-problem problemname)
      (if filename
	  (let* ((file (find-problem-file
			(merge-pathnames filename *lisp-type-path*)))
		 (basename (pathname-name file)))
	    (progn
	      (load-domains
	       ;; pathname-directory result can't be passed to merge-pathnames
	       :dirs (list (directory-namestring file))
	       :domains (list basename))
	      (or (get-problem problemname)
		  (error "Didn't find problem ~a in file ~a" problemname filename))))
	(error "Please provide the name of the file containing problem ~a"
	       problemname))))


;; problemname of the problem
;; if the problem is not already loaded, the user should supply the
;; filename (just the name, not the full path) without the .lisp

(defun solve-find-problem (problemname &optional max-steps start-steps
						 filename experimentnum)
  (let ((prob (find-problem problemname filename)))
    (if prob
	(solve-problem prob max-steps start-steps experimentnum)
      (error "Problem ~a not found" problemname))))


;; Removes all the files created by solving a plan.
(defun medic-clean-directory ()
  ;; Perhaps also get rid of anything with the extensions followed by ".BAK".
  (let ((extensions '("*.transl" "*.cnf" "*.data" "*.info" "*-rs.out" "*-tb.out" "*-ws.out" "*.clauses" "*.stat" "*.percent" "*.litnum" "*.cnf-size")))
    #+:allegro (excl:run-shell-command (format nil "rm~{ ~a~}" extensions) :wait t)
    #+:cmu (dolist (ext extensions)
	     (dolist (file (directory ext))
		(delete-file file)))
    ))
