;; file-io.lisp

(in-package medic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write CNF files for tableau and walksat input
;;;

;; Takes a CNF formula and a filename.
;; Writes a walksat/tableau input file of the form
;;    p cnf 2 3
;;    -1 2 0
;;    -1 0
;;    2 0
;; Also writes a .litnum file giving mappings between literals and numbers.
;; Also writes a .goals file giving goal clauses.
;; Returns a hashtable mapping positive literals to numbers.
(defun write-cnf-file (cnf-formula basename)
  (declare (string basename))
  (assert (cnf-p cnf-formula))
  (assert (stringp basename))
  (flet ((write-clause (ostream clause)
	   (if (numberp clause)
	       (format ostream "~d " clause)
	     (let ((fn (first clause)))
	       (cond ((eq :not fn)
		      (format ostream "-~d " (second clause)))
		     ((eq :or fn)
		      (dolist (e (cdr clause))
			(if (listp e)
			    (format ostream "-~d " (second e))
			  (format ostream "~d " e))))
		     (t
		      (error "bad combiner ~s" fn)))))
	   (format ostream "0~%")))
    (if (member 'cnf *output-files*)
	(let ((cnf-filename (format nil "~a.cnf" basename)))
	  (multiple-value-bind (new-cnf pos-lit-table) (numberize-formula cnf-formula)
	    ;; used to use :rename-and-delete
	    (with-open-file (ostream cnf-filename :direction :output :if-does-not-exist :create :if-exists :rename)
	      (declare (type (or stream (member nil)) ostream))
	      (if (not ostream)
		  (error "Couldn't open cnf output file ~a" cnf-filename))
	      (format ostream "p cnf ~d ~d~%"
		      (hash-table-count pos-lit-table)
		      ;; Ouch, expensive.  Don't I have this somewhere else?
		      (length (cnf-clauses new-cnf)))

	      ;;now handle output of each clause
	      (dolist (clause (cnf-clauses new-cnf))
		(write-clause ostream clause)))
	    (assert (probe-file cnf-filename))
	    (if (member 'litnum *output-files*)
		(let ((litnum-filename (format nil "~a.litnum" basename)))
		  (with-open-file (ostream litnum-filename :direction :output :if-does-not-exist :create :if-exists :rename-and-delete)
		    (maphash #'(lambda (key value)
				 (format ostream "~s ~s~%" key value))
			     pos-lit-table))
		  (assert (probe-file litnum-filename))))
	    (if (member 'goals *output-files*)
		(let* ((old-hash-table-count (hash-table-count pos-lit-table))
		       (new-goal-cnf (numberize-formula *goal-cnf* pos-lit-table)))
		  ;; Comment out either the declare or the assert.
		  ;; (declare (ignore old-hash-table-count))
		  ;; If this assertion fails, then simplification was done;
		  ;; more to the point, the hash table contains more literals
		  ;; than the .cnf file, which can lead to confusion.
		  (assert (= old-hash-table-count (hash-table-count pos-lit-table)))
		  (let ((goals-filename (format nil "~a.goals" basename)))
		    (with-open-file (ostream goals-filename :direction :output :if-does-not-exist :create :if-exists :rename-and-delete)
		      (dolist (clause (cnf-clauses new-goal-cnf))
			(write-clause ostream clause)))
		    (assert (probe-file goals-filename)))))
	    pos-lit-table)))))

;; Takes a formula with arbitrary literal names.
;; Returns two values:
;;   * new cnf formula with numbers substituted for the literals
;;   * hashtable mapping positive literals to numbers
(defun numberize-formula (formula &optional (pos-lit-table (make-sprop-hash-table)))
  (declare (type hash-table pos-lit-table))
  ;; get all the literals
  (let ((num-pos-lits (hash-table-count pos-lit-table)))
    (declare (fixnum num-pos-lits))
    (labels ((convert-formula (f)
	       (if (and (listp f)
			(formula-combiner-p (car f)))
		   ;; Don't bother to do test if a literal operator
		   (cons (car f)
			 (mapcar #'convert-formula (cdr f)))
		 ;; Question: is it better to put only nonnegated propositions
		 ;; in the hash table, or put both and make sure they share
		 ;; the same value?
		 (let* ((nonneg (sprop-nonnegated f))
			(num (gethash nonneg pos-lit-table)))
		   (declare (type (or fixnum (member nil)) num))
		   ;; (format t "nonneg ~s~%" nonneg)
		   (if (not num)
		       (progn
			 (incf num-pos-lits)
			 (setq num num-pos-lits)
			 (setf (gethash nonneg pos-lit-table) num)))
		   (assert (not (zerop num)))
		   (if (sprop-negated-p f)
		       (- num)
		     num)))))
      (let ((numberized-formula (convert-formula formula)))
	(values numberized-formula pos-lit-table)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read CNF files
;;;

;;; Old, inefficient implementation
;; ;;gives the number of VARs in a cnf file
;; ;; Returns nil if no "p" line found
;; (defun number-of-variables (cnf-file)
;;   (let ((temporary-file (format nil "/tmp/~a-numvars" cnf-file)))
;;     (if (probe-file temporary-file)
;; 	(delete-file temporary-file))
;;     (run-program
;;      "awk"
;;      "($1 == \"p\") { print $3 }  (($1 != \"c\") && ($1 != \"p\")) { exit }"
;;      :input cnf-file
;;      :output temporary-file)
;;     (prog1			; execute forms in order, return first value
;;         (with-open-file (instream temporary-file :direction :input)
;; 	  (read instream nil nil))
;;       (if (probe-file temporary-file)
;; 	  (delete-file temporary-file)
;; 	;; how could the test fail??
;; 	))))

;; gives the number of VARs in a cnf file
;; Returns nil if no "p" line found
;; This version doesn't call external processes, create files, etc.
;; But does the "thereis" clause work?
(defun number-of-variables (cnf-file)
  (let ((result (with-open-file (istream cnf-file :direction :input)
		  (loop thereis (let* ((line-first-elt (read istream)))
				  (cond ((eq 'p line-first-elt)
					 (read istream)
					 ;; return third item on line
					 (read istream))
					((eq 'c line-first-elt)
					 (read-line istream)
					 nil)
					(t
					 ;; bail out
					 t)))))))
    (declare (type (or integer (member t)) result))
    (and (numberp result)
	 result)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read CNF solver output files
;;;

(defun read-from-string-after (wholestring prefix)
  (declare (type simple-base-string prefix))
  (let ((prefixlen (length prefix)))
    (assert (< prefixlen (length wholestring)))
    (assert (string= prefix wholestring :start2 0 :end2 prefixlen))
    (read-from-string wholestring t 'done :start prefixlen)))



;; Returns a list of 7 elements:
;;      list of truth assignments (booleans), or nil if no assignment found
;;	   Consider turning this into a vector instead, for efficiency.
;;      Number of clauses
;;      Number of variables
;;      Number of literals
;;      Elapsed time solving (msec)
;;      Mean flips until assign (part of solve output)
;;      Mean restarts until assign (part of solve output)
(defun read-walksat-file (filename)
  (declare (string filename))
  (assert (stringp filename))
  (let (numsolutions numatom numclause numliterals milliseconds
	(mean-flips-until-assign 0)
	(mean-restarts-until-assign 0)
	curr-line curr-item eofp)
    (with-open-file (istream filename :direction :input :if-does-not-exist :error)
      (if (not istream)
	  (error "Couldn't open cnf input file ~a" filename))

     (setq curr-line (read-line istream))
     (if (or (string-prefix-p curr-line "Bad argument ")
	     (search "error" curr-line)
	     (search "ERROR" curr-line))
	 (error "walksat: ~a" curr-line))
     (loop while (not (eq 'numatom (read istream)))
	 do (read-line istream))

     (setq curr-item (read istream))
     (assert (eq '= curr-item))
     (setq numatom (read istream))	; number of variables
     (setq curr-item (read-char istream))
     (assert (eq #\, curr-item))
     (setq curr-item (read istream))
     (assert (eq 'numclause curr-item))
     (setq curr-item (read istream))
     (assert (eq '= curr-item))
     (setq numclause (read istream))
     (setq curr-item (read-char istream))
     (assert (eq #\, curr-item))
     (setq curr-item (read istream))
     (assert (eq 'numliterals curr-item))
     (setq curr-item (read istream))
     (assert (eq '= curr-item))
     (setq numliterals (read istream))
     ;; make sure walksat didn't notice an error
     (setq curr-item (read istream))
     (assert (eq 'wff curr-item))
     (setq curr-item (read istream))
     (assert (eq 'read curr-item))
     (setq curr-item (read istream))
     (assert (eq 'in curr-item))
     (setq curr-line (read-line istream nil))
     (loop
	while (and curr-line
		   (not (string-prefix-p curr-line "total elapsed seconds = ")))
	do
	  (setq curr-line (read-line istream nil)))
     (if (not curr-line)
	 ;; end-of-file was reached; this can happen if walksat is killed
	 (list nil numclause numatom numliterals milliseconds
	       mean-flips-until-assign mean-restarts-until-assign)
       (progn
	 (setq milliseconds (round (* 1000 (read-from-string-after
					    curr-line "total elapsed seconds = "))))
	 (setq curr-line (read-line istream))
	 (assert (string-prefix-p curr-line "average flips per second = "))
	 (setq curr-line (read-line istream))
	 (setq numsolutions (read-from-string-after curr-line "number solutions found = "))
	 (assert (or (= numsolutions 0) (= numsolutions 1)))
	 (if (zerop numsolutions)
	     (progn
	       (setq curr-line (read-line istream))
	       (assert (string= "ASSIGNMENT NOT FOUND" curr-line))
	       (list nil numclause numatom numliterals milliseconds
		     mean-flips-until-assign mean-restarts-until-assign))
	   (progn
	     (setq curr-line (read-line istream))
	     (setq mean-flips-until-assign
		   (read-from-string-after curr-line
					   "mean flips until assign = "))
	     (setq curr-line (read-line istream))
	     (assert (string-prefix-p curr-line "mean seconds until assign = "))
	     (setq curr-line (read-line istream))
	     (setq mean-restarts-until-assign
		   (read-from-string-after curr-line
					   "mean restarts until assign = "))
	     (setq curr-line (read-line istream))
	     (assert (string= "ASSIGNMENT FOUND" curr-line))
	     (list
	      (loop
	       while
	       (progn
		 (multiple-value-setq (curr-line eofp) (read-line istream nil))
		 (not eofp))
	       collect
	       ;; nil if var is negated, t if var is nonnegated
	       (not (char= (char curr-line 2) #\-)))
	      numclause numatom numliterals milliseconds
	      mean-flips-until-assign mean-restarts-until-assign))))))))


;; Returns a list of the following values
;;      List of numbers of true literals, or nil if no assignment found.
;;          The list of returned numbers always contains 0 at the beginning,
;;          to distinguish no true literals from no assignment; strip off the
;;          0 before using.
;;      Number of clauses
;;      Number of variables
;;      Number of binary horn clauses
;;    ? Elapsed time solving (msec)

(defun read-tableau-file (filename)
  (declare (string filename))
  (assert (stringp filename))
  (let ((positive-literals nil)
	(clauses 0) (binary-horn-clauses 0) (terms 0)
	(milliseconds 0)
	curr-line curr-item)
    (with-open-file (istream filename :direction :input :if-does-not-exist :error)
      (setq curr-line (read-line istream))
      (assert (string= "c type: solution" curr-line))
      (setq curr-line (read-line istream))
      (assert (string= "c solver: tableau -p \"filter -bmn\"" curr-line))
      (read-line istream)		; c solved: Fri Dec 13 11:36:14 PST 1996
      (setq curr-line (read-line istream))
      ;; "line 42: contradiction in input for literal '2'"
      (cond ((search "contradiction" curr-line :start2 8 :end2 30)
	     (list positive-literals clauses terms binary-horn-clauses milliseconds))
	    ((string-suffix-p curr-line ":  not found")
	     (error curr-line))
	    (t
	     (assert (string-prefix-p curr-line "line "))
	     (with-input-from-string (strstream curr-line :start (1+ (position #\: curr-line)))
	       (setq curr-item (read strstream))
	       (assert (eq 'loaded curr-item))
	       (setq clauses (read strstream))
	       (setq curr-item (read strstream))
	       (assert (eq 'clauses curr-item))
	       (setq curr-item (read-char strstream))
	       (assert (eq #\, curr-item))
	       (setq binary-horn-clauses (read strstream))
	       (setq curr-item (read strstream))
	       (assert (eq 'binary curr-item))
	       (setq curr-item (read strstream))
	       (assert (eq 'horn curr-item))
	       (setq curr-item (read strstream))
	       (assert (eq 'clauses curr-item))
	       (setq curr-item (read-char strstream))
	       (assert (eq #\, curr-item))
	       (setq terms (read strstream))
	       (setq curr-item (read strstream))
	       (assert (eq 'terms curr-item))
	       )
	     (setq curr-line (read-line istream))
	     (if (string= curr-line "contradiction in input after unit resolution")
		 (list positive-literals clauses terms binary-horn-clauses milliseconds)
	       (progn
		 (assert (string= "" curr-line))
		 (setq curr-line (read-line istream))
		 (assert (string= "query term? 0" curr-line))
		 (setq curr-line (read-line istream))
		 (assert (string= "using picker 'filter'" curr-line))
		 (setq curr-line (read-line istream))
		 (loop
		  while
		  (string-prefix-p curr-line "assertions ")
		  do
		  (setq curr-line (read-line istream)))
		 (if (string= curr-line "positive literals:")
		     (let (curr-item)
		       (setq positive-literals
			     (cons 0
				   (loop
				    while
				    (progn
				      (setq curr-item (read istream))
				      (numberp curr-item))
				    collect
				    ;; nil if var is negated, t if var is nonnegated
				    curr-item)))
		       (setq curr-line (read-line istream))
		       (assert (string= curr-line "end of literals"))
		       (setq curr-line (read-line istream))
		       (assert (string= curr-line "good model"))
		       (setq curr-line (read-line istream))))
		 (assert (string-prefix-p curr-line "number of non-horn clauses "))
		 (setq curr-line (read-line istream))
		 (setq curr-line (read-line istream))
		 (setq curr-line (read-line istream))
		 (setq curr-line (read-line istream))
		 (setq curr-line (read-line istream))
		 (setq curr-line (read-line istream))
		 (setq milliseconds
		       (round (* 1000 (read-from-string-after
				       curr-line "run time                   "))))
		 (list positive-literals clauses terms binary-horn-clauses milliseconds))))))))


;; Returns a list of the following values:
;;      list of truth assignments (booleans), or nil if no assignment found
;;	   Consider turning this into a vector instead, for efficiency.
;;	elapsed time (CPU msec)
;;	learn order (a number; recommended values are 3 or 4)
;;	branch points explored
;;	variables labeled
;;	failed count (number of tries before success)
(defun read-relsat-file (filename)
  (declare (string filename))
  (assert (stringp filename))

  (let (truth-assignments cpu-msec learn-order branch-points labeled failed-count
	curr-line)
    (with-open-file (istream filename :direction :input :if-does-not-exist :error)
     (setq curr-line (read-line istream))
     (if (not (string-prefix-p curr-line "Instance is: "))
	 (error "relsat file should start with input filename: ~a" curr-line))
     (setq curr-line (read-line istream))
     (if (not (string-prefix-p curr-line "Random number seed: "))
	 (error "Expected random number seed in relsat file: ~a" curr-line))
     (setq learn-order (read-from-string-after (read-line istream) "Learn order is: "))
     (setq truth-assignments
       (let ((thisval (read istream (not *run-program-timeout*) 'EOF)))
	 (cond ((eq thisval 'UNSAT)
		nil)
	       ((eq thisval 'SAT)
		(error "SAT, but no assignments?"))
	       ((eq thisval 'TIME)	; TIME LIMIT EXPIRED
		(read istream)
		(read istream)
		nil)
	       ((eq thisval 'EOF)
		'EOF)
	       (t
		(let ((thisnum 1))
		  (loop while (not (eq thisval 'SAT))
		      collect (progn
				(assert (= thisnum (abs thisval)))
				;; (check-for-invariants read-relsat-file (int thisnum thisval))
				(prog1
				    (plusp thisval)
				  (setq thisnum (1+ thisnum)
					thisval (read istream))))))))))
     (if (eq truth-assignments 'EOF)
	 (setq truth-assignments nil
	       cpu-msec (* 1000 *run-program-timeout*))
       (progn
     (setq branch-points (read-from-string-after (read-line istream) "Branch Points Explored: "))
     (setq labeled (read-from-string-after (read-line istream) "Variables Labelled    : "))
     (setq cpu-msec (round (* 1000 (read-from-string-after (read-line istream) "Elapsed Time (cpu sec): "))))
	 (setq failed-count (read-from-string-after (read-line istream) "Failed count: "))))
     (list truth-assignments cpu-msec learn-order branch-points labeled failed-count))))
;; Testing
;; (read-relsat-file "/homes/gws/mernst/tmp/relsat-output")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read translation file produced by compact (ntab) program
;;;

;; Returns a list of indices; -1 means true, -2 means false.
(defun read-compact-translation (translation-file)
  (declare (string translation-file))
  (assert (stringp translation-file))
  (with-open-file (istream translation-file
		   :direction :input :if-does-not-exist :error)
    (let ((curr-index 0)
	  curr-line eofp)
      (loop
	  while
	    (progn
	      (multiple-value-setq (curr-line eofp) (read-line istream nil))
	      (not eofp))
	  collect
	    ;; nil if var is negated, t if var is nonnegated
	    (progn
	      (incf curr-index)
	      (with-input-from-string (str curr-line)
		(let ((old (read str))
		      (new (read str)))
		  ;; The below is too much paranoia even for me
		  ;; (multiple-value-bind (dummy eofp) (read str)
		  ;;    (declare (ignore dummy))
		  ;;     (assert eofp))
		  ;; Comment out either the declare or the assert.
		  ;; (declare (ignore old))
		  (assert (= curr-index old))
		  new)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write "experiment" output files (.info and .data)
;;;


;; Writes the info file in the following format:
;;   Group: 0
;;   Experiment: 0
;;   Encoder: -*F* SIMPLE -*E* COMPLICATED
;;   Solve:
;;   Decoder:
;;   Machine: foo
;;   Comment: comment

(defun write-info-file (experimentnum problemname  encoder-switches solve-switches decoder-switches machine-type comment)
  (declare (fixnum experimentnum))
  (declare (type (or string list) encoder-switches))
  (declare (type (or string list) solve-switches))
  (declare (type (or string list) decoder-switches))
  (declare (string machine-type))
  (declare (string comment))

  (assert (integerp experimentnum))
  ;; Switches are either a string or
  ;;   a list of two-element lists of '(*frame-strategy* 'classical)
  (assert (or (stringp encoder-switches) (listp encoder-switches)))
  (assert (or (stringp solve-switches) (listp solve-switches)))
  (assert (or (stringp decoder-switches) (listp decoder-switches)))
  (assert (stringp machine-type))
  (assert (stringp comment))
  (with-open-file (ostream (info-file-name experimentnum problemname )
			   :direction :output :if-does-not-exist :create :if-exists :rename-and-delete)
    (flet ((write-flags (switches)
	       (if (stringp switches)
		   (write-string switches ostream)
		 ;; switches is a list of 2-element lists
		 (dolist (pair switches)
		   (format ostream "~a=~a " (car pair) (second pair))))
	       (fresh-line ostream)))
      (write-line "Group: 2" ostream)
      (write-string "Experiment: " ostream)
      (write-line (prin1-to-string experimentnum) ostream)
      (write-string "Encoder: " ostream)
      (write-flags encoder-switches)
      ;; (write-string (flags) ostream)
      (write-string "Solve: " ostream)
      (write-flags solve-switches)
      (write-string "Decoder: " ostream)
      (write-flags decoder-switches)
      (write-string "Machine: " ostream)
      (write-line machine-type ostream)
      (write-string "Comment: " ostream)
      (write-line comment ostream))))


;; uses the list *file-info* which contains
;; success-flag num-vars num-clauses num-lits elapsed-time-solving mean-flips mean-restarts
;;uses *start-encode-time* *end-encode-time* *start-decode-time*
;;*end-decode-time* to get elapsed time encoding(msec) and Elapsed time
;;decoding (msec) (zero if solve fails)
;; Takes an experiment number, filename, max of time steps,
;; Number of time steps in plan produced (ignores noops),
;;  Number of plans created (zero if not UCPOP), Number of plans visited (zero if not UCPOP)
;; and finally a ostream to write the next line of output to.
;; writes a line of the following form:
;;0 3 FIXIT -1 f 0 0 0 0 76320 0 0 0 12 12 23701 14878
(defun write-data-line-ucpop (expnum probname maxstep num-plan-steps
		              num-plans-created num-plans-visited ostream)
  (setq *end-decode-time* (get-internal-run-time))
  (let* (
	 ;;values from *file-info* list
	 (success (first *file-info*))
	 (numvars (second *file-info*))
	 (numclauses (third *file-info*))
	 (numlits (fourth *file-info*))
	 (elapsed-time-solving (fifth *file-info*))
	 (mean-flips (sixth *file-info*)) ; or learn-order, for relsat
	 (mean-restarts (seventh *file-info*)) ; or branch-points, for relsat
	 ;;timing values
	 (elapsed-time-decode (- *end-decode-time* *start-decode-time*))
	 (elapsed-time-encoding (- *end-encode-time* *start-encode-time*)))

    (assert (integerp expnum))
    (assert (symbolp probname))
    (assert (integerp maxstep))
    (assert (integerp numclauses))
    (assert (integerp numvars))
    (assert (integerp numlits))
    (assert (or (not elapsed-time-solving) (integerp elapsed-time-solving)))
    (assert (integerp elapsed-time-encoding))
    (assert (integerp elapsed-time-decode))
    (assert (numberp mean-flips))
    (assert (numberp mean-restarts))
    (assert (integerp num-plans-visited))
    (assert (integerp num-plan-steps))
    (assert (integerp num-plans-created))
    (assert (streamp ostream))
    (format ostream "2 ~d ~a ~d ~s ~d ~d ~d ~d ~d ~d ~d ~d ~d ~d ~d ~d"
	    expnum
	    probname
	    maxstep
	    (if success 't 'f)
	    numclauses numvars numlits
	    elapsed-time-encoding elapsed-time-solving elapsed-time-decode
	    mean-flips mean-restarts
	    num-plan-steps num-plan-steps ;; same since linear encoding
	    num-plans-created num-plans-visited)
    ))

;; Like write-data-line-ucpop, but doesn't take two of the args:
;;   num-plans-created num-plans-visited
(defun write-data-line-non-ucpop (expnum probname maxstep num-plan-steps ostream)
  (write-data-line-ucpop expnum probname maxstep num-plan-steps 0 0 ostream))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write statistics files
;;;

(defun write-domain-percent-stat-file (name num-fluents num-actions types probname)
  (let ((filename (format nil "~a.percent" name))
	(flags (brief-flags)))
    (if (probe-file filename)
	;; What is this "vesin" test for??  Is that the first set of tests,
	;; so "percent" is just written once?  FIX!
	(if (equal "vesin" flags)
	    (with-open-file (ostream filename :direction :output
			     :if-exists :append)
	      (format ostream "~a~%" probname))
	  nil)
      (with-open-file (ostream filename
		       :direction :output :if-does-not-exist :create
		       :if-exists nil)
	;; Per fluent info
	(format ostream "~a~%~%" filename)
	(format ostream "% of actions which initially affect the given fluent~%")
	(format ostream "(may change due to cloning or action elimination)~%")
 	(dolist (f *fluent-appearances-in-effects*)
 	  (format ostream "~a	~a	~a%~%" (car f) (* 100.0 (/ (length (cdr f)) num-actions)) (cdr f)))
 	(dolist (f types)
 	  (format ostream "~s 0.0%~%" f))
	;; Per action info
	(format ostream "~%% of fluents which the given action affects~%")
	(dolist (a *action-fluents*)
	  (format ostream "~a ~a%~%" (car a) (* 100.0 (/ (length (cdr a)) num-fluents))))
	(format ostream "~%Associated problems:~%")
	(if (equal "vesin" flags)
	    (format ostream "~a~%" probname))))))


(defun write-problem-clause-breakdown-file (problemname op-len frame-len excl-len)
  (let ((filename (format nil "~a-~a.clauses" problemname (brief-flags))))
    (if (probe-file filename) nil
      (with-open-file (ostream filename :direction :output
		       :if-does-not-exist :create :if-exists nil)
	(write-line filename ostream)
	(write-line "" ostream)
	(write-string "Number of act-clauses:  " ostream)
	(write op-len :stream ostream)
	(write-line "" ostream)
	(write-string "Number of frame-clauses:  " ostream)
	(write frame-len :stream ostream)
	(write-line "" ostream)
	(write-string "Number of exclusion-clauses:  " ostream)
	(write excl-len :stream ostream)
	(write-line "" ostream)))))


(defun write-prob-size-stat-file (filename cnf-form)
  (setq *end-encode-time* (get-internal-run-time))
  (with-open-file (ostream filename :direction :output
		   :if-does-not-exist :create :if-exists :rename-and-delete)
    (write-line filename ostream)
    (write-line "" ostream)
    (write (formula-size cnf-form) :stream ostream)
    (write-line "" ostream)
    (write-string "Encode time: " ostream)
    (write (- *end-encode-time* *start-encode-time*) :stream ostream)
    ))
