;;; Frame axioms

(in-package medic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User parameters
;;;

(defvar *frame-strategy* 'explanatory)
(defvar *frame-strategies* '(classical explanatory))
(deftype frame-strategy ()
  `(member ,@*frame-strategies*))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Local variables
;;;

(defvar *debug-classical-factored-frames* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Frame clauses
;;;

(defun frame-clauses ()
  (if *show-stages* (format t "~&Entering frame-clauses"))
  (cond ((eq *frame-strategy* 'classical)
	 (classical-frame-clauses))
	((eq *frame-strategy* 'explanatory)
	 (explanatory-frame-clauses))
	(t
	 (error "Bad value for *frame-strategy* ~s" *frame-strategy*))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Classical frame axioms
;;;

;; uses the global variables *ground-actions* and *ground-effects* to get a
;; list of actions and the fluents they affect and the global variable
;; *initial-state* to get a list of all possible fluents.  Assumes every
;; proposition in *initial-state* is represented as a list.
;; Returns a TNPF and substitutions to handle classical frame clauses.


;;;
;;; Classical frame axioms
;;;

(defun classical-frame-clauses ()
  (case *action-representation*
    ((regular simple-split overloaded-split bitwise)
     (classical-unfactored-frame-clauses))
    ((simple-split-factored overloaded-split-factored)
     (classical-factored-frame-clauses))
    (otherwise
     (error "What action representation? ~s" *action-representation*))))

(defun classical-unfactored-frame-clauses ()
  ;; prop-list is just *fluents*, but grounded for arguments.
  ;; *unary-types* (time-invariant fluents) should be removed from this, because
  ;;   classical frame axioms can ignore anything time-invariant.
  ;; Gross code.
  (let* ((prop-list
	  ;; *initial-state* is guaranteed to have time :beginning (or :none)
	  (mapcar #'(lambda (x)
		      (sprop-remove-time (sprop-nonnegated x)))
		  (prop-list-no-types)))
	 ;; act-prop-list omits ground fluents that are mentioned positively
	 ;; or negatively, because classical frame axioms are about what does
	 ;; not (cannot) change.  (Adding a clause about the fluent as it
	 ;; appears in the effect is extraneous, since it's already implied
	 ;; by the action axiom.  Adding a clause about its negation can
	 ;; be inconsistent, since we know its negation can't be true, as its
	 ;; nonnegated form is implied by the action axiom.)
	 (act-prop-list
	  (get-act-prop-list *ground-actions* *ground-effects* nil))
	 (clauses (get-simp-frame-clauses act-prop-list prop-list)))
    (values (eliminate-instantiated-non-removable-types-clauses-sprop clauses)
	    nil)))

(defun prop-list-no-types ()
  (cond ((member *types-strategy* '(time-only eliminate))
	 (remove-if #'(lambda (x)
			(eq ':none (sprop-time x)))
		    *initial-state*))
	((eq *types-strategy* 'no-types)
	 *initial-state*)
	(t
	 (error "Bad value for *types-strategy* ~s" *types-strategy*))))


;; This function is given a list of lists associating each action with the
;; propositions that that action changes and a list of all propositions.
;; The actions and the propositions are grounded for all possible arguments
;; but they are not grounded for time and they have none of the :before,
;; :during, :after keywords in them.  It returns the documented output of
;; new clauses and nil as the substitutions.

;; Returns a list of clauses (not a formula).
;; For better abstraction, perhaps should only take one argument.
(defun unchanged-prop->simp-frame-clauses (unchanged-prop action)
  (let* ((prop-before (sprop-add-time unchanged-prop ':before))
	 (prop-after (sprop-add-time unchanged-prop ':after))
	 (act-during (sprop-add-time action ':during))
	 (negated-act-during (sprop-negate act-during)))
    (list (list :or (sprop-negate prop-before)
		negated-act-during
		prop-after)
	  (list :or prop-before
		negated-act-during
		(sprop-negate prop-after)))))


;; Returns a list of frame clauses
;; ACT-CPROPS is a two-element list of (GROUNDED-ACTION CHANGED-PROP-LIST)
;;  where CHANGED-PROP-LIST is the grounded props changed by GROUNDED-ACTION
;; PROP-LIST is a list of timeless propositions that are not types.
(defun act-prop->simp-frame-clauses (act-cprops prop-list)
  (let* ((grounded-action (first act-cprops))
	 (changed-props (second act-cprops))
	 (unchanged-props (set-difference prop-list changed-props :test #'sprop-equal)))
    (mapcan #'(lambda (unchanged-prop)
		(unchanged-prop->simp-frame-clauses
		 unchanged-prop grounded-action))
	    unchanged-props)))

(defun get-simp-frame-clauses (act-prop-list prop-list)
  (mapcan #'(lambda (act-plus-chprops)
	      (act-prop->simp-frame-clauses
	       act-plus-chprops prop-list))
	  act-prop-list))


;; Old special case for axioms not mentioned; subsumed by general case in
;; classical-factored-frame-clauses below.
;; (let	 (non-mentioned-fluent-axioms
;; 	  (let ((result '()))
;; 	    (dolist (fluent+instantiations *possibly-true-fluents*)
;; 	      (let ((fluent-name (car fluent+instantiations))
;; 		    (fluent-instantiations (cdr fluent+instantiations)))
;; 		(dolist (args fluent-instantiations)
;; 		  ;; fluent and actionis(A) => fluent
;; 		  (push (convert-to-cnf
;; 			 (make-implies
;; 			  (list :and (make-sprop fluent-name args :before nil)
;; 				actionis)
;; 			  (make-sprop fluent-name args :after nil)))
;; 			result)
;; 		  ;; not(fluent) and actionis(A) => not(fluent)
;; 		  (push (convert-to-cnf
;; 			 (make-implies
;; 			  (list :and (make-sprop fluent-name args :before t)
;; 				actionis)
;; 			  (make-sprop fluent-name args :after t)))
;; 			result))))
;; 	    result))
;;  )

(defun classical-factored-frame-clauses ()
  (values
   (let ((op-frame-clauses (mapcan #'classical-factored-frame-clauses-for-op
				   (domain-operators *domain*))))
     (append (true-and-false-fluent-frames) op-frame-clauses))
   '()))


(defun true-and-false-fluent-frames ()
  (let ((result '()))
    ;; This takes care of all time: initial state already contains
    ;; closed-world assumption.
    ;; Need to do this for all type strategies, not just for none.
    (dolist (fluent+arglists *true-fluents*)
      (if (or (factored-splitting-p)
	      (not (length-one (car (cdr fluent+arglists)))))
	  (dolist (args (cdr fluent+arglists))
	    (push (make-sprop (car fluent+arglists) args :after nil)
		  result))))
    (dolist (fluent+arglists *false-fluents*)
      (if (or (factored-splitting-p)
	      (not (length-one (car (cdr fluent+arglists)))))
	  (dolist (args (cdr fluent+arglists))
	    (push (make-sprop (car fluent+arglists) args :after t)
		  result))))
    result))


;; Suppose foo(arg1,_) has effect clear(arg1).  We can treat all
;;   mentioned (negatively or positively) fluents the same:
;;     clear(A) and foo(not(A),_) => clear(A)
;;     not(clear(A)) and foo(not(A),_) => not(clear(A))
;;   or we can slightly optimize:
;;     clear(A) and actionis(foo) => clear(A)
;;     not(clear(A)) and foo(not(A),_) => not(clear(A))
;;   This code does the latter.
(defun classical-factored-frame-clauses-for-op-OLD-VERSION (op)
  ;; The effect is a conjunction.
  ;; If any conjunct is a disjunct, give up on factoring that part.
  ;;   (I think that can't happen, so I've deferred implementing it.)
  (assert (notany #'(lambda (x) (and (consp x) (member x '(:and :or)))) (conjunctive-clauses (op-effect op))))

  (let* ((actionis (make-actionis op))
	 ;; For each mentioned fluent, separately collect the effect conjuncts
	 ;;   that mention it positively and those that mention it negatively.
	 ;; fluent-alist is a list of 4-element lists of
	 ;;   (name pos-occurrences neg-occurrences instantiation-arglists).
	 (fluent-alist
	  (let ((alist (mapcar #'(lambda (finfo)
				   (list (car finfo) (list) (list) (cdr finfo)))
			       *variant-fluents*)))
	    (dolist (effect-as-list (conjunctive-clauses (op-effect op)))
	      (if *debug-classical-factored-frames*
		  (progn (format t "Working on ~s~%" effect-as-list)
			 (format t "Looking for ~s in ~s~%"
				 (car effect-as-list) alist)))
	      (let* ((effect (make-sprop-from-list-or-negation effect-as-list))
		     (fluent-assoc (assoc (sprop-name effect) alist)))
		(assert fluent-assoc)
		(assert (keyword-or-number-p (car (last (if (sprop-negated-p effect) (second effect-as-list) effect-as-list)))))
		(if (sprop-negated-p effect)
		    (push effect (third fluent-assoc))
		  (push effect (second fluent-assoc)))))
	    alist))
	 ;; The final result
	 (axioms '()))
    (if *debug-classical-factored-frames*
	(format t "actionis: ~s~%" actionis))
    (dolist (fluent+pos+neg+inst fluent-alist)
      ;; If it doesn't appear negatively, the positive case is as if
      ;; the fluent didn't appear at all.  (This is very slightly
      ;; redundant, but has a smaller representation.)
      (let ((fluent-name (car fluent+pos+neg+inst))
	    (pos-occurrences (second fluent+pos+neg+inst))
	    (neg-occurrences (third fluent+pos+neg+inst))
	    (instantiation-arglists (fourth fluent+pos+neg+inst))
	    (var-argnum-alist (let ((argnum 0))
				(declare (fixnum argnum))
				(mapcar #'(lambda (argname)
					    (incf argnum)
					    (cons argname argnum))
					;; might include time; no harm to do so
					(progn
					  (assert (keyword-or-number-p (op-time op)))
					  (op-formals op))))))
	(assert (every #'sprop-p pos-occurrences))
	(assert (every #'sprop-p neg-occurrences))
	(if *debug-classical-factored-frames*
	    (format t "~s: pos=~s, neg=~s~%" fluent-name pos-occurrences neg-occurrences))
	(assert (notany #'keywordp (mapcar #'car var-argnum-alist)))
	(if *debug-classical-factored-frames*
	    (format t "var-argnum-alist: ~s~%" var-argnum-alist))
	(dolist (fluent-args instantiation-arglists)
	  ;; Cope with positive occurrences, which leads to
	  ;;   not(fluent) and actionis(A) => not(fluent)
	  ;; if there are no such occurrences.
	  (let* ((pos-before (make-sprop fluent-name fluent-args :before nil))
		 (pos-after (make-sprop fluent-name fluent-args :after nil))
		 (neg-before (sprop-negate pos-before))
		 (neg-after (sprop-negate pos-after)))
	    ;; Generate frame axioms if the fluent has arguments or does not
	    ;; appear.  (Skip this frame for argumentless fluents that appear.)
	    (if (or (not (null fluent-args))
		    (not (or pos-occurrences neg-occurrences)))
		(progn
		  (push (cnf-fix-sprop-negation
			 (convert-to-cnf
			  (make-implies (list :and neg-before actionis
					      (formula-negate
					       (matching-var-sprops
						fluent-args pos-occurrences
						op var-argnum-alist)))
					neg-after)))
			axioms)
		  (push (cnf-fix-sprop-negation
			 (convert-to-cnf
			  (make-implies (list :and pos-before actionis
					      (formula-negate
					       (matching-var-sprops
						fluent-args neg-occurrences
						op var-argnum-alist)))
					pos-after)))
			axioms)))
	    (if *debug-classical-factored-frames*
		(format t "pushed: ~s~%        ~s~%" (second axioms) (first axioms)))
	    ))))
    ;; axioms is a list of CNF clauses
    (assert (every #'cnf-p axioms))
    ;; The call to delete-if is needed for unary op
    ;; when there is only one possible argument; maybe in other cases too.
    (if *debug-classical-factored-frames*
	(format t "axioms: ~s~%" axioms))
    (delete-if #'clause-contains-sprop-and-negation
	       (mapcan #'conjunctive-clauses axioms))))


;; This function is essentially the same as the OLD VERSION above, except
;; it calls NON-MATCHING-VAR-SPROPS instead of MATCHING-VAR-SPROPS (see below).
;; It also uses the returned result a bit differently.

(defun classical-factored-frame-clauses-for-op (op)
  ;; The effect is a conjunction.
  ;; If any conjunct is a disjunct, give up on factoring that part.
  ;;   (I think that can't happen, so I've deferred implementing it.)
  (assert (notany #'(lambda (x) (and (consp x) (member x '(:and :or)))) (conjunctive-clauses (op-effect op))))

  (let* ((actionis (make-actionis op))
	 ;; For each mentioned fluent, separately collect the effect conjuncts
	 ;;   that mention it positively and those that mention it negatively.
	 ;; fluent-alist is a list of 4-element lists of
	 ;;   (name pos-occurrences neg-occurrences instantiation-arglists).
	 (fluent-alist
	  (let ((alist (mapcar #'(lambda (finfo)
				   (list (car finfo) (list) (list) (cdr finfo)))
			       *variant-fluents*)))
	    (dolist (effect-as-list (conjunctive-clauses (op-effect op)))
	      (if *debug-classical-factored-frames*
		  (progn (format t "Working on ~s~%" effect-as-list)
			 (format t "Looking for ~s in ~s~%"
				 (car effect-as-list) alist)))
	      (let* ((effect (make-sprop-from-list-or-negation effect-as-list))
		     (fluent-assoc (assoc (sprop-name effect) alist)))
		(assert fluent-assoc)
		(assert (keyword-or-number-p (car (last (if (sprop-negated-p effect) (second effect-as-list) effect-as-list)))))
		(if (sprop-negated-p effect)
		    (push effect (third fluent-assoc))
		  (push effect (second fluent-assoc)))))
	    alist))
	 ;; The final result
	 (axioms '()))
    (declare (list axioms))
    (if *debug-classical-factored-frames*
	(format t "actionis: ~s~%" actionis))
    (dolist (fluent+pos+neg+inst fluent-alist)
      ;; If it doesn't appear negatively, the positive case is as if
      ;; the fluent didn't appear at all.  (This is very slightly
      ;; redundant, but has a smaller representation.)
      (let ((fluent-name (car fluent+pos+neg+inst))
	    (pos-occurrences (second fluent+pos+neg+inst))
	    (neg-occurrences (third fluent+pos+neg+inst))
	    (instantiation-arglists (fourth fluent+pos+neg+inst))
	    (var-argnum-alist (let ((argnum 0))
				(mapcar #'(lambda (argname)
					    (incf argnum)
					    (cons argname argnum))
					;; might include time; no harm to do so
					(progn
					  (assert (keyword-or-number-p (op-time op)))
					  (op-formals op))))))
	(assert (every #'sprop-p pos-occurrences))
	(assert (every #'sprop-p neg-occurrences))
	(if *debug-classical-factored-frames*
	    (format t "~s: pos=~s, neg=~s~%" fluent-name pos-occurrences neg-occurrences))
	(assert (notany #'keywordp (mapcar #'car var-argnum-alist)))
	(if *debug-classical-factored-frames*
	    (format t "var-argnum-alist: ~s~%" var-argnum-alist))
	(dolist (fluent-args instantiation-arglists)
	  ;; Cope with positive occurrences, which leads to
	  ;;   not(fluent) and actionis(A) => not(fluent)
	  ;; if there are no such occurrences.
	  (let* ((pos-before (make-sprop fluent-name fluent-args :before nil))
		 (pos-after (make-sprop fluent-name fluent-args :after nil))
		 (neg-before (sprop-negate pos-before))
		 (neg-after (sprop-negate pos-after)))
	    ;; Generate frame axioms if the fluent has arguments or does not
	    ;; appear.  (Skip this frame for argumentless fluents that appear.)
	    (if (or (not (null fluent-args))
		    (not (or pos-occurrences neg-occurrences)))
		(progn
		  (push (cnf-fix-sprop-negation
			 (convert-to-cnf
			  (make-implies (list :and neg-before
					      (let ((non-matches
						     (non-matching-var-sprops
						      fluent-args
						      pos-occurrences
						      op var-argnum-alist)))
						(if (eq *action-representation*
							'simple-split-factored)
						    (if (equal non-matches
							       '(:and))
							actionis non-matches)
						  (list :and actionis
							non-matches))))
					neg-after)))
			axioms)
		  (push (cnf-fix-sprop-negation
			 (convert-to-cnf
			  (make-implies (list :and pos-before
					      (let ((non-matches
						     (non-matching-var-sprops
						      fluent-args
						      neg-occurrences
						      op var-argnum-alist)))
						(if (eq *action-representation*
							'simple-split-factored)
						    (if (equal non-matches
							       '(:and))
							actionis non-matches)
						  (list :and actionis
							non-matches))))
					pos-after)))
			axioms)))
	    ))))
    ;; axioms is a list of CNF clauses
    (assert (every #'cnf-p axioms))
    ;; The call to delete-if is needed for unary op
    ;; when there is only one possible argument; maybe in other cases too.
    (if *debug-classical-factored-frames*
	(format t "axioms: ~s~%" axioms))
    (delete-if #'clause-contains-sprop-and-negation
	       (mapcan #'conjunctive-clauses axioms))))



;; Given args to fluent and its occurrences, determine
;; variable bindings that could lead to the fluent being true.
;; (Prohibit these in the frame axiom that says nothing changed.)
;; For example, suppose
;;      f-args (a c)
;;      varnum (?x 1) (?y 2) (?z 3)
;; f-occ  (fl a ?y) (fl ?y ?z) (fl ?y b)
;;  => (or  (and (and) (arg2 c))  (and (arg2 a) (arg3 c))  (or))
;; f-occ  (fl b c) (fl a b)
;;  => (or  (and (or) (and))  (and (and) (or)))
;; f-occ  (fl a c) (fl b c) (fl a b)
;;  => (or  (and (and) (and))  (and (or) (and))  (and (and) (or)))
;; f-occ  (fl ?x ?y) (fl ?y ?z)
;;  => (or (and (arg1 a) (arg2 c)) (and (arg2 a) (arg3 c)))
(defun matching-var-sprops (fluent-args fluent-occurrences op var-argnum-alist)
  (if *debug-classical-factored-frames*
      (format t "  excluded-var-sprops (len=~d) <= ~s ~s~%"
	      (op-arity op) fluent-args fluent-occurrences))
  (cons :or
	(delete-if
	 #'(lambda (conjct)
	     (member (cnf-false) conjct :test #'equal))
	 (mapcar
	  #'(lambda (occurrence)
	      (declare (type sprop occurrence))
	      (assert (sprop-p occurrence))
	      (assert (= (length (sprop-args occurrence)) (length fluent-args)))

	      ;; A certain object may be a legal argument to the fluent
	      ;; but may have been deduced to be an illegal argument to the
	      ;; current action.  This checks for that situation.
	      ;; Skip this fluent occurrence (ie, don't exclude its
	      ;; variables) if, for some fluent arg and
	      ;; corresponding appearance in a precond/effect, the
	      ;; appearance is a variable which is excluded by the
	      ;; operator's obj-lst.
	      (let ((skip-occurrence-p
		     (some #'(lambda (actual op-formal)
			       (declare (symbol actual) (symbol op-formal))
			       (assert (symbolp actual))
			       (assert (symbolp op-formal))
			       (let* ((formal-assoc
				       (assoc op-formal var-argnum-alist))
				      (op-argnum (cdr formal-assoc)))
				 (declare (fixnum op-argnum))
				 ;; If no formal-assoc, this must have been
				 ;; a constant, not a variable.
				 ;; Return true if this is a variable and it
				 ;; cannot be the op-argnum'th argument.
				 (and formal-assoc
				      (not (member actual
						   (nth (1- op-argnum) (op-obj-lst op)))))))
			   fluent-args
			   (sprop-args occurrence))))
		(if *debug-classical-factored-frames*
		    (format t "    skip-occurrence-p: ~s ~s ~s~%"
			    skip-occurrence-p
			    fluent-args (sprop-args occurrence)))
		(if skip-occurrence-p
		    (list :and (cnf-false))
		  (cons :and
			;; ought to do (delete-duplicates ... :test #'sprop-equal)
			;; (except that not all the elements are necessarily sprops).
			(delete-if
			 #'cnf-true-p
			 (mapcar
			  #'(lambda (actual op-formal)
			      (declare (symbol actual) (symbol op-formal))
			      (assert (symbolp actual))
			      (assert (symbolp op-formal))
			      (let* ((formal-assoc
				      (assoc op-formal var-argnum-alist)))
				(cond (formal-assoc
				       ;; variable
				       (let ((op-argnum (cdr formal-assoc)))
					 (split-argnum-sprop op op-argnum actual :during)))
				      ;; constant
				      ((eq actual op-formal)
				       ;; two equal constants
				       (cnf-true))
				      (t
				       (assert (not (varp op-formal)))
				       ;; two unequal constants
				       (cnf-false)))))
			  fluent-args
			  (sprop-args occurrence)))))))
	  fluent-occurrences))))


;; This function is like MATCHING-VAR-SPROPS execpt that it returns a formula
;; stating which bindings will make it so that the fluent-occurrences do not
;; satisfy the fluent-args.  A key feature is that it must state this in a
;; positive way.  That is, it says which bindings these are, rather than
;; stating the negation of which bindings these aren't.  The latter is
;; what the original implementation of classical factored frames does, and
;; this subverts the idea of classical frames.
(defun non-matching-var-sprops (fluent-args fluent-occurrences op
				var-argnum-alist)
  (if *debug-classical-factored-frames*
      (format t "  excluded-var-sprops (len=~d) <= ~s ~s~%"
	      (op-arity op) fluent-args fluent-occurrences))
  (cons :and
	(delete-if
	 #'(lambda (conjct)
	     (member (cnf-true) conjct :test #'equal))
	 (mapcar
	  #'(lambda (occurrence)
	      (declare (type sprop occurrence))
	      (assert (sprop-p occurrence))
	      (assert (= (length (sprop-args occurrence)) (length fluent-args)))

	      ;; A certain object may be a legal argument to the fluent
	      ;; but may have been deduced to be an illegal argument to the
	      ;; current action.  This checks for that situation.
	      ;; Skip this fluent occurrence (ie, don't exclude its
	      ;; variables) if, for some fluent arg and
	      ;; corresponding appearance in a precond/effect, the
	      ;; appearance is a variable which is excluded by the
	      ;; operator's obj-lst.
	      (let ((skip-occurrence-p
		     (some #'(lambda (actual op-formal)
			       (declare (symbol actual) (symbol op-formal))
			       (assert (symbolp actual))
			       (assert (symbolp op-formal))
			       (let* ((formal-assoc
				       (assoc op-formal var-argnum-alist))
				      (op-argnum (cdr formal-assoc)))
				 ;; If no formal-assoc, this must have been
				 ;; a constant, not a variable.
				 ;; Return true if this is a variable and it
				 ;; cannot be the op-argnum'th argument.
				 (and formal-assoc
				      (not (member actual
						   (nth (1- op-argnum) (op-obj-lst op)))))))
			   fluent-args
			   (sprop-args occurrence))))
		(if *debug-classical-factored-frames*
		    (format t "    skip-occurrence-p: ~s ~s ~s~%"
			    skip-occurrence-p
			    fluent-args (sprop-args occurrence)))
		(if skip-occurrence-p
		    (list :and (cnf-true))
		  (cons :or
			;; ought to do (delete-duplicates ... :test #'sprop-equal)
			;; (except that not all the elements are necessarily sprops).
			(delete-if
			 #'cnf-false-p
			 (mapcar
			  #'(lambda (actual op-formal)
			      (declare (symbol actual) (symbol op-formal))
			      (assert (symbolp actual))
			      (assert (symbolp op-formal))
			      (let* ((formal-assoc
				      (assoc op-formal var-argnum-alist)))
				(cond (formal-assoc
				       ;; variable
				       (let* ((op-argnum (cdr formal-assoc))
					     (all-sprops
					      (mapcar
					       #'(lambda (const)
						   (split-argnum-sprop op
								       op-argnum const :during))
					       (nth (1- op-argnum)
						    (if
							(eq *action-representation*
							    'simple-split-factored)
							(op-obj-lst op)
						      (op-obj-lst op))))))
					 (cons :or (remove
						    (split-argnum-sprop op
									op-argnum actual :during)
						    all-sprops))))
				      ;; constant
				      ((eq actual op-formal)
				       ;; two equal constants
				       (cnf-false))
				      (t
				       (assert (not (varp op-formal)))
				       ;; two unequal constants
				       (cnf-true)))))
			  fluent-args
			  (sprop-args occurrence)))))))
	  fluent-occurrences))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Explanatory frame axioms
;;;

(defun explanatory-frame-clauses ()
  (case *action-representation*
    ((regular simple-split overloaded-split bitwise)
     (explanatory-unfactored-frame-clauses))
    ((simple-split-factored overloaded-split-factored)
     (explanatory-factored-frame-clauses))
    (otherwise
     (error "What action representation? ~s" *action-representation*))))


;; gets explanatory frame axioms
;; uses the global variables *ground-actions* and *ground-effects*
(defun explanatory-unfactored-frame-clauses ()
  (let* ((act-prop-list
	  (get-act-prop-list *ground-actions* *ground-effects* t))
	 (clauses (get-explanatory-frame-clauses act-prop-list)))
    (values (eliminate-instantiated-non-removable-types-clauses-sprop clauses)
	    nil)))

;; For each element in the effect set, list actions that have it as
;; a postcond.
;; ACT-PROP-LIST is a list of two-element lists of (ACTION-NAME PROP-LIST)
;;   where PROP-LIST is a list of effects (possibly negated fluents).
(defun act-prop-list->effect-act-set (act-prop-list)
  (let ((result '()))
    (dolist (act-prop act-prop-list)
      (let ((act (first act-prop)))
	(dolist (effect (second act-prop))
	  (declare (type sprop effect))
	  (assert (sprop-p effect))
	  (let* ((name effect)
		 (assoc (assoc name result)))
	    (if assoc
		(push act (cdr assoc))
	      (push (cons name (list act))
		    result)))
;;	    (assoc-push effect act result)
	  )))
    result))

;; for every proposition that can be changed, outputs
;; '(implies ((:and p1 (:not p1)) (the set of actions that can change p1 this way)))
;; eg
;; ((implies (:and (:not (prop1 :BEFORE)) (prop1 :AFTER)) (act1 :DURING))
;;  (implies (:and (:not (prop2 :BEFORE)) (prop3 :AFTER)) (act3 :DURING))
;;  (implies (:and (:not (prop3 :BEFORE)) (prop3 :AFTER))
;;           (or (act2 :DURING) (act3 :DURING)))
;;  (implies (:and (prop1 :BEFORE) (:not (prop1 :AFTER))) (act2 :DURING)))

;; ACT-PROP-LIST is a list of two-element lists of (ACTION-NAME PROP-LIST)
;;   where PROP-LIST is a list of effects (possibly negated fluents).
(defun get-explanatory-frame-clauses (act-prop-list)
  (flet ((before-or-not-after (fluent)
           (list :or
		  (sprop-add-time fluent ':before)
		  (sprop-negate (sprop-add-time fluent ':after)))))
    (let* ((effect-act-set (act-prop-list->effect-act-set act-prop-list))
	   (clauses (mapcar #'(lambda (prop-acts)
				(let ((prop (car prop-acts))
				      (act-list (cdr prop-acts)))
				  (append
				   (before-or-not-after prop)
				   (mapcar #'(lambda (a)
					       (sprop-add-time a ':during))
					   act-list))))
			    effect-act-set))
	   ;; The fluents (and their negations) that don't appear in any
	   ;; action.  For instance, if no action makes FOO true, then it can
	   ;; never be the case that (and (not (foo :before)) (foo :after))
	   ;; It would work to put this in with the above, since (or)
	   ;; reduces to false, but separate it for clarity.
	   ;; [Actually, it's clearer to merge it in, I think; do so, if it is.]
	   (never-clauses
	    (let ((result '()))
	      (dolist (fluent *initial-state*)
		(let ((unneg-fluent (sprop-remove-time (sprop-nonnegated fluent))))
		  (if (or
		       ;; This test is necessary because *{partial-}types* seem
		       ;; to get set even when not inferring types.
		       (eq *types-strategy* 'no-types)
		       (and (not (member (sprop-name unneg-fluent) *unary-types* :test #'eq))
			    (not (member (sprop-name unneg-fluent) *non-unary-types* :test #'eq))))
		      ;; Need to add frame axioms, because either we're not
		      ;; inferring types or this isn't a type.
		      (let ((neg-fluent (sprop-negate unneg-fluent)))
			(if (not (assoc unneg-fluent effect-act-set :test #'sprop-equal))
			    (push (before-or-not-after unneg-fluent)
				  result))
			(if (not (assoc neg-fluent effect-act-set :test #'sprop-equal))
			    (push (before-or-not-after neg-fluent)
				  result))))))
	      result)))
      (nconc clauses never-clauses))))


;; FACTORED EXPLANATORY
(defun explanatory-factored-frame-clauses ()
  (values
   (let ((op-frame-clauses (explanatory-factored-frame-clauses-for-ops
				   (domain-operators *domain*))))
     (append (true-and-false-fluent-frames) op-frame-clauses))
   '()))


(defun explanatory-factored-frame-clauses-for-ops (ops)
  (let* ((op-stuff (mapcar #'explanatory-factored-frame-clauses-for-op
			   ops))
	 (combine-fluent-lsts (apply #'mapcar #'list op-stuff))
	 (axiom-rhs (mapcar #'(lambda (flu)
				(apply #'mapcar #'list flu))
			    combine-fluent-lsts))
	 (no-ands (mapcar #'(lambda (flu)
			      (mapcar #'(lambda (rhs)
					  (cons :or
						rhs))
				      flu))
			  axiom-rhs))
	 (fluents (mapcan #'(lambda (flu)
			      (let ((name (car flu)))
				(mapcar #'(lambda (args)
					    (cons name args))
					(cdr flu))))
			  *variant-fluents*))
	 (num-flus-minus-1 (- (length fluents) 1))
	 )
    ;; (format t "op-stuff: ~s~%" op-stuff)
    ;; (format t "axiom-rhs: ~s~%" axiom-rhs)
    ;; (format t "no-ands: ~s~%" no-ands)
    ;; (format t "fluents: ~s~%" fluents)
    (apply #'append
	   (loop for i from 0 to num-flus-minus-1
	       collect
		 (let* ((cur-flu (nth i fluents))
			(cur-flu-name (car cur-flu))
			(cur-flu-args (cdr cur-flu))
			(pos-before (make-sprop cur-flu-name cur-flu-args :before nil))
			(pos-after (make-sprop cur-flu-name cur-flu-args :after nil))
			(neg-before (sprop-negate pos-before))
			(neg-after (sprop-negate pos-after))
			(cur-rhs (nth i no-ands))
			(pos
			 (cnf-fix-sprop-negation
			  (convert-to-cnf
			   (list :or pos-before neg-after (first cur-rhs)))))
			(neg
			 (cnf-fix-sprop-negation
			  (convert-to-cnf
			   (list :or neg-before pos-after (second cur-rhs))))))
		   (append (conjunctive-clauses pos)
			   (conjunctive-clauses neg)))))))


(defun explanatory-factored-frame-clauses-for-op (op)
  (let* ((actionis (make-actionis op))
	 ;; For each mentioned fluent, separately collect the effect conjuncts
	 ;;   that mention it positively and those that mention it negatively.
	 ;; fluent-alist is a list of 4-element lists of
	 ;;   (name pos-occurrences neg-occurrences instantiation-arglists).
	 (fluent-alist
	  (let ((alist (mapcar #'(lambda (finfo)
				   (list (car finfo) (list) (list) (cdr finfo)))
			       *variant-fluents*)))
	    (dolist (effect-as-list (conjunctive-clauses (op-effect op)))
	      (let* ((effect (make-sprop-from-list-or-negation effect-as-list))
		     (fluent-assoc (assoc (sprop-name effect) alist)))
		(assert fluent-assoc)
		(assert (keyword-or-number-p (car (last (if (sprop-negated-p effect) (second effect-as-list) effect-as-list)))))
		(if (sprop-negated-p effect)
		    (push effect (third fluent-assoc))
		  (push effect (second fluent-assoc)))))
	    alist))
	 ;; The final result
	 (axioms '()))
    (dolist (fluent+pos+neg+inst fluent-alist)
      ;; If it doesn't appear negatively, the positive case is as if
      ;; the fluent didn't appear at all.  (This is very slightly
      ;; redundant, but has a smaller representation.)
      (let (;; (fluent-name (car fluent+pos+neg+inst))
	    (pos-occurrences (second fluent+pos+neg+inst))
	    (neg-occurrences (third fluent+pos+neg+inst))
	    (instantiation-arglists (fourth fluent+pos+neg+inst))
	    (var-argnum-alist (let ((argnum 0))
				(mapcar #'(lambda (argname)
					    (incf argnum)
					    (cons argname argnum))
					;; might include time; no harm to do so
					(progn
					  (assert (keyword-or-number-p (op-time op)))
					  (op-formals op))))))
	(assert (every #'sprop-p pos-occurrences))
	(assert (every #'sprop-p neg-occurrences))
	(assert (notany #'keywordp (mapcar #'car var-argnum-alist)))
	(dolist (fluent-args instantiation-arglists)
	  ;; Cope with positive occurrences, which leads to
	  ;;   not(fluent) and actionis(A) => not(fluent)
	  ;; if there are no such occurrences.
	  (setf axioms
	    (append axioms
		    (list
		     (list
		      (if (null pos-occurrences)
			  '(:or)
			(cons :and
			      (cons
			       (matching-var-sprops
				fluent-args pos-occurrences
				op var-argnum-alist)
			       (if (and
				    (eq *action-representation*
					'simple-split-factored)
				    (not (null fluent-args)))
				   '()
				 (list actionis)))))
		      (if (null neg-occurrences)
			  '(:or)
			(cons :and
			      (cons
			       (matching-var-sprops
				fluent-args neg-occurrences
				op var-argnum-alist)
			       (if (and
				    (eq *action-representation*
					'simple-split-factored)
				    (not (null fluent-args)))
				   '()
				 (list actionis)))))))))
	  )))
    axioms))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Utilities
;;;

;; returns the action-proposition list.
;; It is a list of two-element lists of (GROUNDED-ACTION CHANGED-PROP-LIST)
;;  where CHANGED-PROP-LIST is the grounded props changed by GROUNDED-ACTION
;; [by changed, do we mean either positively or negatively?]
;; LEAVE-NOTS is t if affected clauses, nil if classical clauses.
(defun get-act-prop-list (ground-acts ground-effs leave-nots)
  (mapcar #'(lambda (act effect-expr)
	      (list (sprop-remove-time act)
		    (if (sprop-p effect-expr)
			(list (sprop-remove-time
			       (if (and (not leave-nots)
					(sprop-negated-p effect-expr))
				   (sprop-negate effect-expr)
				 effect-expr)))
		      (if leave-nots
			  (flatten-expr-leave-nots effect-expr)
			(flatten-expr-remove-nots effect-expr)))))
	  ground-acts ground-effs))

;;; Should merge these two

;; takes a boolean expression, returns a list of the literals in that expr.
(defun flatten-expr-remove-nots (expr)
  (if (sprop-p expr)
      (list (sprop-remove-time (sprop-nonnegated expr)))
    (let ((fn (car expr)))
      (cond
       ((eq fn :not) (flatten-expr-remove-nots (second expr)))
       ((or (eq fn :and)
	    (eq fn :or) ;; not really needed
	    )
	(mapcan #'flatten-expr-remove-nots (cdr expr)))
       (t
	(error "What expr? ~s" expr))))))

;; returns a list of possibly negated atomic predicates
;; just like flatten-expr, but leaves "not"s
(defun flatten-expr-leave-nots (expr)
  (if (sprop-p expr)
      (list (sprop-remove-time expr))
    (cond
     ((or (eq (car expr) :and)
	  (eq (car expr) :or)      ;; not really needed
	  )
      (mapcan #'flatten-expr-leave-nots (cdr expr)))
     ;; No need to check for not?
     (t
      (error "What expr? ~s" expr)))))
