;;; main.lisp: load other files

;; To use this file:
;;   (load "main") (in-package medic) (load-all) (load-domains)
;; To find errors and reduce runtime, do (compile-all)
;; If compiled files give trouble, do (delete-all-compiled)

(unless (cl:find-package "DOMAINS")
  (cl:defpackage "DOMAINS" (:use :common-lisp :cl-user)))
(unless (cl:find-package "MEDIC")
  (cl:defpackage "MEDIC" (:use :common-lisp :cl-user :domains)
	    ;; This is a hack to permit reuse of the existing VARIABLE package
	    (:nicknames "VARIABLE")))

(in-package medic)

;; For the time being, remove CLIM features unconditionally.
;;(setq *features* (nset-difference *features* '(:clim :clim-2 :clim-2.0 :clim-motif)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User parameters
;;;

;; can't do merge-pathnames on result of (pathname-directory ...).
(defvar *medic-code-directory* (directory-namestring *load-truename*)
  "Directory containing Medic code (Lisp source).")

;; (setq *print-circle* t)

(defvar *cnf-converter* '->cnf
  "Either 'form-cnf or '->cnf; used by convert-to-cnf.
A third option for CNF-conversion on UCPOP-style well-formed formulae
is Marc Friedman's (never finished) /projects/ai/planners/fabian/logic.lisp.")

(defvar *domains-directories*
    '("/projects/ai/planners/domains/"
      "/projects/ai/planners/domains/medic-tests/")
  "List of directories containing domain/problem files.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Global variables
;;;

;; debugging
(defvar *show-stages* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Load non-medics utilities
;;;

;; "domains" must precede "verify" (and probably everything else) here

(defvar external-basenames '("domains" "verify" "clause" "normal" "utilities"))
(defvar external-absolute-basenames
    (mapcar #'(lambda (base) (merge-pathnames base *medic-code-directory*))
	    external-basenames))

(if (not (fboundp 'domains::print-domain))
    (load (merge-pathnames "domains" *medic-code-directory*)))

(if (not (fboundp 'verify-clean))
    (load (merge-pathnames "verify" *medic-code-directory*)))
(setq *verify-verbosely* nil)

(if (not (fboundp '->cnf))
    (progn
      (load (merge-pathnames "utilities" *medic-code-directory*)) ; for normal.lisp
      (load (merge-pathnames "normal" *medic-code-directory*))
      ))
(if (not (fboundp 'form-cnf))
    (load (merge-pathnames "clause" *medic-code-directory*)))

(defun convert-to-cnf (x)
  (let ((result (cond ((eq *cnf-converter* 'form-cnf) (form-cnf x))
		      ((eq *cnf-converter* '->cnf) (->cnf x))
		      (t (error "Invalid value for *cnf-converter*")))))
    ;; Don't return a formula containing 'true, 'false, t, or nil.
    (cond ((cnf-true-p result) (cnf-true))
	  ((cnf-false-p result) (cnf-false))
	  (t result))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Load domains and problems
;;;

;; (in-package cl-user)
;; 
;; (export 'load-domains)

;; Load from director(y/ies) DIRS (default *domains-directories*) the files
;; named in the list DOMAINS, which may be either symbols or strings.
;; DIRS may be either a single directory or a list of directories.
;; If DOMAINS is omitted, all the domains in the directory are loaded.
(defun load-domains (&key (dirs *domains-directories*) domains verbose)
  (if (not (listp dirs))
      (setq dirs (list dirs)))
  ;; Avoid errors incurred by specifying "foo/bar" instead of "foo/bar/"
  (assert (notany #'pathname-name dirs))

  (let ((domain-pathnames
	 (loop for dir in dirs
	     append (if domains
			(loop for domain in domains
			    collect (merge-pathnames (if (symbolp domain)
							 (symbol-name domain)
						       domain)
						     dir))
		      (directory (merge-pathnames dir "*.lisp"))))))
    (dolist (domain-pathname domain-pathnames)
      (if verbose
	  (format t "~&; Loading ~A" (pathname-name domain-pathname)))
      (load domain-pathname :verbose verbose))))

;; (in-package medic)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Medics files
;;;

;; does not include domains, verify, or CNF converters
(defvar *medic-file-basenames*)
;; setq so reloading this file causes changes to have effect
(setq *medic-file-basenames*
  '("main"
    "util"
    "prop"
    "convert"
    "frames"
    "action-repr"
    "partial-action"
    "exclusion"
    "driver"
    "substitute"
    "file-io"
    "dpll"
    "test"))

(defconstant *fasl-type-path* (make-pathname :type
					     #+:allegro excl:*fasl-default-type*
					     #+:cmu "sparcf"
					     ))
(defconstant *lisp-type-path* (make-pathname :type "lisp"))

(defvar *medic-file-absolute-basenames*)
;; setq so reloading this file causes changes to have effect
(setq *medic-file-absolute-basenames*
  (mapcar #'(lambda (base) (merge-pathnames base *medic-code-directory*))
	  *medic-file-basenames*))

(defun load-all (&optional uncompiled)
  (dolist (basename *medic-file-absolute-basenames*)
    (load-carefully basename uncompiled)))

(defun load-carefully (file-basename &optional uncompiled)
  (let ((uncompiled-file (merge-pathnames *lisp-type-path* file-basename)))
    (if uncompiled
	(load uncompiled-file)
      (let* ((fasl (probe-file (merge-pathnames *fasl-type-path* file-basename)))
	     (lisp (probe-file uncompiled-file))
	     (fasl-date (and fasl (file-write-date fasl)))
	     (lisp-date (and lisp (file-write-date lisp))))
	(declare (type (or unsigned-byte (member nil)) fasl-date))
	(declare (type (or unsigned-byte (member nil)) lisp-date))
	(if (or (not fasl)
		(and lisp-date fasl-date (> lisp-date fasl-date))
		(with-open-file (istream fasl)
		  (zerop (file-length istream))))
	    (load uncompiled-file)
	  (load file-basename))))))

(defun compile-file-maybe (basename &optional force)
  (if (or force
	  (let ((fasl (merge-pathnames *fasl-type-path* basename))
		(lisp (merge-pathnames *lisp-type-path* basename)))
	    (or (not (probe-file fasl))
		(not (probe-file lisp))
		(> (file-write-date lisp) (file-write-date fasl))
		(with-open-file (istream fasl)
		  (zerop (file-length istream))))))
      (compile-file basename)))

(defun compile-all (&optional force)
  ;; call (delete-all-compiled) first?
  (dolist (basename *medic-file-absolute-basenames*)
    (compile-file-maybe basename force))
  ;; Warning!  load-all doesn't get new versions of these files;
  ;; you need to reload manually or restart Lisp.
  (dolist (basename external-absolute-basenames)
    (compile-file-maybe basename force)))
;; (compile-all)

(defun delete-all-compiled (&optional only-medics-p)
  (dolist (basename (if only-medics-p
			*medic-file-absolute-basenames*
		      (append external-absolute-basenames
			      *medic-file-absolute-basenames*)))
    (let ((fname (merge-pathnames (make-pathname :type "sfasl42") basename)))
      (if (probe-file fname) (delete-file fname)))
    (let ((fname (merge-pathnames (make-pathname :type "gfasl42") basename)))
      (if (probe-file fname) (delete-file fname)))
    (let ((fname (merge-pathnames (make-pathname :type "sparcf") basename)))
      (if (probe-file fname) (delete-file fname)))))
