;;; -*- Mode: Lisp; Syntax: Common-Lisp; -*- File: logic/normal.lisp

;; From /projects/ai/573/code/logic/normal.lisp, with only the following
;; additions.
;; (in-package 'logic-normal)
;; (use-package 'common-lisp-user)
;; (export '(->cnf ->inf ->horn logic))
;; (in-package 'common-lisp-user)
(in-package medic)

;; From aima.lisp:
(defun mklist (x)
  "If x is a list, return it; otherwise return a singleton list, (x)."
  (if (listp x) x (list x)))

;; From unify.lisp, with the following renamings to avoid clash with verify:
;; new-variable => new-nvariable
;; variable? => nvariable?
;; get-binding => get-nbinding
;; binding-var => get-nbinding-var
;; binding-val => get-nbinding-val
;; make-binding => make-nbinding
;; lookup => lookup-nbinding
;; extend-bindings => extend-nbindings

(defvar +NO-BINDINGS+)
(defvar +FAIL+)
(defvar *new-variable-counter* 0)
(defun new-nvariable (var)
  "Create a new variable.  Assumes user never types variables of form ?X_9"
  (intern (format nil "~A~A_~D" (if (nvariable? var) "" "?")
		  var (incf *new-variable-counter*))))
(defun nvariable? (x)
  "Is x a variable (a symbol beginning with `?')?"
  (and (symbolp x) (equal (char (symbol-name x) 0) #\?)))
(defun get-nbinding (var bindings)
  "Find a (variable . value) pair in a binding list."
  (assoc var bindings))
;; (defun nbinding-var (binding)
;;   "Get the variable part of a single binding."
;;   (car binding))
(defun nbinding-val (binding)
  "Get the value part of a single binding."
  (cdr binding))
(defun make-nbinding (var val) (cons var val))
(defun lookup-nbinding (var bindings)
  "Get the value part (for var) from a binding list."
  (nbinding-val (get-nbinding var bindings)))
(defun extend-nbindings (var val bindings)
  "Add a (var . value) pair to a binding list."
  (cons (make-nbinding var val)
        ;; Once we add a "real" binding,
        ;; we can get rid of the dummy +no-bindings+
        (if (eq bindings +no-bindings+)
            nil
            bindings)))

;;; end of additions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;; Convert Expressions to Normal Form (Conjunctive, Implicative or Horn)

;;; This could be done much more efficiently.  Most often, a special
;;; representation is used for CNF, which eliminates the explicit ANDs
;;; and ORs.  This code is meant to be informative, not efficient.

;;;; Top-Level Functions

(defun ->cnf (p &optional vars)
  "Convert a sentence p to conjunctive normal form [p 279-280]."
  ;; That is, return (and (or ...) ...) where
  ;; each of the conjuncts has all literal disjuncts.
  ;; VARS is a list of universally quantified variables that P is in scope of.
  (setf p (eliminate-implications (logic p)))
  (case (op p)
    (:NOT (let ((p-2 (move-not-inwards (arg1 p))))
	   (if (literal-clause? p-2) p-2 (->cnf p-2 vars))))
    (:AND (conjunction (mappend #'(lambda (q) (conjuncts (->cnf q vars)))
				(args p))))
    (:OR  (merge-disjuncts (mapcar #'(lambda (q) (->cnf q vars))
				  (args p))))
    (:ALL (let ((new-vars (mapcar #'new-nvariable  (mklist (arg1 p)))))
	   (->cnf (sublis (mapcar #'cons  (mklist (arg1 p)) new-vars)
			  (arg2 p))
		  (append new-vars vars))))
    (:EXI (->cnf (skolemize (arg2 p) (arg1 p) vars) vars))
    (t   p) ; p is atomic
    ))

(defun ->inf (p)
  "Convert a sentence p to implicative normal form [p 282]."
  (conjunction (mapcar #'cnf1->inf1 (conjuncts (->cnf p)))))

(defun ->horn (p)
  "Try to convert sentence to a Horn clause, or a conjunction of Horn clauses.
  Signal an error if this cannot be done."
  (let ((q (->inf p)))
    (when (not (every #'horn-clause? (conjuncts q)))
      (warn "~A, converted to ~A, is not a Horn clause." p q))
    q))

;; This converts
(defun logic (sentence)
  "Canonicalize a sentence into proper logical form."
  (cond
   ;; Commented out by MDE.  This requires infix.lisp, which converts an
   ;; infix string representation into a Lisp-style s-expression.
   ;; ((stringp sentence) (string->prefix sentence))
   (t sentence)))

;;;; Auxiliary Functions

(defun cnf1->inf1 (p)
  ;; P is of the form (or (not a) (not b) ... c d ...)
  ;; Convert to: (=> (and a b ...) (or c d ...))
  ;; where a,b,c,d ... are positive atomic clauses
  (let ((lhs (mapcar #'arg1 (remove-if-not #'negative-clause? (disjuncts p))))
	(rhs (remove-if #'negative-clause? (disjuncts p))))
    `(=> ,(conjunction lhs) ,(disjunction rhs))))

(defun eliminate-implications (p)
  (if (literal-clause? p)
      p
    (case (op p)
      (=>  `(:or ,(arg2 p) (:not ,(arg1 p))))
      (<=> `(:or (:and ,(arg1 p) ,(arg2 p))
		(:and (:not ,(arg1 p)) (:not ,(arg2 p)))))
      (t   (cons (op p) (mapcar #'eliminate-implications (args p)))))))

(defun move-not-inwards (p)
  "Given P, return ~P, but with the negation moved as far in as possible."
  (case (op p)
    (:TRUE :false)
    (:FALSE :true)
    (:NOT (arg1 p))
    (:AND (disjunction (mapcar #'move-not-inwards (args p))))
    (:OR  (conjunction (mapcar #'move-not-inwards (args p))))
    (:ALL (make-exp 'EXI (arg1 p) (move-not-inwards (arg2 p))))
    (:EXI (make-exp 'ALL (arg1 p) (move-not-inwards (arg2 p))))
    (t (make-exp :not p))))

(defun merge-disjuncts (disjuncts)
  "Return a CNF expression for the disjunction."
  ;; The argument is a list of disjuncts, each in CNF.
  ;; The second argument is a list of conjuncts built so far.
  (case (length disjuncts)
    (0 :false)
    (1 (first disjuncts))
    (t (conjunction
	(with-collection ()
	  (for each y in (conjuncts (merge-disjuncts (rest disjuncts))) do
	       (for each x in (conjuncts (first disjuncts)) do
		    (collect (disjunction (append (disjuncts x)
						  (disjuncts y)))))))))))

(defun skolemize (p vars outside-vars)
  "Within the proposition P, replace each of VARS with a skolem constant,
  or if OUTSIDE-VARS is non-null, a skolem function of them."
  (sublis (mapcar #'(lambda (var)
		      (cons var (if (null outside-vars)
				    (skolem-constant var)
				  (cons (skolem-constant var) outside-vars))))
		  (mklist vars))
	  p))

(defun skolem-constant (name)
  "Return a unique skolem constant, a symbol starting with '$'."
  (intern (format nil "$~A_~D" name (incf *new-variable-counter*))))

(defun renaming? (p q &optional (bindings +no-bindings+))
  "Are p and q renamings of each other? (That is, expressions that differ
  only in variable names?)"
  (cond ((eq bindings +fail+) +fail+)
	((equal p q) bindings)
	((and (consp p) (consp q))
	 (renaming? (rest p) (rest q)
		    (renaming? (first p) (first q) bindings)))
	((not (and (nvariable? p) (nvariable? q)))
	 +fail+)
	;; P and Q are both variables from here on
	((and (not (get-nbinding p bindings)) (not (get-nbinding q bindings)))
	 (extend-nbindings p q bindings))
	((or (eq (lookup-nbinding p bindings) q) (eq p (lookup-nbinding q bindings)))
	 bindings)
	(t +fail+)))

;;;; Utility Predicates and Accessors

(defconstant +logical-connectives+ '(:and :or :not => <=>))
(defconstant +logical-quantifiers+ '(:all :exi))

(defun atomic-clause? (sentence)
  (not (or (member (op sentence) +logical-connectives+)
	   (member (op sentence) +logical-quantifiers+))))

(defun literal-clause? (sentence)
  (or (atomic-clause? sentence)
      (and (eq (op sentence) :not) (atomic-clause? (arg1 sentence)))))

(defun negative-clause? (sentence)
  (eq (op sentence) :not))

(defun conjunction? (sentence)
  (eq (op sentence) :and))

(defun horn-clause? (sentence)
  (and (eq (op sentence) '=>)
       (every #'atomic-clause? (conjuncts (arg1 sentence)))
       (atomic-clause? (arg2 sentence))))

(defun conjuncts (sentence)
  "Return a list of the conjuncts in this sentence."
  (cond ((eq (op sentence) :and) (args sentence))
	((eq sentence :true) nil)
	(t (list sentence))))

(defun disjuncts (sentence)
  "Return a list of the disjuncts in this sentence."
  (cond ((eq (op sentence) :or) (args sentence))
	((eq sentence :false) nil)
	(t (list sentence))))

(defun conjunction (args)
  "Form a conjunction with these args."
  (case (length args)
    (0 :true)
    (1 (first args))
    (t (cons :and args))))

(defun disjunction (args)
  "Form a disjunction with these args."
  (case (length args)
    (0 :false)
    (1 (first args))
    (t (cons :or args))))
