(in-package medic)


;; This file contains partial action elimination and at-least-one


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Eliminating Partial Actions (used when doing factored splitting)
;;
;; code for ensuring that there are no partial action assignments
;; Method 3 (see the e-mail about the three proposed methods)
;; This function takes an operator and returns the partial action clauses
;; for this operator.  It creates a new sprop called ACTION which has the
;; operator's name as a parameter.

(defun partial-action-elim-extra-vars (ops)
  (if (eq *action-representation* 'simple-split-factored)
      (mapcan #'partial-action-elim-simple-extra-vars ops)
    (partial-action-elim-overloaded-extra-vars ops)))

;; returns a list of disjunctions
(defun partial-action-elim-simple-extra-vars (op)
  (let* ((name (op-name op))
	 (new-var (make-actionat-sprop name :during))
	 (negated-new-var (sprop-negate new-var))
	 (n (op-arity op))
	 (splits (loop for i from 1 to n collect
		       (let ((cur-obj-lst (nth (- i 1) (op-obj-lst op))))
			 (mapcar #'(lambda (obj)
				     (split-argnum-sprop op i obj :during))
				 cur-obj-lst)))))
    ;; (format t "action-elim-extra-vars: ~s~%" (op-action op))
    (mapcan #'(lambda (s)
		(partial-action-elim-aux-extra-vars new-var negated-new-var s))
	    splits)))


;; returns a list of disjunctions
(defun partial-action-elim-overloaded-extra-vars (ops)
  (let* ((new-var (make-sprop 'overloaded () :during nil))
	 (negated-new-var (sprop-negate new-var))
	 (action-vars (mapcar #'(lambda (op)
				    (make-actionis op))
			      ops))
	 (n (max-arity ops))
	 (unioned-obj-lsts *unioned-obj-lsts*)
	 ;; list of lists; list i contains all possible sprops for arg i
	 (splits (loop for i from 1 to n collect
		       (let ((cur-obj-lst (nth (- i 1) unioned-obj-lsts)))
			 (mapcar #'(lambda (obj)
				     (overloaded-split-argnum-sprop
				      i obj :during))
				 cur-obj-lst))))
	 (all-sprops (cons action-vars splits)))
    (mapcan #'(lambda (s)
		(partial-action-elim-aux-extra-vars new-var negated-new-var s))
	    all-sprops)))

;; returns a list of disjunctions
;; if we're doing classical frames, we don't need the only-if direction,
;; and we further don't need the new OVERLOAD variable, since it's always true
(defun partial-action-elim-aux-extra-vars (new-var neg-new-var splits)
  (let ((if-direction (cons :or (if (or (eq *frame-strategy* 'explanatory)
					(eq *action-representation*
					    'simple-split-factored))
				    (cons neg-new-var splits)
				  splits)))
	(only-if-direction (if (or (eq *frame-strategy* 'explanatory)
				   (eq *action-representation*
					    'simple-split-factored))
			       (mapcar #'(lambda (s)
					   (list :or new-var (sprop-negate s)))
				       splits)
			     ())))
    (cons if-direction only-if-direction)))

;; code for ensuring that there are no partial action assignments
;; Method 2 (see the e-mail about the three proposed methods)
;; This function takes an operator and returns the partial action clauses
;; for this operator.

(defun partial-action-elim-ordinary (ops)
  (if (eq *action-representation* 'simple-split-factored)
      (mapcan #'partial-action-elim-simple-ordinary ops)
    (partial-action-elim-overloaded-ordinary ops)))


(defun partial-action-elim-simple-ordinary (op)
  (let* ((n (op-arity op))
	 (splits (loop for i from 1 to n
		       collect (let ((cur-obj-lst (nth (- i 1) (op-obj-lst op))))
				 (mapcar #'(lambda (o)
					     (split-argnum-sprop op i o :during))
					 cur-obj-lst)))))
    ;; Perhaps here add in clauses forbidding arguments that can't
    ;; appear here, as `partial-action-elim-overloaded-ordinary' does.
    (mapcan #'(lambda (s)
		(partial-action-elim-aux-ordinary s splits))
	    splits)))

(defun partial-action-elim-overloaded-ordinary (ops)
  (let* ((action-vars (mapcar #'(lambda (op)
				  (make-actionis op))
			      ops))
	 (n (max-arity ops))
	 (unioned-obj-lsts *unioned-obj-lsts*)
	 ;; List of (sprops for) all objects that can appear in each position.
	 (splits (loop for i from 1 to n
		       collect (let ((cur-obj-lst (nth (- i 1) unioned-obj-lsts)))
				 (mapcar #'(lambda (obj)
					     (overloaded-split-argnum-sprop
					      i obj :during))
					 cur-obj-lst))))
	 ;; These arguments can't appear in these positions.
	 ;; This is useful when *types-strategy* is not 'eliminate.
	 (forbidden-arguments
	  (loop for i from 1 to n
		nconc (let ((cur-obj-lst (nth (- i 1) unioned-obj-lsts)))
			  (mapcar #'(lambda (obj)
				      (sprop-negate
				       (overloaded-split-argnum-sprop
					i obj :during)))
				  (set-difference *objects* cur-obj-lst)))))
	 (all-sprops (cons action-vars splits)))
    (nconc forbidden-arguments
	   (mapcan #'(lambda (s)
		       (partial-action-elim-aux-ordinary s all-sprops))
		   all-sprops))))

(defun partial-action-elim-aux-ordinary (split splits)
  (mapcan #'(lambda (s)
	      (if (sprop-equal (car s) (car split))
		  '()
		;; (not sprop) or s  ==  sprop => s
		;; ie, if this action occurs, one of these is the argument.
		(mapcar #'(lambda (sprop)
			    (cons :or (cons (sprop-negate sprop) s)))
			split)))
	  splits))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; At-least-one clauses
;;;

;; this gives us the AT LEAST ONE clause when using factored splitting method
;; number 3
(defun factored-at-least-one-extra-vars (ops)
  (cons :or
	(mapcar #'(lambda (op)
		    (if (null (op-formals op))
			(make-sprop (op-name op) '() :during nil)
		      (make-actionat-sprop (op-name op) :during)))
		ops)))

(defun factored-at-least-one-ordinary (ops)
  (if (eq *action-representation* 'simple-split-factored)
      (cons :or (mapcan #'(lambda (op)
			    (if (null (op-formals op))
				(list (make-sprop (op-name op) () :during nil))
			      (mapcar #'(lambda (obj)
					  (split-argnum-sprop op 1
							      obj :during))
				      (car (op-obj-lst op)))))
			ops))
    (let* ((action-vars (mapcar #'(lambda (op)
				    (make-actionis op))
				ops)))
      (cons :or action-vars))))
