(in-package medic)

;; (This is a separate file mostly so that I can skip it completely when
;; doing replacement of sprop for lprop everywhere else in the code.)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Debugging variables
;;;

;; Debugging
(defvar *sprop-print-curly-p* nil
  "If non-nil, print sprops surrounded by curly braces, to emphasize they
are not mere lists.  If nil, print sprops surrounded by parentheses.")
(defvar *sprop-print-omit-not-p* nil
  "If non-nil, then printed sprops may have either 3 or 4 elements; if nil,
they always have 4 elements, the first of which is blank if non-negated.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Propositions
;;;

;; A proposition is either an action or a fluent, representated
;; identically; this is also known as a literal.  The proposition carries
;; information regarding whether it is negated or not.

;; Alternate representations:
;;  * lprop: list of (name arg1 arg2 ...), or (not (name arg1 arg2 ...)).
;;     I've added 'lprop to the beginning for testing.
;;  * sprop: struct of (name args time negated), indexed through hash table
;;     to avoid duplication; perhaps each one also has memoized pointer to its
;;     negation, its timeless version.


;; Needed for both, can't go at end of file.
(defun make-equal-hash-table ()
  (make-hash-table :test #'equal))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SPROP
;;;

;; Be careful with this function!  You get <unprintable object> if this errs.
(defun sprop-print-function (sprop stream depth)
  (declare (ignore depth))
  ;; (format stream "~a~a" (sprop->list sprop) (sxhash sprop))
  (if *sprop-print-curly-p*
      (if *sprop-print-omit-not-p*
	  (if (sprop-negated-p sprop)
	      (format stream "{~s ~s ~s}"
		      (sprop-name sprop) (sprop-args sprop) (sprop-time sprop))
	    (format stream "{~s ~s ~s ~s}"
		    :NOT (sprop-name sprop) (sprop-args sprop) (sprop-time sprop)))
	(format stream "{~a ~s ~s ~s}"
		(if (sprop-negated-p sprop) :NOT " ")
		(sprop-name sprop) (sprop-args sprop) (sprop-time sprop)))
    (if (sprop-negated-p sprop)
	(format stream "(not ~s)" (sprop->list (sprop-negate sprop)))
      (format stream "~s" (sprop->list sprop)))))


;; An action represented as a structure.
(defstruct (sprop
	    ;; This constructor doesn't check types.  Nor does it check the
	    ;; hashtable.  For these and other reasons, under no circumstances
	    ;; should any function other than make-sprop call it.
	    (:constructor make-sprop-internal (name args time negated-p))
	    (:print-function sprop-print-function)
	    )
  (name (retuired) :type symbol)
  (args (required) :type list)			; list of symbols
  (time (required) :type (or symbol fixnum))	; nil, keyword, or small int
  (negated-p (required) :type bool)
  (negation nil :type (or sprop (member nil))) ; nil, or negation of this sprop
  )

(defun make-sprop-hash-table ()
  (make-hash-table :test #'eq))

;; Clear this out for each new problem??  Probably, to avoid bloat, but that
;; forces us to recreate some sprops, maybe.

;; This is an equal hash table, not an eq hash table, because it maps from
;; lists, not from sprops.  Use make-sprop-hash-table only for mapping from
;; sprops.
(defvar *sprop-hash-table* (make-equal-hash-table))
(defvar *sprop-args-hash-table* (make-hash-table :test #'equal))

(defun canonicalize-sprop-args (args)
  (and args
       (let ((hash-result (gethash args *sprop-args-hash-table*)))
	 (or hash-result
	     (and (setf (gethash args *sprop-args-hash-table*) args)
		  args)))))


;; The time argument, if any, appears last.
(defun make-sprop (name args time negated-p)
  (declare (symbol name) (list args)
	   (type (or keyword-or-number (member nil)) time))
  (assert (symbolp name))
  (assert (not (null name)))
  (assert (not (eq 'sprop name)))
  (assert (listp args))
  (assert (every #'symbol-or-number-p args))
  (assert (notany #'keywordp args))
  ;; Disable for creation of non-time-neutral formulae?  Probably not.
  (assert (or (not time) (keyword-or-number-p time)))
  ;; No longer needed, since *sprop-hash-table* uses #'equal.
  ;; (setq args (canonicalize-sprop-args args))
  (let* ((hash-key (list name args time negated-p))
	 (prop (gethash hash-key *sprop-hash-table*)))
    ;; (format t "~&hash result ~s~%" prop)
    (or prop
	(let ((prop (make-sprop-internal name args time negated-p)))
	  (setf (gethash hash-key *sprop-hash-table*) prop)
	  prop))))

(defun make-sprop-from-list (x negated-p)
  (multiple-value-bind (name args time) (parse-prop-list x)
    (make-sprop name args time negated-p)))

(defun make-sprop-from-list-or-negation (x)
  (let ((negated-p (eq :not (car x))))
    (make-sprop-from-list (if negated-p (second x) x) negated-p)))

(defun make-nonnegated-sprop-from-list (x)
  (make-sprop-from-list x nil))

(defun make-timeless-sprop-from-list (x negated-p)
  (multiple-value-bind (name args time) (parse-prop-list x)
    (declare (ignore time))
    (make-sprop name args nil negated-p)))

(defun make-timeless-sprop-from-list-or-negation (x)
  (let ((negated-p (eq :not (car x))))
    (make-timeless-sprop-from-list (if negated-p (second x) x) negated-p)))

(defun make-timeless-nonnegated-sprop-from-list (x)
  (make-timeless-sprop-from-list x nil))


(defun sprop->list (sprop)
  ;; arbitrary restriction
  (assert (not (sprop-negated-p sprop)))
  (let ((sans-time (cons (sprop-name sprop) (sprop-args sprop))))
    (if (sprop-time sprop)
	(append sans-time (list (sprop-time sprop)))
      sans-time)))

(defun sprop->list-no-time (sprop)
  ;; arbitrary restriction
  (assert (not (sprop-negated-p sprop)))
  (cons (sprop-name sprop) (sprop-args sprop)))

;; Predicates

;; Could change to eq depending on representation of propositions
(defun sprop-equal (p1 p2)
  (declare (type sprop p1) (type sprop p2))
  (assert (sprop-p p1))
  (assert (sprop-p p2))
  (eq p1 p2))

;; Does no check of types
(defun equal-like-sprop (p1 p2)
  (eq p1 p2))

(defun sprop-equal-ignore-time (p1 p2)
  (declare (type sprop p1) (type sprop p2))
  (assert (sprop-p p1))
  (assert (sprop-p p2))
  (eq p1 p2))
;; Perhaps this is more efficient?
;; (setf (symbol-function sprop-equal) #'eq)

;; Modifiers

;; Don't want a setfmethod because don't want to side-effect the object.
(defun sprop-remove-time (sprop)
  (make-sprop (sprop-name sprop) (sprop-args sprop) nil (sprop-negated-p sprop)))

(defun sprop-add-time (sprop time)
  (declare (type sprop sprop))
  (declare (type keyword-or-number time))
  (assert (keyword-or-number-p time))
  (make-sprop (sprop-name sprop) (sprop-args sprop) time (sprop-negated-p sprop)))

(defun sprop-negate (sprop)
  (or (sprop-negation sprop)
      (let ((negated (make-sprop (sprop-name sprop) (sprop-args sprop) (sprop-time sprop) (not (sprop-negated-p sprop)))))
	(setf (sprop-negation sprop) negated)
	(setf (sprop-negation negated) sprop)
	negated)))

(defun sprop-nonnegated (sprop)
  (if (sprop-negated-p sprop)
      (sprop-negate sprop)
    sprop))


;;; Representation-independent; for simplicity, just copy

;; Don't call on timeless propositions; no check, for efficiency.
(defun sprop-time-lessp (p1 p2)
  (< (sprop-time p1) (sprop-time p2)))

(defun spropify-formula (f)
  (assert (not (sprop-p f)))
  (let ((fn (car f)))
    (cond ((eq :not fn)
	   (assert (length-one (cdr f)))
	   (let ((negated (spropify-formula (second f))))
	     (if (sprop-p negated)
		 (sprop-negate negated)
	       (list :not negated))))
	  ((formula-combiner-p fn)
	   (cons fn (mapcar #'spropify-formula (cdr f))))
	  ((prop-as-list-p f)
	   (make-sprop-from-list f nil))
	  (t (error "What formula? ~s" f)))))

(defun spropify-formulas (l)
  (mapcar #'spropify-formula l))

(defun timeless-spropify-formula (f)
  (assert (not (sprop-p f)))
  (let ((fn (car f)))
    (cond ((eq :not fn)
	   (assert (length-one (cdr f)))
	   (let ((negated (timeless-spropify-formula (second f))))
	     (if (sprop-p negated)
		 (sprop-negate negated)
	       (list :not negated))))
	  ((formula-combiner-p fn)
	   (cons fn (mapcar #'timeless-spropify-formula (cdr f))))
	  ((prop-as-list-p f)
	   (make-timeless-sprop-from-list f nil))
	  (t (error "What formula? ~s" f)))))

(defun spropified-formula-p (f)
  (cond ((sprop-p f)
	 t)
	((symbolp f)
	 (error "non-list formula? ~s" f))
	((and (eq :not (car f)) (sprop-p (second f)))
	 nil)
	((or (eq :and (car f)) (eq :or (car f)))
	 (every #'spropified-formula-p (cdr f)))
	(t
	 nil)))


(defun clause-contains-duplicate-sprop (arg)
  (assert (or (sprop-p arg) (formula-combiner-p (car arg))))
  (and (not (sprop-p arg))
       (contains-duplicate (cdr arg) :test #'sprop-equal)))

(defun clause-contains-sprop-and-negation (arg)
  (assert (or (sprop-p arg) (formula-combiner-p (car arg))))
  (and (not (sprop-p arg))
       (contains-duplicate (cdr arg) :test #'sprop-negation-p)))

;; Don't do this; remove the clause, as it's vacuously true.
;; (defun clause-remove-irrelevant-sprops (arg)
;;   ;; This is the right implementation if testing for duplicates is cheaper
;;   ;; than figuring out which elements are duplicates; most arguments should
;;   ;; be short anyway.
;;   (if (clause-contains-sprop-and-negation arg)
;;       (let ((irrelevant (duplicates (cdr arg) :test #'sprop-negation-p)))
;; 	(cons (car arg)
;; 	      (set-difference (cdr arg) irrelevant :test #'sprop-equal)))
;;     arg))


(defun sprop-negation-p (sprop1 sprop2)
  (sprop-equal sprop1 (sprop-negate sprop2)))

(defun intersection-sprop-equal (arg1 arg2)
  (intersection arg1 arg2 :test #'sprop-equal))

(defvar sprop-hash-function #'eq)


;;; These leave :NONE, because it's inconvenient to remove it using nsubst,
;;; and it does no harm.
;; Takes a step number (a positive integer), returns an instantiated step
;; which uses our odd-action, even-state convention.
;; Perhaps add :PREVIOUS-STEP, :NEXT-STEP.
;; Or add arbitrary math, but I think that's going too far: don't need it,
;;   it's error-prone, and it substantially complicates the implementation.
(defun sprop-step-instantiate (formula stepno)
  (if (sprop-p formula)
      (case (sprop-time formula)
	(:AFTER (make-sprop (sprop-name formula) (sprop-args formula)
			    (* 2 stepno) (sprop-negated-p formula)))
	(:DURING (make-sprop (sprop-name formula) (sprop-args formula)
			     (- (* 2 stepno) 1) (sprop-negated-p formula)))
	(:BEFORE (make-sprop (sprop-name formula) (sprop-args formula)
			     (- (* 2 stepno) 2) (sprop-negated-p formula)))
	(otherwise formula))
    (cons (car formula)
	  (mapcar #'(lambda (x) (sprop-step-instantiate x stepno))
		  (cdr formula)))))
(defun sprop-goal-instantiate (formula max-stepno)
  (if (sprop-p formula)
      (if (eq :END (sprop-time formula))
	  (make-sprop (sprop-name formula) (sprop-args formula)
		      (* 2 max-stepno) (sprop-negated-p formula))
	formula)
    (cons (car formula)
	  (mapcar #'(lambda (x) (sprop-goal-instantiate x max-stepno))
		  (cdr formula)))))
(defun sprop-start-state-instantiate (formula)
  (if (sprop-p formula)
      (if (eq :BEGINNING (sprop-time formula))
	  (make-sprop (sprop-name formula) (sprop-args formula)
		      0 (sprop-negated-p formula))
	formula)
    (cons (car formula)
	  (mapcar #'sprop-start-state-instantiate (cdr formula)))))

;; Works for either simple or overloaded splitting; works for unary operators.
(defun split-argnum-sprop (op argnum arg time)
  (declare (type operator op) (fixnum argnum))
  (assert (operator-p op))
  (let ((arity (op-arity op)))
    (assert (plusp arity))
    (assert (plusp argnum))
    (assert (<= argnum arity))
    (assert (member arg (nth (1- argnum) (op-obj-lst op))))
    (case *action-representation*
      ((simple-split-factored)
       (if (= 1 arity)
	   (make-sprop (op-name op) (list arg) time nil)
	 (simple-split-argnum-sprop op argnum arg time)))
      ((overloaded-split-factored)
       (if (and (= 1 arity) (not (factored-splitting-p)))
	   (make-sprop (op-name op) (list arg) time nil)
	 (overloaded-split-argnum-sprop argnum arg time)))
      ((regular simple-split overloaded-split bitwise)
       (error "not doing splitting; oughtn't call split-argnum-sprop"))
      (otherwise
       (error "what *action-representation* ~s" *action-representation*)))))

;; Use this for simple operator splitting.
(defun simple-split-argnum-sprop (op argnum arg time)
  (declare (type operator op) (fixnum argnum))
  (assert (operator-p op))
  (assert (plusp (op-arity op)))
  (assert (plusp argnum))
  (assert (<= argnum (op-arity op)))
  (assert (member arg (nth (1- argnum) (op-obj-lst op))))
  (make-sprop (if (= 1 (op-arity op))
		  (op-name op)
		(simple-split-argnum-name (op-name op) argnum))
	      (list arg) time nil))

;; Use these for overloaded operator splitting.
(defun overloaded-split-argnum-sprop (argnum arg time)
  (make-sprop (overloaded-argnum-name argnum) (list arg) time nil))
(defun make-actionat-sprop (name time)
  (make-sprop 'action (list name) time nil))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Representation-independent
;;;

;; I don't expect there to be much here.
(defun prop-as-list-p (x)
  (and (listp x)
       (every #'symbol-or-number-p x)))

;; Global type declaration
(declaim (ftype (function (list) (values symbol list (or keyword fixnum (member nil))))
		parse-prop-list))

(defun parse-prop-list (x)
  (declare (list x))
  (let ((time (let ((last-arg (car (last x))))
		(and (or (keywordp last-arg) (numberp last-arg))
		     last-arg))))
    (values (car x) (if time (butlast (cdr x)) (cdr x)) time)))

