(in-package medic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Debugging variables
;;;

(defvar *debug-substitute* nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Perform substition
;;;

(defun substitution-p (s)
  (and (every #'sprop-p s)
       (cdr s)				; length >= 1
       (not (sprop-negated-p (car s)))))
(defun substitutions-p (s)
  (every #'substitution-p s))

;; Return the formula with the substitutions applied.
;; Does no recursive substitutions.
;; Works by side effect.
(defun prop-substitute (clauses substitutions)
  ;; (setq prev-formula (copy-tree formula))  ; ouch, expensive copy
  (assert (substitutions-p substitutions))
  (if *debug-substitute* (format t "~&Entered prop-substitute~%"))
  (dolist (sub substitutions)
    (prop-substitute-one clauses sub))
  (if *debug-substitute*
      (format t "~&Distributing ors over ands (~d).~%" (length clauses)))
  ;; this is extremely slow.  Try pushing onto a global instead.
  ;; (mapcan #'distribute-ors-over-ands clauses)
  (let ((result '()))
    (dolist (clause clauses)
      ;; (format t "dooa=~d " (if (sprop-p clause) 1 (length clause)))
      ;; Don't side-effect result of d-o-o-a; might share list structure.
      (dolist (new-clause (distribute-ors-over-ands clause))
	(push new-clause result)))
    ;; (format t "~%")
    result))

;; ;; A simple one-pass algorithm won't work, since substitutions may insert
;; ;; something that a previous pass eliminated, so either
;; ;;  * iterate until no more changes, or
;; ;;  * substitute into both the formula and the substitution right-hand-sides.
;; ;; This implimentation uses the first strategy
;; (defun prop-substitute (formula substitutions)
;;   (if (null substitutions)
;;       formula
;;     (let ((prev-formula nil))
;;       (loop
;; 	  while
;; 	    (not (equal prev-formula formula)) ; ouch, expensive test
;; 	  do
;; 	    (setq prev-formula (copy-tree formula))  ; ouch, expensive copy
;; 	    (dolist (sub substitutions)
;; 	      (prop-substitute-one formula sub)))
;;       prev-formula)))


;; Modify CLAUSES by side effect.  That means that you shouldn't try to
;; substitute for the top level.
;; SUBSTITUTION is (from . list-of-to-conjuncts)
(defun prop-substitute-one (clauses substitution)
  (assert (clauses-p clauses))
  (let ((from (car substitution))
	;; copy, because inserted literally and might be side-effected later
	(to-positive (if (length-one substitution)
			 (first substitution)
		       (cons :and (copy-tree (cdr substitution)))))
	(to-negative (if (length-one substitution)
			 (sprop-negate (first substitution))
		       (cons :or (mapcar #'sprop-negate (cdr substitution))))))
    ;; Can't use #'sprop-equal: substituting into tree structure.
    ;; (We know it's only one level deep, but still...)
    (nsubst to-positive from clauses :test #'equal-like-sprop)
    (nsubst to-negative (sprop-negate from) clauses :test #'equal-like-sprop)))

;;; Do I need to worry about ordering constraints when two clauses-functions
;;; are adding to this?  Probably just iterate, as for prop-substitute.
;; Given a list of propositions, count which ones are true when.
;; If time-steps is 7, the possible times are 1, 3, 5, 7, 9, 11, 13.
;; We need not worry about negation, because we're interested in discovering
;; what is true, not what is false.
(defun prop-unsubstitute (propositions substitutions time-steps)
  (if (null substitutions)
      propositions
    (let* ((all-times (loop for i from 1 to (1- (* 2 time-steps)) by 2
			  collect i))
	   (added-props-when
	    ;; Assoc list of propositions added in substitutions, and during
	    ;; which time steps they are true (initially the emtpy list).
	    (let ((new-props '()))
	      (dolist (substitution substitutions)
		(dolist (prop (cdr substitution))
		  ;; (print "prop") (print prop)
		  (setq new-props
		    (adjoin (sprop-remove-time (sprop-nonnegated prop))
			    new-props
			    :test #'sprop-equal))))
	      (mapcar #'list new-props)))
	   (props-sans-added
	    ;; propositions, less those added by prop-substitute.
	    ;; by side-effect, add substituted clauses to added-props-when.
	    (remove-if #'(lambda (prop)
			   (let* ((timeless (sprop-remove-time prop))
				  (added-assoc (assoc timeless added-props-when
						      :test #'sprop-equal)))
			     ;; (print "timeless") (print timeless)
			     ;; (print "added-assoc") (print added-assoc)
			     (if added-assoc
				 (progn
				   (push (sprop-time prop)
					 (cdr added-assoc))
				   ;; (print "new added-assoc") (print added-assoc)
				   t)
			       (assoc (sprop-negate timeless) added-props-when
				      :test #'sprop-equal))))
		       propositions)))
      (if *debug-substitute* (format t "added-props-when ~s~%" added-props-when))
      (if *debug-substitute* (format t "props-sans-added ~s~%" props-sans-added))

      (dolist (substitution substitutions)
	(if *debug-substitute* (format t "substitution ~s~%" substitution))
	;; More efficient to exit early if intersection becomes empty,
	;; but this will do.
	(let* ((orig-prop (sprop-remove-time (car substitution)))
	       ;; When the substituted propositions are found negated
	       ;; and non-negated; "-orig" is probably gratuitous in name.
	       when-pos-origs when-neg-origs)
	  (dolist (added-prop (cdr substitution))
	    (if (sprop-negated-p added-prop)
		(push (cdr (assoc (sprop-remove-time
				   (sprop-negate added-prop)) ; certainly negated
				  added-props-when
				  :test #'sprop-equal))
		      when-neg-origs)
	      (push (cdr (assoc (sprop-remove-time added-prop)
				added-props-when
				:test #'sprop-equal))
		    when-pos-origs)))
	  (if *debug-substitute* (format t "when-pos-origs ~a when-neg-origs ~a~%"
					 when-pos-origs when-neg-origs))
	  ;; when-pos-origs is a list of times at which the substituted
	  ;; propositions are true -- one for each proposition which is
	  ;; substituted positively.  when-neg-origs is similar, for
	  ;; propositions which are substituted negatively.
	  (let* (;; list of times at which all positively substituted props true
		(when-all-pos-origs (if (null when-pos-origs)
					all-times
				      (reduce #'intersection when-pos-origs)))
		;; list of times at which any negatively substituted props true
		(when-any-neg-origs (reduce #'union when-neg-origs :initial-value '()))
		(when-all (set-difference when-all-pos-origs when-any-neg-origs)))
	    (if *debug-substitute* (format t "orig ~s when-all-pos-origs ~s when-any-neg-origs ~s when-all ~s~%"
					   orig-prop when-all-pos-origs when-any-neg-origs when-all))
	    (dolist (when when-all)
	      (push (sprop-add-time orig-prop when) props-sans-added)))))
       props-sans-added)))
