
;; code that checks whether a given action sequence is
;; a solution to a given UCPOP planning problem.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Unfortunately, this has diverged some from Nick's original version.
;; In an ideal world, the versions would be re-merged.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package medic)

;; nick kushmerick / nick@cs.washington.edu / 25.11.96

;; note that though this code takes as input UCPOP's
;; operator and problem representation, it is completely
;; independant of the UCPOP system per se.

;; usage: load this file, and then
;;
;;   (use-package 'uv)
;;   (verify domain problem plan)
;;
;; which returns t if the plan is a solution and nil
;; otherwise (and optionally prints some helpful output
;; if *verify-verbosely* is non-nil).
;; 
;; the form of the inputs is as follows:
;; 
;;  o  domain is a sexp of the form
;;          (define (domain ...)
;;             (:operator ...)
;;                  ...
;;             (:operator ...))
;;
;;  o  problem is a sexp of the form
;;          (define (problem ...)
;;              :domain 'domain
;;              :inits (...)
;;              :goal (...))
;;
;;  o  plan is a sexp of the form
;;          ((operator1 arg11 arg12 ...)
;;                  ...
;;           (operatorN argN1 argN2 ...))
;;
;; note that no check is done to ensure that the given problem
;; does indeed belong to the given domain.

;; a special interface has been established for the CLEAN file
;; format (see ucpop/domains/clean/):
;;
;;   (verify-clean probname plan)
;;
;; probname is the name of the problem, and plan is as before.

;; to verify a parallel plan, use
;;
;;   (verify-clean-parallel probname parallel-plan)
;;
;; probname is the name of the problem, and parallel-plan is a list of
;; lists of actions; in each list of actions, they may occur in any order.

;; BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT
;;
;;   not yet implemented:
;;    o  axioms
;;    o  safety whatever-they-ares
;;    o  facts
;;    o  FORALL with more than one quantified variable
;;         ie, you can't say (forall (?x ?y) GD)
;;         instead say (forall (?x) (forall (?y) GD))
;;
;; BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT BUT

;; this source code is best viewed on a window at least 100 characters wide

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *verify-verbosely* t)

(defvar *verify-max-plans* 100
  "Maximum number of parallel or alternate plans to verify exhaustively.
If there are more than this many plans, select this many randomly (with
replacement) to verify.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; also see verify-clean-parallel

(defun verify-clean (probname plan)
  (let* ((problem (get-problem-copy probname))
	 (domain (get-domain-copy (problem-domain problem))))
    (verify domain problem plan)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun verify (domain problem plan)
  (if (and plan (listp (first (first plan))))
      (error "attempt to verify parallel plan using (sequential) verify~%~a"
	     plan))
  (let ((operators (build-operators (domain-operators domain))))
    (multiple-value-bind (initial-state goal)
	(build-problem problem)
      (verify-trajectory initial-state goal plan operators))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *last-failed-GD* nil) ;; used for verbose output -- the "leaf" state-satisfies? fns below set this

(defun verify-trajectory (state GD plan operators)
  (when *verify-verbosely*
    (format t "*** Initial state: ~S~%" state))
  (flet ((check-goal ()
	   (when *verify-verbosely*
	     (format t "*** Final state: ~S~%" state))
	   (setf *last-failed-GD* nil)
	   (let ((satisfied? (state-satisfies? state GD nil)))
	     (when *verify-verbosely*
	       (format t "*** Goal is ")
	       (if satisfied?
		   (format t "satisfied")
		 (format t "NOT satisfied, because~%~5T~S~%is not satisfied" *last-failed-GD*))
	       (format t ".~%"))
	     satisfied?)))
    (dolist (action plan (check-goal))
      (when *verify-verbosely*
	(format t "*** Pushing ~S~%    through ~S~%" action state))
      (multiple-value-bind (next-state failed-precondition)
	  (push-action state action operators)
	(cond ((eq next-state :FAILURE)
	       (when *verify-verbosely*
		 (format t "*** FAILURE -- precondition~%~5T~S~%not satisfied~%" failed-precondition))
	       (return nil))
	      (t
	       (setf state next-state)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; building internal data structures from UCPOP (list-like) representation

(defun build-operators (domain-spec)
  (mapcar #'build-operator domain-spec))

(defun build-operator (operator-spec)
  (declare (type domains::operator operator-spec))
  (make-voperator
   :name (op-name operator-spec)
   :parameters (op-formals operator-spec)
   :precondition
   (let ((prec-spec (if (op-formal-types operator-spec)
			(list* :and (op-precondition operator-spec)
			       (mapcar #'(lambda (x)
					   (list (second x) (first x)))
				       (op-formal-types operator-spec)))
		      (op-precondition operator-spec))))
     (when prec-spec
       (build-GD prec-spec)))
   :effect (build-effect (op-effect operator-spec))))

(defun build-problem (problem-spec)
  ;; returns <inits, goal GD>
  (let ((init (flatten-ands (v-forall->and (problem-init problem-spec) problem-spec))))
    (values (if (eq (car init) :and) (cdr init) (list init))
	    (build-GD (problem-goal problem-spec)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; internal data structures

(defstruct GD)

(defstruct (simple-GD (:include GD))
  term)

(defmethod print-object ((GD simple-GD) stream)
  (format stream "#<GD ~S>" (simple-GD-term GD)))

(defstruct (combination-GD (:include GD))
  GDs)

(defstruct (and-GD (:include combination-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD and-GD) stream)
  (format stream "#<GD AND ~S>" (combination-GD-GDs GD)))

(defstruct (or-GD (:include combination-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD or-GD) stream)
  (format stream "#<GD OR ~S>" (combination-GD-GDs GD)))

(defstruct (not-GD (:include GD))
  GD)

(defmethod print-object ((GD not-GD) stream)
  (format stream "#<GD NOT ~S>" (not-GD-GD GD)))

(defstruct (imply-GD (:include GD))
  if
  then)

(defmethod print-object ((GD imply-GD) stream)
  (format stream "#<GD ~S ==> ~S>" (imply-GD-if GD) (imply-GD-then GD)))

(defstruct (quantified-GD (:include GD))
  scope
  GD)

(defstruct (forall-GD (:include quantified-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD forall-GD) stream)
  (format stream "#<GD FORALL ~S: ~S>" (quantified-GD-scope GD) (quantified-GD-GD GD)))

(defstruct (exists-GD (:include quantified-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD exists-GD) stream)
  (format stream "#<GD EXISTS ~S: ~S>" (quantified-GD-scope GD) (quantified-GD-GD GD)))

(defstruct (constraint-GD (:include GD))
  arg1
  arg2)

(defstruct (eq-GD (:include constraint-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD eq-GD) stream)
  (format stream "#<GD ~S EQ ~S>" (constraint-GD-arg1 GD) (constraint-GD-arg2 GD)))

(defstruct (neq-GD (:include constraint-GD))
  ) ;; nothing additional needed

(defmethod print-object ((GD neq-GD) stream)
  (format stream "#<GD ~S NEQ ~S>" (constraint-GD-arg1 GD) (constraint-GD-arg2 GD)))

(deftype GD-type ()
  '(or GD simple-GD combination-GD and-GD or-GD not-GD imply-GD
       quantified-GD forall-GD exists-GD constraint-GD eq-GD neq-GD))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun build-GD (spec)
  ;; (declare (notinline build-GD))
  ;; (format t "spec=~S (=NEQ? ~S)~%" spec (eq 'NEQ (first spec)))
  (let ((operator (string (first spec)))) ;; convert to string to avoid package problems
    (cond ((equalp operator "AND")
	   (make-and-GD :GDs (mapcar #'build-GD (rest spec))))
	  ((equalp operator "OR")
	   (make-or-GD :GDs (mapcar #'build-GD (rest spec))))
	  ((equalp operator "NOT")
	   (make-not-GD :GD (build-GD (second spec))))
	  ((equalp operator "IMPLY")
	   (make-imply-GD :if (build-GD (second spec)) :then (build-GD (third spec))))
	  ((equalp operator "FORALL")
	   (make-forall-GD :scope (second spec) :GD (and (third spec) (build-GD (third spec)))))
	  ((equalp operator "EXISTS")
	   (make-exists-GD :scope (second spec) :GD (and (third spec) (build-GD (third spec)))))
	  ((equalp operator "EQ")
	   (make-eq-GD :arg1 (second spec) :arg2 (third spec)))
	  ((equalp operator "NEQ")
	   (make-neq-GD :arg1 (second spec) :arg2 (third spec)))
	  (t
	   (make-simple-GD :term spec)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defstruct effect)

(defstruct (simple-effect (:include effect))
  term)

(defstruct (add-effect (:include simple-effect))
  ) ;; nothing additional needed

(defstruct (delete-effect (:include simple-effect))
  ) ;; nothing additional needed

(defstruct (compound-effect (:include effect))
  )

(defstruct (when-effect (:include compound-effect))
  GD
  effect)

(defstruct (and-effect (:include compound-effect))
  effects)

(defstruct (forall-effect (:include compound-effect))
  scope
  effect)

(deftype effect-type ()
  '(or effect simple-effect add-effect delete-effect
       compound-effect when-effect and-effect forall-effect))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun build-effect (spec)
  (let ((operator (string (first spec)))) ;; convert to string to avoid package problems
    (cond ((equalp operator "AND")
	   (make-and-effect :effects (mapcar #'build-effect (rest spec))))
	  ((equalp operator "WHEN")
	   (make-when-effect :GD (build-GD (second spec)) :effect (build-effect (third spec))))
	  ((equalp operator "FORALL")
	   (let ((scope (second spec)))
	     (when (and (vvariable? (first scope))
			(> (length scope) 1))
	       (error "FORALL-EFFECT scope must have exactly one item: ~S, ~S~%" scope spec))
	     (make-forall-effect :scope scope :effect (build-effect (third spec)))))
	  ((equalp operator "PRIMARY")
	   (build-effect (second spec)))
	  ((equalp operator "NOT")
	   (make-delete-effect :term (second spec)))
	  (t
	   (make-add-effect :term spec)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Helper function
(defun vrequired ()
  "If this ever gets called, it means something that was required was not
  supplied.  Use as default value for &key args or defstruct slots."
  (error "A required argument was not supplied."))

;; This use of the types should follow their definitions

(defstruct (voperator (:conc-name vop-))
  (name (vrequired) :type symbol)
  (parameters (vrequired) :type list)
  (precondition (vrequired) :type (or GD-type (member nil)))
  (effect (vrequired) :type effect-type))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; variables, bindings, unification

(defstruct (binding (:type list) (:constructor make-binding*))
  var
  val)

(defun make-binding (var val)
  (make-binding* :var var :val val))

(defun get-binding (var bindings)
  (binding-val (find var bindings :key #'binding-var)))

(defun vvariable? (symbol)
  (and (symbolp symbol)
       (eq #\? (elt (string symbol) 0))))

(defun v-instantiate (form bindings)
  (cond ((null form)
	 nil)
	((vvariable? form)
	 (get-binding form bindings))
	((consp form)
	 (cons
	  (v-instantiate (first form) bindings)
	  (v-instantiate (rest form) bindings)))
	(t
	 form)))

(defun baby-unify (pattern data)
  ;; warning not the most general unify in the world but good enough for our needs!
  ;; assumes pattern contains exactly one variable.
  ;; returns a value for that variable if they match and nil otherwise
  (let ((value nil))
    (and (= (length pattern) (length data))
	 (every #'(lambda (x y)
		    (cond ((vvariable? x)
			   (when value
			     (warn "BABY-UNIFY expecting exactly one variable but found more than one: ~S, ~S; ~S, ~S"
				   x y pattern data))
			   (setq value y)
			   t)
			  ((eq x y))))
		pattern
		data)
	 (or value
	     (warn "BABY-UNIFY expecting exactly one variable but found none: ~S, ~S~" pattern data)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; the main event: pushing an action through a state

(defun push-action (state action voperators)
  (assert (every #'voperator-p voperators))
  (let ((voperator (find (first action) voperators :key #'vop-name)))
    (assert voperator () "couldn't find operator ~a in list ~a"
	    (first action) (mapcar #'vop-name voperators))
    (let ((bindings (mapcar #'make-binding (vop-parameters voperator) (rest action))))
      (assert (= (length (vop-parameters voperator)) (length (rest action)))
	      ()
	      "Action ~s was given ~d args, expects ~d." (first action)
	      (length (vop-parameters voperator)) (length (rest action)))
      (let ((precondition (vop-precondition voperator)))
	(setf *last-failed-GD* nil)
	(values
	 (if (state-satisfies? state precondition bindings)
	     (let ((simple-effects
		    (collect-effects state (vop-effect voperator) bindings)))
	       (mapc #'(lambda (eff)
			 (setq state (funcall (car eff) state (cdr eff))))
		     simple-effects)
	       state)
	   :FAILURE)
	 *last-failed-GD*)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; does a state satisfy a GD?

(defgeneric state-satisfies? (state GD bindings))

(defmethod state-satisfies? (state (GD (eql NIL)) bindings)
  ;; action has no precondition
  (declare (ignore state bindings))
  T)

(defmethod state-satisfies? (state (GD simple-GD) bindings)
  (cond ((state-obtains? state (v-instantiate (simple-GD-term GD) bindings)))
	(t
	 (setf *last-failed-GD* GD)
	 NIL)))

(defmethod state-satisfies? (state (GD and-GD) bindings)
  (combination-GD-satisfies? state GD bindings #'every))

(defmethod state-satisfies? (state (GD or-GD) bindings)
  (combination-GD-satisfies? state GD bindings #'some))

(defun combination-GD-satisfies? (state GD bindings combiner)
  (funcall combiner
	   #'(lambda (subGD) (state-satisfies? state subGD bindings))
	   (combination-GD-GDs GD)))

(defmethod state-satisfies? (state (GD not-GD) bindings)
  (cond ((not (state-satisfies? state (not-GD-GD GD) bindings)))
	(t
	 (setf *last-failed-GD* GD)
	 NIL)))

(defmethod state-satisfies? (state (GD imply-GD) bindings)
  (or (state-satisfies? state (imply-GD-then GD) bindings)
      (not (state-satisfies? state (imply-GD-if GD) bindings))))

(defmethod state-satisfies? (state (GD forall-GD) bindings)
  (cond ((quantified-GD-state-satisfies? state GD bindings #'every))
	(t
	 (setf *last-failed-GD* GD)
	 NIL)))

(defmethod state-satisfies? (state (GD exists-GD) bindings)
  (cond ((quantified-GD-state-satisfies? state GD bindings #'some))
	(t
	 (setf *last-failed-GD* GD)
	 NIL)))

(defun quantified-GD-state-satisfies? (state GD bindings combiner)
  ;; TO DO: this code could stand to be integrated with that of FORALL-EFFECT since it is so similar
  (let* ((scope
	  (quantified-GD-scope GD))
	 (subGD
	  (quantified-GD-GD GD))
	 (special-case-scope?
	  ;; special case for (exists (?x) ...)
	  (= (length scope) 1))
	 (var ;; the variable this quantifier quantifies over
	  (if special-case-scope?
	      (first scope)
	    ;; regular case: (forall (P ?x ?y ?z) ...) or (exists (P ?x ?y ?z) ...)
	    (find-if #'(lambda (arg)
			 (and (vvariable? arg)
			      (not (member arg bindings :key #'binding-var))))
		     (cdr scope)))) ;; ignore the 'P'
	 (vals ;; the values to which it can be bound
	  (remove-duplicates
	   (apply #'append
		  (mapcar
		   #'(lambda (term)
		       (let ((val
			      (baby-unify
			       (if special-case-scope?
				   (list (first term) var)
				 scope)
			       (v-instantiate term bindings))))
			 (when val (list val))))
		   state)))))
    (or
     ;; special case to allow (not (exists (on ?x b)))
     (and (null subGD)
	  vals)
     ;; regular case -- try all th variables, stopping when quantifier allows it
     (funcall
      combiner
      #'(lambda (val)
	  (state-satisfies? state subGD (cons (make-binding var val) bindings)))
      vals))))

(defmethod state-satisfies? (state (GD eq-GD) bindings)
  (declare (ignore state))
  (constraint-GD-satisfied? GD bindings))

(defmethod state-satisfies? (state (GD neq-GD) bindings)
  (declare (ignore state))
  (not (constraint-GD-satisfied? GD bindings)))

(defun constraint-GD-satisfied? (GD bindings)
  (cond ((eq (v-instantiate (constraint-GD-arg1 GD) bindings)
	     (v-instantiate (constraint-GD-arg2 GD) bindings)))
	(t
	 (setf *last-failed-GD* GD)
	 NIL)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; state primitives

(defun state-obtains? (state term)
  (member term state :test #'equal))

(defun state-add (state term)
  (if (state-obtains? state term)
      state
    (cons term state)))

(defun state-delete (state term)
  (if (state-obtains? state term)
      (remove term state :test #'equal)
    state))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; push effect through a state

(defgeneric push-effect (state effect bindings))

(defmethod push-effect (state (effect add-effect) bindings)
  (push-simple-effect state #'state-add effect bindings))

(defmethod push-effect (state (effect delete-effect) bindings)
  (push-simple-effect state #'state-delete effect bindings))

(defun push-simple-effect (state pusher effect bindings)
  (funcall pusher state (v-instantiate (simple-effect-term effect) bindings)))


;; Returns a list of (function . arg) conses; iterate over the list calling
;; (function STATE arg) which returns a new state (to be passed to the next
;; invocation).
(defgeneric collect-effects (state effect bindings))

(defmethod collect-effects (state (effect add-effect) bindings)
  (declare (ignore state))
  (list
   (cons #'state-add (v-instantiate (simple-effect-term effect) bindings))))

(defmethod collect-effects (state (effect delete-effect) bindings)
  (declare (ignore state))
  (list
   (cons #'state-delete (v-instantiate (simple-effect-term effect) bindings))))

(defmethod collect-effects (state (effect and-effect) bindings)
  ;; 'reduce' would be lispier?
  (mapcan #'(lambda (subeffect)
	      (collect-effects state subeffect bindings))
	  (and-effect-effects effect)))

(defmethod collect-effects (state (effect when-effect) bindings)
  (if (state-satisfies? state (when-effect-GD effect) bindings)
      (collect-effects state (when-effect-effect effect) bindings)
    '()))

(defmethod collect-effects (state (effect forall-effect) bindings)
  ;; TO DO: this code could stand to be integrated with that of {FORALL,EXISTS}-GD since it is so similar
  (let* ((scope
	  (let ((s (forall-effect-scope effect)))
	    ;; three possibilities: (P ?x) ((P ?x)) ((P ?x) (Q ?y))
	    ;; convert the second to the first; we don't deal with the third yet
	    (cond ((symbolp (car s))
		   s)
		  ((and (listp (car s))
			(= 1 (length s)))
		   (car s))
		  (t (error "Don't understand multiple quantifier bindings ~s" s)))))
	 (subeffect
	  (forall-effect-effect effect))
	 (var ;; the variable this quantifier quantifies over
	  (if (vvariable? (first scope))
	      (first scope)
	    (find-if #'(lambda (arg)
			 (and (vvariable? arg)
			      (not (member arg bindings :key #'binding-var))))
		     (cdr scope)))) ;; (P ?x ?y ?z) -> (?x ?y ?z): ignore predicate slot
	 (vals ;; the values to which it can be bound
	  (remove-duplicates
	   (apply #'append
		  (mapcar
		   (if (vvariable? (first scope))
		       ;; we're dealing with effect of the form (forall (?x) effect)
		       ;; so collect up all possible bindings for ?x
		       #'cdr
		     ;; we're dealing with effect of the form (forall (P .... ?x ...) effect)
		     ;; so collect up all possible bindings for ?x by trying to unify
		     ;; (P ... ?x ...) against all elements of the state
		     #'(lambda (term)
			 (let ((val (baby-unify scope (v-instantiate term bindings))))
			   (when val (list val)))))
		   state)))))
    (mapcan #'(lambda (val)
		(collect-effects state subeffect (cons (make-binding var val) bindings)))
	  vals)))


;;; This old implementation orders the effects and does them one after
;;; another, which is wrong.  Consider the effect
;;;  (:and (:when FOO (:not FOO)) (when (:not FOO) FOO))
;;; The old implementation left everything FOO, nothing (:not FOO).
;;; I'm not sure whether it gave better error messages, though.

;; (defgeneric push-effect (state effect bindings))
;; 
;; (defmethod push-effect (state (effect add-effect) bindings)
;;   (push-simple-effect state #'state-add effect bindings))
;; 
;; (defmethod push-effect (state (effect delete-effect) bindings)
;;   (push-simple-effect state #'state-delete effect bindings))
;; 
;; (defun push-simple-effect (state pusher effect bindings)
;;   (funcall pusher state (v-instantiate (simple-effect-term effect) bindings)))
;; 
;; (defmethod push-effect (state (effect and-effect) bindings)
;;   ;; 'reduce' would be lispier?
;;   (mapc #'(lambda (subeffect)
;; 	    (setf state (push-effect state subeffect bindings)))
;; 	(and-effect-effects effect))
;;   state)
;; 
;; (defmethod push-effect (state (effect when-effect) bindings)
;;   (if (state-satisfies? state (when-effect-GD effect) bindings)
;;       (push-effect state (when-effect-effect effect) bindings)
;;     state))
;; 
;; (defmethod push-effect (state (effect forall-effect) bindings)
;;   ;; TO DO: this code could stand to be integrated with that of {FORALL,EXISTS}-GD since it is so similar
;;   (let* ((scope
;; 	  (let ((s (forall-effect-scope effect)))
;; 	    ;; three possibilities: (P ?x) ((P ?x)) ((P ?x) (Q ?y))
;; 	    ;; convert the second to the first; we don't deal with the third yet
;; 	    (cond ((symbolp (car s))
;; 		   s)
;; 		  ((and (listp (car s))
;; 			(= 1 (length s)))
;; 		   (car s))
;; 		  (t (error "Don't understand multiple quantifier bindings ~s" s)))))
;; 	 (subeffect
;; 	  (forall-effect-effect effect))
;; 	 (var ;; the variable this quantifier quantifies over
;; 	  (if (vvariable? (first scope))
;; 	      (first scope)
;; 	    (find-if #'(lambda (arg)
;; 			 (and (vvariable? arg)
;; 			      (not (member arg bindings :key #'binding-var))))
;; 		     (cdr scope)))) ;; (P ?x ?y ?z) -> (?x ?y ?z): ignore predicate slot
;; 	 (vals ;; the values to which it can be bound
;; 	  (remove-duplicates
;; 	   (apply #'append
;; 		  (mapcar
;; 		   (if (vvariable? (first scope))
;; 		       ;; we're dealing with effect of the form (forall (?x) effect)
;; 		       ;; so collect up all possible bindings for ?x
;; 		       #'cdr
;; 		     ;; we're dealing with effect of the form (forall (P .... ?x ...) effect)
;; 		     ;; so collect up all possible bindings for ?x by trying to unify
;; 		     ;; (P ... ?x ...) against all elements of the state
;; 		     #'(lambda (term)
;; 			 (let ((val (baby-unify scope (v-instantiate term bindings))))
;; 			   (when val (list val)))))
;; 		   state)))))
;;     (mapc #'(lambda (val)
;; 	      (setf state (push-effect state subeffect (cons (make-binding var val) bindings))))
;; 	  vals)
;;     state))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; utilities

(defun extract-named-slot (slotid spec)
  (second (member slotid spec)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;; parallel-action extension

;; Verify parallel plans by verifying all their linearizations.
;; PPLAN is a list of lists of actions; in each list of actions, they may
;; occur in any order.  This is inefficient -- for instance, it redoes file
;; IO over and over -- but it works.
;; Returns two values; if first is nil, second is the failing linearization
(defun verify-clean-parallel (probname pplan)
  ;; (every #'verify-clean probname (parallel-plan->linear-plans pplan))
  (let* ((num-plans (apply #'* (mapcar #'factorial (mapcar #'length pplan))))
	 (verified-plans (if (<= num-plans *verify-max-plans*)
			     (parallel-plan->linear-plans pplan)
			   ;; Needn't really collect them all ahead of time,
			   ;; but this let me reuse the code.
			   (loop for i from 1 to *verify-max-plans*
				 collect (parallel-plan->random-linear-plan pplan)))))
    (assert (= (length verified-plans) (min num-plans *verify-max-plans*)))
    (if *verify-verbosely*
	(if (<= num-plans *verify-max-plans*)
	    (format t "~&Verifying ~d linearizations~%" num-plans)
	  (format t "~&Verifying ~d linearizations (out of ~d)~%"
		  *verify-max-plans* num-plans)))
    (let ((failing-plan
	   (dolist (lplan verified-plans nil)
	     (if (not (verify-clean probname lplan))
		 (return lplan)))))
      (if failing-plan
	  (values nil failing-plan)
	(values t nil)))))

(defun factorial (n)
  (if (< n 3)
      n
    (* n (factorial (- n 1)))))

;; Returns a list of linear (completely ordered) plans consistent with the
;; parallel plan, which is a list of lists of actions.
(defun parallel-plan->linear-plans (pplan)
  (cond ((null pplan)
	 '())
	(t
	 ;; (format t "pplan not null ~s~%" pplan)
	 (let* ((preplans (permutations (car pplan)))
		(postplans (parallel-plan->linear-plans (cdr pplan)))
		(all-plans (if (null postplans)
			       preplans
			     (mapcan #'(lambda (preplan)
					 (mapcar #'(lambda (postplan)
						     (append preplan postplan))
						 postplans))
				     preplans))))
	   ;; (format t "pplan ~s preplans ~s postplans ~s all-plans ~s~%"
	   ;;   pplan preplans postplans all-plans)
	   all-plans))))
;; (parallel-plan->linear-plans '((a b c) (d) (e f)))

;; Returns one random linearization of the plan
(defun parallel-plan->random-linear-plan (pplan)
  (apply #'append (mapcar #'random-permutation pplan)))



;; Verify alternate plans by verifying all their linearizations.
;; APLAN is a list of lists of actions; in each list of actions, any one may
;; occur.  This is inefficient -- for instance, it redoes file
;; IO over and over -- but it works.
;; Returns two values; if first is nil, second is the failing linearization

(defun verify-clean-alternates (probname aplan)
  (let* ((num-plans (apply #'* (mapcar #'length aplan)))
	 (verified-plans (mapcar #'(lambda (l) (remove *no-op-action* l :test #'equal))
				 (if (<= num-plans *verify-max-plans*)
				     (alternate-plan->linear-plans aplan)
				   ;; Needn't really collect them all ahead of
				   ;; time, but this let me reuse the code.
				   (loop for i from 1 to *verify-max-plans*
					 collect (alternate-plan->random-linear-plan aplan))))))
    (assert (= (length verified-plans) (min num-plans *verify-max-plans*)))
    (if *verify-verbosely*
	(if (<= num-plans *verify-max-plans*)
	    (format t "~&Verifying ~d linearizations~%" num-plans)
	  (format t "~&Verifying ~d linearizations (out of ~d)~%"
		  *verify-max-plans* num-plans)))
    (let ((failing-plan
	   (dolist (lplan verified-plans nil)
	     (if (not (verify-clean probname lplan))
		 (return lplan)))))
      (if failing-plan
	  (values nil failing-plan)
	(values t nil)))))

;; Returns a list of linear (completely ordered) plans consistent with the
;; alternate plan, which is a list of lists of actions.
(defun alternate-plan->linear-plans (aplan)
  (cond ((null aplan)
	 '())
	(t
	 ;; (format t "aplan not null ~s~%" aplan)
	 (let* ((preplans (mapcar #'list (car aplan)))
		(postplans (alternate-plan->linear-plans (cdr aplan)))
		(all-plans (if (null postplans)
			       preplans
			     (mapcan #'(lambda (preplan)
					 (mapcar #'(lambda (postplan)
						     (append preplan postplan))
						 postplans))
				     preplans))))
	   ;; (format t "aplan ~s preplans ~s postplans ~s all-plans ~s~%"
	   ;;   aplan preplans postplans all-plans)
	   all-plans))))
;; (alternate-plan->linear-plans '((a b c) (d) (e f)))

;; Returns one random linearization of the plan
(defun alternate-plan->random-linear-plan (pplan)
  (apply #'list (mapcar #'random-element pplan)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Converts a forall into a conjunction of all appropriate objects
(defun v-forall->and (form prob)
  ;; base case: if form is a symbol, return it
  (if (symbolp form)
      form
    (let ((fn (car form)))  ;; The first element in form
      (cond
       ;; if we have anything bu a forall, leave it alone
       ;; and recursively convert the pieces
       ((or (eq :and fn) (eq :or fn) (eq :exists fn) (eq :not fn))
	(cons fn (mapcar #'(lambda (f) (v-forall->and f prob))
			 (cdr form))))
       ;; If we do have a forall, then we find the quantified variable
       ;; and replace it in the expression for each object that
       ;; fits (either all things or all tools, frinstance)
       ;; and conjoin the results into conj-list
       ((eq :forall fn)
	;; A(x)(x is blue)
	;; variable-name is 'x'
	;; things-to-instantiate are all the obejects that could fit x
	(let ((forall-bindings (second form))
	      ;; work from the inside out to get nested scoping right
	      (forall-body (v-forall->and (third form) prob)))
	  (dolist (var+type (v-parse-bindings forall-bindings))
	    (let ((variable-name (first var+type))
		  (things-to-instantiate
		   (objects-of-type (second var+type) prob)))
	      (setq forall-body
		    ;; now we just need to dolist things-to-instantiate, replacing
		    ;; into what each time, then make all that an :and
		    (cons :and (mapcar #'(lambda (item)
					   (subst item variable-name forall-body))
				       things-to-instantiate)))))
	  forall-body))
       (t form)))))


;; Tests
;; (v-forall->and   '(:and (:not (jacked ?h)) (:forall (domains::?y) (:when (on domains::?y ?h) (on domains::?y ground))) (have jack)) (get-problem-copy 'fixit2))
;; (v-forall->and '(:forall (DOMAINS::tool ?x) (in ?x boot)) (get-problem-copy 'fixit2))
;; (v-forall->and '(:AND (:AND (:NOT (JACKED ?H)) (:FORALL (?Y)) (HAVE JACK)) (ON ?Y GROUND)) (get-problem-copy 'fixit2 ))
;; (v-forall->and '(:forall (domains::person domains::?p) (:exists (object ?c) (:and (written-for domains::?p ?c)))) (get-problem-copy 'office1))
;; (v-forall->and '(:forall (?x) (:and (eq ?x y) (:forall (domains::person ?p) (foo ?p ?x)))) (get-problem-copy 'office1))



;; ;; Inits is the problem-inits of the current problem
;; ;; what-to-replace should be of the form nil  or 'tool
;; ;; In the former candidates is a list of all the objects in the
;; ;; universe.  In the latter case it is a list of all the tools
(defun candidates (inits what-to-replace)
  (assert (listp inits))
  (assert (equal :and (car inits)))
  (let ( (criterion ()) (return-list ()))
    (if  (equal () what-to-replace)
	(setf criterion 'anything)
      (setf criterion what-to-replace))
    ;; criterion is the property which  objects in the
    ;; inits need to have ( e.g. tool, block, person)
    ;; or 'anything for all objects...
    ;; to be added to the return-list
    (format t "~%criterion is ~A" criterion)
    (dolist (an-item (cdr inits))
      (when (or (eq (car an-item) criterion) (and (equal 'anything criterion) (equal 2 (length an-item))))
	(setf return-list (adjoin  (second an-item) return-list :test #'equal)))
      (format t "~%~A" return-list)
      )
    (return-from candidates return-list)
    ))
;; ;; Tests
;; ;;  (candidates '((tree a) (Bush b) (Tree c) (rock d) (green a)) 'tree)


;; Not much use anymore, just a way to call candidates...
(defun v-objects-of-type (type problem)
  (candidates (problem-init (get-problem-copy problem)) type)
)
;; Tests
;; (v-objects-of-type nil 'sussman-anomaly)
;; (v-objects-of-type 'domains::place 'get-paid)

;; Returns a list of (varname type), or nil if the (single-variable)
;; binding can't be parsed.
(defun v-parse-binding (binding)
  (cond ((symbolp binding)
	 (and (varp binding)
	      (list binding nil)))
	((and (= 1 (length binding))
	      (varp (car binding)))
	 (list (car binding) nil))
	((and (= 2 (length binding))
	      (symbolp (car binding))
	      (not (varp (car binding)))
	      (varp (second binding)))
	 (list (second binding) (first binding)))
	(t
	 nil)))

;; Returns a list (not a cons) of (varname type) lists, where TYPE may be nil.
;; Will need to be extended for the Nisp-style syntax.
(defun v-parse-bindings (forall-bindings)
  (let ((as-single-binding (v-parse-binding forall-bindings)))
    (if as-single-binding
	(list as-single-binding)
      (let ((result (mapcar #'v-parse-binding forall-bindings)))
	(if (member nil result :test #'eq)
	    (error "Can't parse bindings ~s" forall-bindings))
	result))))
;; Testing
;; (v-parse-bindings '?x)
;; (v-parse-bindings '(?x))
;; (v-parse-bindings '(?x ?y))
;; (v-parse-bindings '(tool ?x))
;; (v-parse-bindings '(?x ?y ?z))
;; (v-parse-bindings '(tool ?x ?y))      ; error
;; (v-parse-bindings '((tool ?x)))
;; (v-parse-bindings '(?x (?y) (tool ?z) (boot ?w)))
;; (v-parse-bindings '(?x (?y) (?z ?z))) ; error
;; (v-parse-bindings '(?x (?y) ((?z ?z)))) ; error
